#!/usr/local/bin/perl -wT
use strict;
#
# This is an automatic installation script for NMS TFmail
#
# You may need to replace "/usr/local/bin/perl" on the first line of
# this script with a different path.  Your hosting provider or system
# administrator will be able to tell you the correct path to the Perl
# interpreter for your system.
#
# You should upload this script in ASCII mode, and you may need to
# enable execute permission after uploading it.
#
# Before you upload this script, you must set a password.  Change the
# string "password" below to some secret word known only to you.  It
# is very important to make this password hard to guess, since anyone
# who knows it will be able to break into your web site, take full
# control of your pages and use your web server to attack other hosts
# on the internet.
my $pw = 'password';
#
# If this script fails to run and gives an error message that looks
# like "too late for -T" then you should delete the 'T' from the end
# of the first line of the script.
#
######################################################################
#
#  NO USER SERVICEABLE PARTS BELOW THIS POINT
#
######################################################################
#

use CGI;
use Fcntl ':flock';
use IO::File;

use vars qw(%Scripts %Modules);

%Scripts = (

    'GBview' => <<'NMSAI_END_CGI_SCRIPT',
#!/usr/bin/perl -wT
use strict;
#
# $Id: GBview.pl,v 1.5 2004/10/12 08:37:21 gellyfish Exp $
#
# USER CONFIGURATION SECTION
# --------------------------
# Modify these to your own settings, see the README file
# for detailed instructions.

use constant DEBUGGING      => 1;
use constant LIBDIR         => '.';
use constant CONFIG_ROOT    => '.';
use constant MAX_DEPTH      => 0;
use constant CONFIG_EXT     => '.trc';
use constant TEMPLATE_EXT   => '.trt';
use constant HTMLFILE_ROOT  => '';
use constant HTMLFILE_EXT   => '.html';
use constant CHARSET        => 'iso-8859-1';

# USER CONFIGURATION << END >>
# ----------------------------
# (no user serviceable parts beyond here)

=head1 NAME

GBview.pl - viewer script for a guestbook generated by TFmail.pl

=head1 DESCRIPTION

This CGI script reads a file in an XML-like format generated by
TFmail.pl, formats the data in the file using a template and
outputs it.

See the C<ADVANCED GUESTBOOK> section near the end of the
F<README> file for more details.

=cut

use Fcntl ':flock';
use CGI;
use lib LIBDIR;
use NMStreq;

BEGIN
{
  use vars qw($VERSION);
  $VERSION = substr q$Revision: 1.5 $, 10, -1;
}

delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};
$ENV{PATH} =~ /(.*)/ and $ENV{PATH} = $1;

use vars qw($done_headers);
$done_headers = 0;

#
# We want to trap die() calls, output an error page and
# then do another die() so that the script aborts and the
# message gets into the server's error log.  If there is
# already a __DIE__ handler installed then we must
# respect it on our final die() call.
#
eval { local $SIG{__DIE__} ; main() };
if ($@)
{
   my $message = $@;
   error_page($message);
   die($message);
}

sub main
{
   my $treq = NMStreq->new(
      ConfigRoot    => CONFIG_ROOT,
      MaxDepth      => MAX_DEPTH,
      ConfigExt     => CONFIG_EXT,
      TemplateExt   => TEMPLATE_EXT,
      EnableUploads => 0,
      CGIPostMax    => 10000,
      Charset       => CHARSET,
   );

   if ( HTMLFILE_ROOT eq '' )
   {
      die "No HTMLFILE_ROOT set, nothing for this script to do\n";
   }

   my $htmlfile = $treq->config('gbview_htmlfile', 'guestbook');
   unless ( $htmlfile =~ /^([\w\-\/]+)$/ )
   {
      die "bad gbview_htmlfile value [$htmlfile] in config file";
   }
   $htmlfile = "@{[  HTMLFILE_ROOT ]}/$1@{[ HTMLFILE_EXT ]}";
    
   open LOCK, ">>$htmlfile.lck" or die "open >>$htmlfile.lck: $!";
   flock LOCK, LOCK_SH or die "flock $htmlfile.lck: $!";

   open IN, "<$htmlfile" or die "open $htmlfile: $!";
   my @entries;
   { local $/ = '</entry>' ; @entries = grep m#<entry>#, <IN> };
   close IN;

   close LOCK;

   unless ( $treq->config('gbview_oldest_first', 0) )
   {
      @entries = reverse @entries;
   }

   my $total = scalar @entries;

   my $startat = $treq->param('startat') || 0;
   $startat =~ /^(\d{1,4})$/ or die "bad startat value [$startat]\n";
   $startat = $1;
   splice @entries, 0, $startat;

   my $perpage = $treq->config('gbview_perpage', 0);
   $perpage =~ /^(\d{1,4})$/ or die "bad gbview_perpage value [$perpage] in config file\n";
   $perpage = $1;
   my ($page_count, $this_is_page) = (1, 1);
   if ($perpage > 0)
   {
      splice @entries, $perpage;
      $page_count   = int( ($total-1) / $perpage ) + 1;
      $this_is_page = int ( $startat / $perpage ) + 1;
   }

   $treq->install_foreach('entry', [map {extract_values($_)} @entries]);

   $treq->install_directive('can_go_back', ($startat > 0 ? 1 : 0));
   $treq->install_directive('can_go_on',   ($startat+$perpage <= $total-1 ? 1 : 0));
   $treq->install_directive('prev_page_start', $startat - $perpage);
   $treq->install_directive('next_page_start', $startat + $perpage);
   $treq->install_directive('page_count', $page_count);
   $treq->install_directive('this_is_page', $this_is_page);
   $treq->install_directive('multiple_pages', ($page_count > 1 ? 1 : 0) );

   $treq->install_foreach('page', [
      map {{  
         page  => $_,
         this  => ($_ == $this_is_page ? 1 : 0),
         start => ($_-1) * $perpage,
      }} (1..$page_count)
   ]);

   my $template = $treq->config('gbview_template', 'gbview');
   html_page($treq, $template);
}

=head1 INTERNAL FUNCTIONS

=over

=item extract_values ( STRING )

Converts a string consisting of named values encoded in an XML-like
format into a hashref.  The format is:

   <value name="$name">$value</value>

Where C<$name> and C<$value> are the strings that end up as keys and
values in the hash.  Stores a reference to the value rather than the
value itself, to prevent HTML metacharacters in the values from being
escaped when they're displayed via a template.  This is nessessary
to avoid double escaping, since HTML metacharacters were escaped by
TFmail.pl when the data was written to the file.

=cut

sub extract_values
{
   my ($string) = @_;

   my %hash;
   while ( $string =~ m#<value name="([\w\-\.\/]+)">(.*?)</value>#sg )
   {
      my ($key, $val) = ($1, $2);
      $hash{$key} = \$val;
   }

   return \%hash;
}
        
=item html_page ( TREQ, TEMPLATE )

Outputs an HTML page using the template TEMPLATE.

=cut

sub html_page
{
   my ($treq, $template) = @_;

   html_header();
   $done_headers = 1;

   $treq->process_template($template, 'html', \*STDOUT);
}

=item error_page ( MESSAGE )

Displays an "S<Application Error>" page, without using a
template since the error may have arisen during template
resolution.

=cut

sub error_page
{
   my ($message) = @_;

   unless ( $done_headers )
   {
      html_header();
      print <<EOERR;
<?xml version="1.0" encoding="@{[ CHARSET ]}"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>Error</title>
  </head>
  <body>
EOERR

      $done_headers = 1;
   }

   if ( DEBUGGING )
   {
      $message = NMSCharset->new(CHARSET)->escape($message);
      $message = "<p>$message</p>";
   }
   else
   {
      $message = '';
   }

   print <<EOERR;
    <h1>Application Error</h1>
    <p>
     An error has occurred in the program
    </p>
    $message
  </body>
</html>
EOERR
}

=item html_header ()

Outputs the CGI header using a content-type of text/html.

=cut

sub html_header {
    if ($CGI::VERSION >= 2.57) {
        # This is the correct way to set the charset
        print CGI::header('-type'=>'text/html', '-charset'=>CHARSET);
    }
    else {
        # However CGI.pm older than version 2.57 doesn't have the
        # -charset option so we cheat:
        print CGI::header('-type' => "text/html; charset=@{[ CHARSET ]}");
    }
}

=back

=head1 MAINTAINERS

The NMS project, E<lt>http://nms-cgi.sourceforge.net/E<gt>

To request support or report bugs, please email
E<lt>nms-cgi-support@lists.sourceforge.netE<gt>

=head1 COPYRIGHT

Copyright 2002 - 2004 London Perl Mongers, All rights reserved

=head1 LICENSE

This script is free software; you are free to redistribute it
and/or modify it under the same terms as Perl itself.

=cut


NMSAI_END_CGI_SCRIPT

    'Config' => <<'NMSAI_END_CGI_SCRIPT',
#!/usr/bin/perl -wT
use strict;
#
# $Id: TFmail_config.pl,v 1.4 2004/07/28 13:57:14 gellyfish Exp $
#
# USER CONFIGURATION SECTION
# --------------------------

# Set these to same values that you used in TFmail.pl
use constant DEBUGGING      => 1;
use constant LIBDIR         => '.';
use constant CONFIG_ROOT    => '.';
use constant CONFIG_EXT     => '.trc';
use constant LOGFILE_ROOT   => '';
use constant LOGFILE_EXT    => '.log';
use constant CHARSET        => 'iso-8859-1';

# The file that the script should use as a lock file and a
# place to count failed password attempts.  This should not
# be the name of an existing file - the script will create
# the file itself.
use constant LOCKFILE       => '.lock';

# Set the password to some word or phrase that nobody will be
# able to guess.  It's important to choose a strong password
# as anybody who guesses it will be able to take control of
# your web site and use your web server to attack other hosts
# on the internet.
use constant PASSWORD       => 'password';

# USER CONFIGURATION << END >>
# ----------------------------
# (no user serviceable parts beyond here)

=head1 NAME

TFmail_config.pl - TFmail configuration editor

=head1 DESCRIPTION

This CGI script provides basic TFmail configuration file editing
facilities.  All templates are inlined, so each configuration is
a single file.

=cut

use Fcntl ':flock';
use lib LIBDIR;
use NMSCharset;
use CGI qw(:standard);
use IO::File;

BEGIN
{
   use vars qw($VERSION);
   $VERSION = substr q$Revision: 1.4 $, 10, -1;
}

delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};
$ENV{PATH} =~ /(.*)/ and $ENV{PATH} = $1;

use vars qw($done_headers);
$done_headers = 0;

use vars qw($cs $cs_strip_nonprint);
$cs = NMSCharset->new(CHARSET);
$cs_strip_nonprint = $cs->strip_nonprint_coderef;
sub strip_nonprint { &{ $cs_strip_nonprint }(@_); }

#
# We want to trap die() calls, output an error page and
# then do another die() so that the script aborts and the
# message gets into the server's error log.  If there is
# already a __DIE__ handler installed then we must
# respect it on our final die() call.
#
eval { local $SIG{__DIE__} ; main() };
if ($@)
{
   my $message = $@;
   error_page($message);
   die($message);
}

sub main
{
   local ($CGI::DISABLE_UPLOADS, $CGI::POST_MAX);
   $CGI::DISABLE_UPLOADS = 1;
   $CGI::POST_MAX        = 100000;

   my $cgi = CGI->new;
   my $gotpass = strip_nonprint( $cgi->param('password') || '' );

   if ( PASSWORD =~ /pa[s5][s5]|w[o0]rd/i )
   {
      die "Password too similar to 'password', aborting\n"
   }
   unless ( $ENV{REQUEST_METHOD} eq 'POST' and length $gotpass )
   {
      html_header();
      print <<END;
<?xml version="1.0" encoding="@{[ CHARSET ]}"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>Password Required</title>
 </head>
 <body>
  <form method="post">
   <p>Please enter the TFmail configuration password.</p>
   <p><input type="password" name="password" /></p>
   <p><input type="submit" value="Continue" /></p>
  </form>
 </body>
</html>
END
      exit;
   }
    
   my $flock = IO::File->new('>>'.LOCKFILE);
   defined $flock or die "open >>@{[ LOCKFILE ]}: $!";
   flock $flock, LOCK_EX or die "flock @{[ LOCKFILE ]}: $!";

   if (-s LOCKFILE > 5)
   {
       html_header();
       print <<END;
<?xml version="1.0" encoding="@{[ CHARSET ]}"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>Too many incorrect passwords</title>
 </head>
 <body>
  <p>
   This script is disabled because there have been too many
   incorrect password attempts.  To fix this, delete the lock
   file from the server.  The LOCKFILE configuration constant
   in the script holds the lock file location.
  </p>
 </body>
</html>
END
      exit;
   }

   if ($gotpass ne PASSWORD)
   { 
      sleep 1;
      seek $flock, 0, 2 or die "seek: $!";
      print $flock "x";
      close $flock;
      html_header();
      print <<END;
<?xml version="1.0" encoding="@{[ CHARSET ]}"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>Password Incorrect</title>
 </head>
 <body>
  <form method="post">
   <p>You entered an incorrect password, please try again.</p>
   <p><input type="password" name="password" /></p>
   <p><input type="submit" value="Continue" /></p>
  </form>
 </body>
</html>
END
      exit;
   }

   truncate LOCK, 0;
   
   my $viewlog = $cgi->param('viewlog') || '';
   my $getlog  = $cgi->param('getlog')  || '';

   if (length $viewlog)
   {
      $viewlog =~ /^([\w\-]{1,100})$/ or die "bad log name [$viewlog]";
      send_log($1, 0);
      return;
   }
   elsif (length $getlog)
   {
      $getlog =~ /^([\w\-]{1,100})$/ or die "bad log name [$getlog]";
      send_log($1, 1);
      return;
   }
      
   my $name = strip_nonprint( $cgi->param('config_name') || 'default' );
   $name =~ /^([a-z0-9_]{2,80})$/ or die "bad config name [$name]";
   $name = $1;

   if ( $cgi->param('save') )
   {
      save_config($name, $cgi);
   }
   else
   {
      show_config($name, $cgi);
   }
}

sub send_log
{
   my ($log, $is_download) = @_;

   my $file = LOGFILE_ROOT . "/$log" . LOGFILE_EXT;
   my $fh = IO::File->new("<$file");
   defined $fh or die "open logfile [$log]: $!";
   my $size = -s $file;

   my $buf;
   $fh->read($buf, $size) or die "read logfile [$log]: $!";
   $fh->close;

   if ($is_download)
   {
      print header('-type' => 'application/octet-stream',
                   '-Content_Disposition' => qq{attachment; filename="$log.log"},
                   '-Content_Length' => $size,
                  );
   }
   else
   {
      print header('-type' => 'text/plain');
   }

   print $buf;
}

sub save_config
{
   my ($name, $cgi) = @_;

   my $config = $cgi->param('config') || '';
   $config =~ s#(\015\012|\012|\015)#\n#g;
   $config = strip_nonprint($config);
   $config =~ s#\s*$##;
   $config .= "\n\n";
   
   foreach my $template (grep /_template$/, $cgi->param)
   {
      my $tpl = $cgi->param($template) || '';
      $tpl =~ s#(\015\012|\012|\015)#\n#g;
      $tpl = strip_nonprint($tpl);
      $tpl =~ m#\n$# or $tpl .= "\n";
   
      if ($tpl =~ /\S/)
      {
         $tpl =~ s#^#%#mg;
         $config .= "$template:\n$tpl\n\n";
      }
   } 

   my $filename = CONFIG_ROOT."/$name".CONFIG_EXT;
   open OUT, ">$filename.tmp" or die "open >$filename.tmp: $!";
   print OUT $config or die "write to $filename.tmp: $!";
   close OUT or die "close $filename.tmp: $!";
   rename "$filename.tmp", $filename or die "rename [$filename.tmp] -> [$filename]: $!";

   page_header('');
   print <<END;
  <p>The configuration <tt>$name</tt> was saved successfully.</p>
END

   if ($name eq 'default')
   {
      print <<END;
  <p>
   This is the default configuration, so TFmail will use it for any
   HTML form that doesn't include a <tt>_config</tt> hidden input.
  </p>
END
   }
   else
   {
      print <<END;
  <p>
   To use the configuration you have just saved, ensure that your HTML
   form includes the following:</p>
<pre>&lt;input type=&quot;hidden&quot; name=&quot;_config&quot; value=&quot;$name&quot; /&gt;</pre>
END
   }

   print <<END;
   </form>
  <hr />
 </body>
</html>
END
}
   
sub show_config
{
   my ($name, $cgi) = @_;

   my $filename = CONFIG_ROOT."/$name".CONFIG_EXT;
   my $config;
   if (open INFILE, "<$filename")
   {
      local $/;
      $config = <INFILE>;
      close INFILE;
   }
   else
   {
      $config = default_config();
   }
   $config .= "\n\n\n";

   my %templates = ();
   while ($config =~ s#\n\s*([\w\-]+)_template:\n(.*?)\n(?!\%)#\n\n#s)
   {
      my ($name, $tpl) = ($1, $2);
      exists $templates{$name} or $templates{$name} = '';
      $tpl =~ s#^%##mg;
      $templates{$name} .= $tpl;
   }
   $config =~ s#\s*$#\n\n#;

   page_header($name);
   print <<END;
   <p>
    See the <tt>README</tt> file that came with the TFmail distribution
    for details of the things that can be configured here.
   </p>
END

   my $linecount = $config =~ tr#\n## + 5;
   print <<END;
   <textarea name="config" cols="80" rows="$linecount">${\( $cs->escape($config) )}</textarea>
   <hr />
END

   foreach my $template (sort keys %templates)
   {
      my $linecount = $templates{$template} =~ tr#\n## + 5;
      my $value = $cs->escape( $templates{$template}||'' );

      print <<END;
   <p>The text below is the <b>$template template</b>, see <tt>README</tt>
   for details of what you can change here.</p>
   <textarea name="${template}_template" cols="80" rows="$linecount">$value</textarea>
   <hr />
END
   }
   
   print <<END;
   <p>
    <input type="submit" name="save" value="Save this configuration" />
    as <input type="text" name="config_name" value="$name" />
   </p>
   <hr />
  </form>
 </body>
</html>
END
}

sub default_config
{
   return <<'END';
%% NMS configuration file %%
#
# TFmail configuration.
#

#
# recipient: the email address(s) to which TFmail should
# send the results of the form submission.
#
recipient: your.email@address.goes.here

#
# The names of the CGI inputs that TFmail should use to
# build the From: header of the email.
#
email_input:    email
realname_input: realname

success_page_template:
%<?xml version="1.0" encoding="iso-8859-1"?>
%<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
%    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
%<html xmlns="http://www.w3.org/1999/xhtml">
%  <head>
%    <title>Thank You</title>
%    <link rel="stylesheet" type="text/css" href="/css/nms.css" />
%    <style>
%       h1.title {
%                   text-align : center;
%                }
%    </style>
%  </head>
%  <body>
%    <h1 class="title">Thank You</h1>
%    <p>Below is what you submitted on {= date =}</p>
%    <hr size="1" width="75%" />
%{= FOREACH input_field =}
%    <p><b>{= name =}:</b> {= value =}</p>
%{= END =}
%    <hr size="1" width="75%" />
%    <p align="center">
%      <font size="-1">
%        <a href="http://nms-cgi.sourceforge.net/">TFmail</a>
%        &copy; 2002 London Perl Mongers
%      </font>
%    </p>
%  </body>
%</html>

email_template:
%Below is the result of your feedback form.  It was submitted
%{= by_submitter =}on {= date =}.
%----------------------------------------------------------------------
%
%{= FOREACH input_field =}
%{= name =}: {= value =}
%
%{= END =}
%----------------------------------------------------------------------

missing_template:
%<?xml version="1.0" encoding="iso-8859-1"?>
%<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
%    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
%<html xmlns="http://www.w3.org/1999/xhtml">
%  <head>
%    <title>Missing Fields</title>
%    <link rel="stylesheet" type="text/css" href="/css/nms.css" />
%    <style>
%       h1.title {
%                   text-align : center;
%                }
%    </style>
%  </head>
%  <body>
%    <h1 class="title">Missing Fields</h1>
%    <p>
%      The following fields were left blank in your submission form:
%    </p>
%    <ul>
%{= FOREACH missing_field =}
%      <li>{= name =}</li>
%{= END =}
%    </ul>
%    <p>
%      These fields must be filled in before you can successfully
%      submit the form.
%    </p>
%    <p>
%      Please use your back button to return to the form and
%      try again.
%    </p>
%    <p align="center">
%      <font size="-1">
%        <a href="http://nms-cgi.sourceforge.net/">TFmail</a>
%        &copy; 2002 London Perl Mongers
%      </font>
%    </p>
%  </body>
%</html>

confirmation_template:
%Thankyou for your form submission

END
}

sub page_header
{
   my ($this_config)  = @_;

   opendir D, CONFIG_ROOT or die "opendir @{[ CONFIG_ROOT ]}: $!";
   my @config = sort
                grep {!/^default$/}
                grep {/^[\w\-]+$/}
		map  {s/\Q${\( CONFIG_EXT )}\E$// ? ($_) : ()} 
		readdir D;
   closedir D;

   my $title = 'TFmail Configuration Editor';
   $title .= " - $this_config" if length $this_config;

   html_header();
   print <<END;
<?xml version="1.0" encoding="@{[ CHARSET ]}"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>$title</title>
 </head>
 <body>
  <h2>$title</h2>
  <form method="post">
   <input type="hidden" name="password" value="@{[ $cs->escape(PASSWORD) ]}" />
END

   $done_headers = 1;

   show_log_files();

   print "<b>Config:</b>\n";
   foreach my $config ('default', @config)
   {
      if ($config eq $this_config)
      {
         print "<b>[$config]</b>\n";
      }
      else
      {
         print qq{<input type="submit" name="config_name" value="$config" />\n};
      }
   }

   print <<END;
  </form>
  <hr />
  <form method="post">
   <input type="hidden" name="password" value="@{[ $cs->escape(PASSWORD) ]}" />
   Create a new config called
   <input type="text" name="config_name" value="" />
   <input type="submit" value="Create" />
  </form>
  <hr />
  <form method="post">
   <input type="hidden" name="password" value="@{[ $cs->escape(PASSWORD) ]}" />
END
}

sub show_log_files
{
   opendir D, LOGFILE_ROOT or return;
   my @logs = map { /([\w\-]+)\Q@{[ LOGFILE_EXT ]}\E$/ ? ($1) : () } readdir D;
   closedir D;

   if (scalar @logs)
   {
      print "<p>View log file:\n";
      foreach my $log (@logs)
      {
         print qq{<input type="submit" name="viewlog" value="$log" />\n};
      }
      print "</p>\nDownload log file:\n";
      foreach my $log (@logs)
      {
         print qq{<input type="submit" name="getlog" value="$log" />\n};
      }
      print "</p>\n<hr />\n";
   }
}

sub error_page
{
   my ($message) = @_;

   unless ( $done_headers )
   {
      html_header();
      print <<EOERR;
<?xml version="1.0" encoding="@{[ CHARSET ]}"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>Error</title>
  </head>
  <body>
EOERR

      $done_headers = 1;
   }

   if ( DEBUGGING )
   {
      $message = NMSCharset->new(CHARSET)->escape($message);
      $message = "<p>$message</p>";
   }
   else
   {
      $message = '';
   }

   print <<EOERR;
    <h1>Application Error</h1>
    <p>
     An error has occurred in the program
    </p>
    $message
  </body>
</html>
EOERR
}

sub html_header {
    if ($CGI::VERSION >= 2.57) {
        # This is the correct way to set the charset
        print header('-type'=>'text/html', '-charset'=>CHARSET);
    }
    else {
        # However CGI.pm older than version 2.57 doesn't have the
        # -charset option so we cheat:
        print header('-type' => "text/html; charset=@{[ CHARSET ]}");
    }
}


NMSAI_END_CGI_SCRIPT

    '' => <<'NMSAI_END_CGI_SCRIPT',
#!/usr/bin/perl -wT
use strict;
#
# $Id: TFmail.pl,v 1.32 2004/10/12 08:37:21 gellyfish Exp $
#
# USER CONFIGURATION SECTION
# --------------------------
# Modify these to your own settings, see the README file
# for detailed instructions.

use constant DEBUGGING      => 1;
use constant LIBDIR         => '.';
use constant MAILPROG       => '/usr/lib/sendmail -oi -t';
use constant POSTMASTER     => 'me@my.domain';
use constant CONFIG_ROOT    => '.';
use constant SESSION_DIR    => '.';
use constant MAX_DEPTH      => 0;
use constant CONFIG_EXT     => '.trc';
use constant TEMPLATE_EXT   => '.trt';
use constant ENABLE_UPLOADS => 0;
use constant USE_MIME_LITE  => 1;
use constant LOGFILE_ROOT   => '';
use constant LOGFILE_EXT    => '.log';
use constant HTMLFILE_ROOT  => '';
use constant HTMLFILE_EXT   => '.html';
use constant CHARSET        => 'iso-8859-1';

# USER CONFIGURATION << END >>
# ----------------------------
# (no user serviceable parts beyond here)

=head1 NAME

TFmail.pl - template and config file driven formmail CGI

=head1 DESCRIPTION

This CGI script converts form input to an email message.  It
gets its configuration from a configuration file, and uses a
minimal templating system to generate output HTML and the
email message bodies.

See the F<README> file for instructions.

=cut

use constant MIME_LITE => USE_MIME_LITE || ENABLE_UPLOADS;

use Fcntl ':flock';
use IO::File;
use CGI qw(header);
use lib LIBDIR;
use NMStreq;
use NMSCharset;
BEGIN
{
   if (MIME_LITE)
   {
      # Use installed MIME::Lite if available, falling back to
      # the copy of MIME/Lite.pm distributed with the script.
      eval { local $SIG{__DIE__} ; require MIME::Lite };
      require MIME_Lite if $@;
      import MIME::Lite;
   }

   if ( MAILPROG =~ /^SMTP:/i )
   {
      require IO::Socket;
      import IO::Socket;
   }

   use vars qw($VERSION);
   $VERSION = substr q$Revision: 1.32 $, 10, -1;
}

delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};
$ENV{PATH} =~ /(.*)/ and $ENV{PATH} = $1;

use vars qw($done_headers);
$done_headers = 0;

#
# We want to trap die() calls, output an error page and
# then do another die() so that the script aborts and the
# message gets into the server's error log.  If there is
# already a __DIE__ handler installed then we must
# respect it on our final die() call.
#
eval { local $SIG{__DIE__} ; main() };
if ($@)
{
   my $message = $@;
   error_page($message);
   die($message);
}

sub main
{
   local ($CGI::DISABLE_UPLOADS, $CGI::POST_MAX);

   my $treq = NMStreq->new(
      ConfigRoot    => CONFIG_ROOT,
      MaxDepth      => MAX_DEPTH,
      ConfigExt     => CONFIG_EXT,
      TemplateExt   => TEMPLATE_EXT,
      EnableUploads => ENABLE_UPLOADS,
      CGIPostMax    => 1000000,
      Charset       => CHARSET,
   );

   if ( POSTMASTER eq 'me@my.domain' )
   {
      die "You must configure the POSTMASTER constant in the script\n";
   }

   if ( $treq->config('block_lists',''))
   {
      foreach my $zone (split /[\s,]+/, $treq->config('block_lists',''))
      { 
         if (! rbl_check($ENV{REMOTE_ADDR}, $zone ) )
         {
            my $stat = $treq->config('block_status','403 Forbidden');
            my @extra = (-status => $stat);
            if ( $treq->config('blocked_template',''))
            {
               html_page($treq,  $treq->config('blocked_template',''),@extra);
            }
            else
            {
               print header(@extra);
            }
            exit;
         }
      }
   }

   if ( $ENV{REQUEST_METHOD} eq 'POST' )
   {
      check_session($treq) or die "Bad or missing session information";

      my $recipients = check_recipients($treq);
   
      if ( check_required_fields($treq) )
      {
         setup_input_fields($treq);
         my $confto = send_main_email($treq, $recipients);
         if ( HTMLFILE_ROOT ne '' )
         {
            insert_into_html_files($treq);
         }
         if ( LOGFILE_ROOT ne '' )
         {
            log_to_file($treq);
         }
         send_confirmation_email($treq, $confto);
         if ( $treq->config('no_content',0))
         {
            print $treq->cgi()->header(-status => 204 );
            exit;
         }
         else
         {
            return_html($treq);
         }
      }
      else
      {
          missing_html($treq);
      }
   }
   elsif ( can_handle_get($treq))
   {
      if ( $ENV{REQUEST_METHOD} eq 'GET' )
      {
         handle_get($treq);
      }
      else
      {
         bad_method($treq,'POST or GET');
      }
   }
   else
   {
      bad_method($treq,'POST');
   }
}

=head1 INTERNAL FUNCTIONS

=over 4

=item check_session ( TREQ )

If L<use_session> would return a true value this will determine the appropiate
method of determining the session id (either cookie or form field) and 
retrieve the session id then check for its existence, returning true if the
session exists and false if it doesn't.  The session will be removed if it
exists.  It will always return true if sessions are not in use.

=cut

sub check_session
{
   my ( $treq ) = @_;
   
   my $session_ok = 1;
   if ( use_session($treq) )
   {
      $session_ok = 0;

      my $session_id;
      if ( $treq->config('session_cookie',0) )
      {
         $session_id = $treq->cgi()->cookie('SessionID');
      }
      else
      {
         $session_id = $treq->param($treq->config('session_field','session'));
      }

      if ( $session_id )
      {
         $session_id =~ /([a-fA-F0-9]+)/ or die "Bad Session id";
         $session_id = $1;
         
         my $session_file = "@{[ SESSION_DIR ]}/$session_id";

         if ( -f $session_file )
         {
            $session_ok = 1;
            unlink $session_file or die "Can't delete session [$session_file]";
         }
      }

   }

   return $session_ok;
}

=item create_session ( TREQ )

This creates the new session file in SESSION_DIR and returns the number of
the session created.  It will die if it is unable to create the session file.

=cut

sub create_session
{
   my ( $treq ) = @_;

   my $session_id = session_id();
   my $session_file = "@{[ SESSION_DIR ]}/$session_id";

   open TFILE, ">$session_file" or die "Unable to create session: $!";
   print TFILE $ENV{REMOTE_ADDR};
   close TFILE;

   return $session_id;
}


=item session_id

This returns a hexadecimal number that is suitable to be used as a session ID

=cut

=for developers

Please review the uniqueness of this - I tested with ~ 1.5m calls to this
code and didn't find any duplicates but different OS, levels of concurrency
and other factors may impact this.

=cut

sub session_id
{
   return sprintf("%x%x%x", (time() +  $$) * rand, {} * rand,[] *rand)
}

=item use_session ( TREQ )

This returns a true value if either the configuration items 'session_cookie'
or session_field are set, indicating that for a GET request the appropriate
session should be generated and for a POST the existence of the session 
should be checked before any further actions are taken.

=cut

sub use_session
{
   my ( $treq ) = @_;

   if ( $treq->config('session_cookie','') || $treq->config('session_field',''))
   {
      return 1;
   }
   else
   {
      return 0;
   }

}

=item can_handle_get ( TREQ )

Will return a true value if either of the configuration items 'get_redirect'
or 'get_template' has been set, the default is to return false.

=cut

sub can_handle_get
{
   my ( $treq) = @_;

   if ( $treq->config('get_redirect','') || $treq->config('get_template',''))
   {
      return 1;
   }
   else
   {
      return 0;
   }
}

=item handle_get ( TREQ )

This will take the appropriate action for a GET request depending on the
configuration.  If the 'get_redirect' configuration is set then a redirect
will be requested to the specified URL, otherwise if the 'get_template' item
is specified then it will attempt to use that template to present the outpur
based on the query parameters. There is no default behaviour - the assumption
is that can_handle_get() was checked first.

=cut

sub handle_get
{
   my ( $treq ) = @_;

   my $redir = $treq->config('get_redirect');
   if ( defined $redir )
   {
      print "Location: $redir\n\n";
   }
   else
   {
      my @cookie = ();

      setup_input_fields($treq);

      if (use_session($treq) )
      {
         my $session_id = create_session($treq);

         my $me = $treq->cgi()->script_name();

         if ( $treq->config('session_cookie',0) )
         {
            my $cookie = $treq->cgi()->cookie('-name'  => 'SessionID',
                                              '-value' => $session_id,
                                              '-path'  => $me );
            @cookie = ('-cookie' => $cookie);
         }
         else
         {
            $treq->install_directive('session_id', $session_id);
         }
            
      }

      html_page($treq, $treq->config('get_template'),@cookie);
   }
}

=item  bad_method (TREQ, MESSAGE)

Performs the appropriate action as per the configuration if the request
method is not allowed - the default behaviour is to die with an error
noting the allowed methods supplied in MESSAGE, if the bad_method_status
configuaration is set to a true value then it will return a Response with
a "Request Method Not Allowed" (405) status and exit the program.

=cut

sub bad_method
{
   my ( $treq, $message ) = @_;

   if ( $treq->config('bad_method_status',0))
   {
      my @extra = (-status => "405 Request method not allowed");
      if ( $treq->config('bad_method_template',''))
      {
         html_page($treq, $$treq->config('bad_method_template',''),@extra);
      }
      else
      {
         print header(@extra);
      }
      exit;
   }
   else
   {
      $treq->error("request method must be " . $message);
   }
}

=item check_recipients ( TREQ )

Checks that all configured recipients are reasonable email
addresses, and returns a string suitable for use as the value
of a To header.  Dies if any configured recipient is bad.
Returns the empty string if the 'no_email' configuration
setting is true.  It will check first if there is the 'recipient_input'
configuration defined in which case it will attempt to use the value
for the recipient, otherwise it will use the 'recipient' configuration.

=cut

sub check_recipients
{
   my ($treq) = @_;

   $treq->config('no_email', '') and return '';

   my @recip;

   if ( my $recip_field = $treq->config('recipient_input', '' ) )
   {
      $recip[0] = $treq->param($recip_field);
   }
   else
   {
      @recip = split /[\s,]+/, $treq->config('recipient', '');
   }
   scalar @recip or die 'no recipients specified in the config file';
   foreach my $rec (@recip)
   {
      address_ok($rec) or die
         "malformed recipient [$rec] specified in config file";
   }
   return join ', ', @recip;
}

=item address_ok ( ADDRESS )

Returns true if ADDRESS is a reasonable email address, false
otherwise.  Allows leading and trailing spaces and tabs on the
address.

=cut

sub address_ok
{
   my ($addr) = @_;

   $addr =~ m#^[ \t]*[\w\-\.\*]{1,100}\@[\w\-\.]{1,100}[ \t]*$# ? 1 : 0;
}

=item check_required_fields ( TREQ )

Returns false if any fields configured as "required" have
been left blank, true otherwise.

=cut

sub check_required_fields
{
   my ($treq) = @_;

   my @require = split /\s*,\s*/, $treq->config('required', '');

   my @missing = ();
   foreach my $r (@require)
   {
      if ($r =~ /^_?email$/)
      {
         push @missing, $r unless address_ok($treq->param($r));
      }
      else
      {
         push @missing, $r if $treq->param($r) =~ /^\s*$/;
      }
   }

   if (scalar @missing)
   {
      $treq->install_foreach('missing_field', [map {{name=>$_}} @missing]);
      return 0;
   }
   else
   {
      return 1;
   }
}

=item setup_input_fields ( TREQ )

Installs a FOREACH directive in the TREQ object to
iterate over the names and values of input fields.

Honors the 'sort' and 'print_blank_fields' configuration
settings.

=cut

sub setup_input_fields
{
   my ($treq) = @_;

   my @fields;
   my $sort = $treq->config('sort', '');
   if ($sort =~ s/^\s*order\s*:\s*//i)
   {
      @fields = split /\s*,\s*/, $sort;
   }
   else
   {
      @fields = grep {/^[^_]/} $treq->param_list;
      @fields = grep {!($treq->config('session_field',0) 
                   and ($_ eq $treq->config('session_field',''))) } @fields;
      if ($sort =~ /^alpha/i)
      {
         @fields = sort @fields;
      }
   }

   unless ( $treq->config('print_blank_fields', '') )
   {
      @fields = grep {$treq->param($_) !~ /^\s*$/} @fields;
   }

   $treq->install_foreach('input_field',
     [map { {name => $_, value => $treq->param($_)} } @fields]
   );
}

=item dangerous_recipient ( TREQ )

This will return true if the recipient that the main mail message will be
sent to is not directly under the control, currently this will be the case
if the configuration directive 'recipient_input' is being used.  

=cut

=for developers

It is important to keep this function up to date if allowing input from
anywhere other than the config file as we MUST prohibit any other user
supplied input from being sent in the e-mail if this is the case.

=cut

sub dangerous_recipient
{
   my ($treq) = @_;

   my $ret = defined $treq->config('recipient_input', undef) ? 1 : 0;

   return $ret;
}

=item send_main_email ( TREQ, RECIPIENTS )

Sends the main email to the configured recipient.

Any file uploads configured will be attached to the main email,
with "content/type" forced to "application/octet-stream" so
that mail software will do nothing with the attachments other
than allow them to be saved.

Returns the email address of the user if a valid one was
supplied, the empty string otherwise.

Dies on error.

=cut

sub send_main_email
{
   my ($treq, $recipients) = @_;

   my $email_input  = $treq->config('email_input', '');
   my $realname_input  = $treq->config('realname_input', '');

   my $from = POSTMASTER;
   my $confto = '';
   if ($email_input and address_ok($treq->param($email_input)) )
   {
      $from = $treq->param($email_input);
      $from =~ s#\s+##g;
      $confto = $from;
      if ($realname_input)
      {
         my $realname = join ' ', map {$treq->param($_, '')} split /\s+/, $realname_input;
         $realname =~ tr#a-zA-Z0-9_\-\.\,\'\241-\377# #cs;
         $realname = substr $realname, 0, 100;
         $from = build_from_address($treq,$from,$realname);
      }
      my $by = $treq->config('by_submitter_by', 'by');
      $treq->install_directive('by_submitter', "$by $from ") ;
                                      
   }

   return $confto unless length $recipients;

   my $template = $treq->config('email_template', 'email');

   my $subject = $treq->config('subject', 'WWW Form Submission');
   $subject = $treq->process_template("\%$subject", 'email', undef);
   $subject =~ tr#\r\n\t # #s;
   $subject =~ s#\s*$##;
   $subject = substr $subject, 0, 150;

   my $msg = {
      To       => $recipients,
      From     => $from,
      Subject  => $subject,
   };

   if (!dangerous_recipient($treq) and ENABLE_UPLOADS)
   {
      my $cthash = {};
      $msg->{attach} = [];
      my $cgi = $treq->cgi;
      foreach my $param ($treq->param_list)
      {
         next if $param !~ /^(\w+)$/;
         $param = $1;

         my @goodext = split /\s+/, $treq->config("upload_$param", '');
         next unless scalar @goodext;
         my %goodext = map {lc $_=>$_} @goodext;

         my $filename = $cgi->param($param);
         my $info = $cgi->uploadInfo($filename);
         next unless defined $info;
         my $ct = $info->{'Content-Type'} || $info->{'Content-type'} || '';
         $cthash->{$param} = $ct;

         my $filehandle = $cgi->param($param);
         next unless defined $filehandle;
         my $data;
         { local $/; $data = <$filehandle> }

         my $bestext = $goodext[-1];
         if ( $filename =~ m#\.(\w{1,8})$# and exists $goodext{lc $1} )
         {
            $bestext = $goodext{lc $1};
         }
         elsif ( $ct =~ m#^[\w\-]+/(\w{1,8})$# and exists $goodext{lc $1} )
         {
            $bestext = $goodext{lc $1};
         }

         # Some versions of MIME::Lite can loop forever in some circumstances
         # when fed on tainted data.
         $data =~ /^(.*)$/s or die;
         $data = $1;
         $bestext =~ /^([\w\-]+)$/ or die "suspect file extension [$bestext]";
         $bestext = $1;

         push @{ $msg->{attach} }, {
            Type        => 'application/octet-stream',
            Filename    => "$param.$bestext",
            Data        => $data,
            Disposition => 'attachment',
            Encoding    => 'base64',
         };
      }
      $treq->install_directive('content_type', $cthash);
   }

   my $save;

   if ( dangerous_recipient($treq))
   {
      $save = clean_template($treq);
   }

   $msg->{body} = $treq->process_template($template, 'email', undef);

   if ( dangerous_recipient($treq))
   {
      restore_template($treq, $save);
   }

   send_email($treq,$msg);

   return $confto;
}

=item build_from_address ( TREQ , FROM, REALNAME )

Will build the From: address as used in the mail email depending on the
value of the address_style configuration item - the default is:
$email ($realname), a value of 1 specifies "$realname <$email>".

=cut

sub build_from_address
{
   my ( $treq, $from, $realname ) = @_;

   my $new_from;

   if ( $realname !~ /^\s+$/ )
   {
      if ( $treq->config('address_style',0))
      {
         $new_from = "$realname <$from>";
      }
      else
      {
         $new_from = "$from ($realname)";
      }
   }
   else
   {
      $new_from = $from;
   }

   return $new_from;
}

=item send_confirmation_email ( TREQ, CONFTO )

Sends the confirmation email back to the user if configured
to do so and we have a reasonable email address for the user.

The CONFTO parameter must be the sanity checked user's email
address or the empty string it no valid email address was
given.

Dies on error.

=cut

sub send_confirmation_email
{
   my ($treq, $confto) = @_;

   return unless length $confto;

   my $conftemp = $treq->config('confirmation_template', '');
   return unless length $conftemp;

   my $save = clean_template($treq);
   my $body = $treq->process_template($conftemp, 'email', undef);
   restore_template($treq, $save);

   send_email($treq, {
      To      => $confto,
      From    => $treq->config('confirmation_email_from',POSTMASTER),
      Subject => $treq->config('confirmation_subject', 'Thanks'),
      body    => $body,
   });
}

=item clean_template ( TREQ )

This will remove all of the template directives that would place user
supplied input into the processed template - this includes the param
and foreach directives.  It should be used when output (particularly an
e-mail) is going to a recipient which is derived from user input.
A scalar is returned that can be fed to L<restore_template> in order to
put the template directives back.

=cut

=for developers

This should be kept up to date if any further directives are introduced
ideally we should be tracking the installed directives globally.

=cut

sub clean_template
{
   my ( $treq ) = @_;

   my $save = {
     'param'        => $treq->uninstall_directive('param'),
     'param_values' => $treq->uninstall_directive('param_values'),
     'env'          => $treq->uninstall_directive('env'),
     'by_submitter' => $treq->uninstall_directive('by_submitter'),
   };
   my $save_foreach = $treq->uninstall_foreach('input_field');

   return { save => $save, save_foreach => $save_foreach };
}

=item restore_template (TREQ, HASHREF)

This will restore the template directives removed previously by a
L<clean_template> when supplied with the output of that subroutine.
It is recommended to always restore in case a new output is defined
that is considered "safe" and requires templating.

=cut

sub restore_template
{
   my ( $treq, $restore ) = @_;

   foreach my $k (keys %{$restore->{save}})
   {
     $treq->install_directive($k, $restore->{save}{$k});
   }
   $treq->install_foreach('input_field', $restore->{save_foreach});
}

=item send_email ( TREQ, HASHREF )

Adds abuse tracing headers to an outgoing email stored in a
hashref, and sends it.  Dies on error.

=cut

sub send_email
{
   my ($treq, $msg) = @_;

   my $remote_addr = $ENV{REMOTE_ADDR};
   $remote_addr =~ /^[\d\.]+$/ or die "weird remote_addr [$remote_addr]";

   my $x_remote = "[$remote_addr]";
   my $x_gen_by = "NMS TFmail v$VERSION (NMStreq $NMStreq::VERSION)";

   email_start( POSTMASTER, split /\s*,\s*/, $msg->{To} );


   if (MIME_LITE)
   {
      my $ml = MIME::Lite->new(
         To               => $msg->{To},
         From             => $msg->{From},
         Subject          => $msg->{Subject},
         'X-Http-Client'  => $x_remote,
         'X-Generated-By' => $x_gen_by,
         Type             => 'text/plain; charset=' . CHARSET,
         Data             => $msg->{body},
         Date             => '',
         Encoding         => 'quoted-printable',
      );

      foreach my $a (@{ $msg->{attach} || [] })
      {
         $ml->attach( %$a );
      }

      email_data( $ml->as_string );
   }
   else
   {
      email_data(<<END);
X-Http-Client: $x_remote
X-Generated-By: $x_gen_by
To: $msg->{To}
From: $msg->{From}
Subject: $msg->{Subject}

$msg->{body}
END
   }

   email_end();
}

=item email_start( SENDER, RECIPIENT [,...] )

Starts sending a new outgoing email.  SENDER is the envelope
sender and one or more recipient email addresses must be given.

=cut

use vars qw($smtp);
sub email_start {
  my ($sender, @recipients) = @_;

  if (MAILPROG =~ /^SMTP:([\w\-\.]+(:\d+)?)$/i) {
    my $mailhost = $1;
    $mailhost .= ':25' unless $mailhost =~ /:/;
    $smtp = IO::Socket::INET->new($mailhost);
    defined $smtp or die "SMTP connect to [$mailhost]: $!";

    my $banner = smtp_response();
    $banner =~ /^2/ or die "bad SMTP banner [$banner] from [$mailhost]";

    my $helohost = ($ENV{SERVER_NAME} =~ /^([\w\-\.]+)$/ ? $1 : '.');
    smtp_command("HELO $helohost");
    smtp_command("MAIL FROM:<$sender>");
    foreach my $r (@recipients) {
      smtp_command("RCPT TO:<$r>");
    }
    smtp_command("DATA", '3');
  }
  else {
    my $command = MAILPROG . " -f '$sender'";
    my $result;
    eval { local $SIG{__DIE__};
           $result = open SENDMAIL, "| $command"
         };
    if ($@) {
      die $@ unless $@ =~ /Insecure directory/;
      delete $ENV{PATH};
      $result = open SENDMAIL, "| $command";
    }

    die "Can't open mailprog [$command]\n" unless $result;
  }
}

=item email_data ( DATA )

Called one or more times after a call to email_start(), this sub
sends part of the email data.

=cut

sub email_data {
  my ($data) = @_;

  if (defined $smtp) {
    $data =~ s#\n#\015\012#g;
    $data =~ s#^\.#..#mg;
    $smtp->print($data) or die "write to SMTP server: $!";
  } else {
    print SENDMAIL $data or die "write to sendmail pipe: $!";
  }
}

=item email_end ()

This sub must be called for the end of each email.

=cut

sub email_end {
  if (defined $smtp) {
    smtp_command(".");
    smtp_command("QUIT");
    undef $smtp;
  } else {
    close SENDMAIL or die "close sendmail pipe failed, mailprog=[".MAILPROG."]";
  }
}

=item smtp_command ( COMMAND, [EXPECT] )

Sends a single SMTP command to the remote server, and dies unless
the response starts with the character EXPECT.  EXPECT defaults to
'2'.

=cut

sub smtp_command {
  my ($cmd, $want) = @_;
  defined $want or $want = '2';

  $smtp->print("$cmd\015\012")
      or die "write [$cmd] to SMTP server: $!";

  my $resp = smtp_response();
  unless (substr($resp, 0, 1) eq $want) {
    die "SMTP command [$cmd] gave response [$resp]";
  }
}

=item smtp_response ()

Reads a response from the remote SMTP server, and returns it as
a single string.  The returned string may have multiple lines.

=cut

sub smtp_response {
  my $line = smtp_getline();
  my $resp = $line;
  while ($line =~ /^\d\d\d\-/) {
    $line = smtp_getline();
    $resp .= $line;
  }
  return $resp;
}

=item smtp_getline ()

Reads a single line from the remote SMTP server, and returns it
as a string.

=cut

sub smtp_getline {
  my $line = <$smtp>;
  defined $line or die "read from SMTP server: $!";
  return $line;
}  

=item log_to_file ( TREQ )

Appends to a log file, if configured to do so

=cut

sub log_to_file
{
   my ($treq) = @_;

   my $file = $treq->config('logfile', '');
   return unless $file;
   $file =~ m#^([\/\-\w]{1,100})$# or die "bad logfile name [$file]";
   $file = $1;

   open LOG, ">>@{[ LOGFILE_ROOT ]}/$file@{[ LOGFILE_EXT ]}" or die "open [$file]: $!";
   flock LOG, LOCK_EX or die "flock [$file]: $!";
   seek LOG, 0, 2 or die "seek to end of [$file]: $!";

   $treq->process_template(
      $treq->config('log_template', 'log'),
      'email',
      \*LOG
   );

   close LOG or die "close [$file] after append: $!";
}

=item insert_into_html_files ( TREQ )

Inserts template output into one or more HTML files, if configured
to do so.

=cut

sub insert_into_html_files
{
   my ($treq) = @_;

   my @files = split /[\s ,]+/, $treq->config('modify_html_files', '');
   foreach my $file (@files)
   {
      $file =~ m#^([\/\-\w]{1,100})$# or die "bad htmlfile name [$file]";
      $file = $1;
      my $path = "@{[ HTMLFILE_ROOT ]}/$file@{[ HTMLFILE_EXT ]}";

      my $template = $treq->config("htmlfile_template_$file", '');
      die "missing [htmlfile_template_$file] config directive" unless $template;

      rewrite_html_file($treq, $path, $template);
   }
}

=item rewrite_html_file ( TREQ, FILENAME, TEMPLATE )

Rewrites the HTML file FILENAME, inserting the result of running
the template TEMPLATE either above or below the HTML comment
that marks the correct location.

If the HTML comment isn't found, then we default to appending the
template output to the file.

=cut

sub rewrite_html_file
{
   my ($treq, $filename, $template) = @_;

   my $done = 0;

   my $lock = IO::File->new(">>$filename.lck") or die
      "open $filename.lck: $!";
   flock $lock, LOCK_EX or die "flock $filename: $!";

   my $temp = IO::File->new(">$filename.tmp") or die
      "open >$filename.tmp: $!";

   my $in = IO::File->new("<$filename") or die
      "open <$filename: $!";

   while( defined(my $line = <$in>) )
   {
      if ($line =~ /^<!-- NMS insert (above|below) -->\s*$/)
      {
         if ($1 eq 'above')
         {
            $treq->process_template($template, 'html', $temp);
            $temp->print($line);
         }
         else
         {
            $temp->print($line);
            $treq->process_template($template, 'html', $temp);
         }
         $done = 1;
      }
      else
      {
         $temp->print($line);
      }
   }
 
   unless ($done)
   {
      $treq->process_template($template, 'html', $temp);
   }

   unless ($temp->close)
   {
      my $close_err = $!;
      unlink "$filename.tmp";
      die "close $filename.tmp: $close_err";
   }

   $in->close;

   rename "$filename.tmp", $filename or die
      "rename $filename.tmp -> $filename: $!";

   $lock->close;
}

=item missing_html ( TREQ )

Generates the output page in the case where some inputs that
were configured as required have been left blank.

=cut

sub missing_html
{
   my ($treq) = @_;

   my $redirect = $treq->config('missing_fields_redirect');
   if ( $redirect )
   {
      print "Location: $redirect\n\n";
   }
   else
   {
      html_page($treq, $treq->config('missing_template','missing'));
   }
}

=item return_html ( TREQ )

Generates the output page in the case where the email has been
successfully sent.

=cut

sub return_html
{
   my ($treq) = @_;

   my $redirect = $treq->config('redirect');
   if ( defined $redirect )
   {
      print "Location: $redirect\n\n";
   }
   else
   {
      html_page($treq, $treq->config('success_page_template','spage'));
   }
}

=item html_page ( TREQ, TEMPLATE, EXTRA )

Outputs an HTML page using the template TEMPLATE.  EXTRA is an array that is
passed directlyn to L<html_header>.

=cut

sub html_page
{
   my ($treq, $template, @extra) = @_;

   html_header(@extra);
   $done_headers = 1;

   $treq->process_template($template, 'html', \*STDOUT);
}

=item error_page ( MESSAGE )

Displays an "S<Application Error>" page, without using a
template since the error may have arisen during template
resolution.

=cut

sub error_page
{
   my ($message) = @_;

   unless ( $done_headers )
   {
      html_header();
      print <<EOERR;
<?xml version="1.0" encoding="@{[ CHARSET ]}"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>Error</title>
  </head>
  <body>
EOERR

      $done_headers = 1;
   }

   if ( DEBUGGING )
   {
      $message = NMSCharset->new(CHARSET)->escape($message);
      $message = "<p>$message</p>";
   }
   else
   {
      $message = '';
   }

   print <<EOERR;
    <h1>Application Error</h1>
    <p>
     An error has occurred in the program
    </p>
    $message
  </body>
</html>
EOERR
}

=item html_header (EXTRA)

Outputs the CGI header using a content-type of text/html. The optional
argument EXTRA comprise an array of key/value pairs that will be passed
directly to header() method of the CGI module.

=cut

sub html_header {
    my @extra = @_;
    if ($CGI::VERSION >= 2.57) {
        # This is the correct way to set the charset
        print header('-type'=>'text/html', '-charset'=>CHARSET, @extra);
    }
    else {
        # However CGI.pm older than version 2.57 doesn't have the
        # -charset option so we cheat:
        print header('-type' => "text/html; charset=@{[ CHARSET ]}", @extra);
    }
}

=item rbl_check (IP, ZONE )

This performs a dns block list lookup of the supplied IP in the specified
zone, returning false if there is an entry listed and true otherwise.
It can block for a long time if the SOA for the supplied zone is busy or
unavailable.  It is only really useful if the DNSBL zone provided is one
that lists open HTTP proxies and know exploited machines that may be used
by spammers or crackers.  

=cut

=for developers

This has only been tested against a local DNSBL which I can put my own
IP in, so it could probably be tested more thoroughly against a real
DNSBL using some known proxies.

=cut

sub rbl_check 
{
    my ( $ip, $zone ) = @_;

    my $rc = 1;
    if ( $ip =~ /(\d+)\.(\d+).(\d+)\.(\d+)/ ) {
        my $query = "$4.$3.$2.$1.$zone.";
        my $res   = gethostbyname($query);
        if ( defined $res ) {
            $rc = 0;
        }
    }

    return $rc;
}

=back

=head1 MAINTAINERS

The NMS project, E<lt>http://nms-cgi.sourceforge.net/E<gt>

To request support or report bugs, please email
E<lt>nms-cgi-support@lists.sourceforge.netE<gt>

=head1 COPYRIGHT

Copyright 2002 -2004 London Perl Mongers, All rights reserved

=head1 LICENSE

This script is free software; you are free to redistribute it
and/or modify it under the same terms as Perl itself.

=cut


NMSAI_END_CGI_SCRIPT

);

%Modules = (

    'MIME/Lite.pm' => <<'NMSAI_END_PM_FILE',
package MIME::Lite;


=head1 NAME

MIME::Lite - low-calorie MIME generator


=head1 SYNOPSIS

    use MIME::Lite;
   
Create a single-part message:

    ### Create a new single-part message, to send a GIF file:
    $msg = MIME::Lite->new(
                 From     =>'me@myhost.com',
                 To       =>'you@yourhost.com',
                 Cc       =>'some@other.com, some@more.com',
                 Subject  =>'Helloooooo, nurse!',
                 Type     =>'image/gif',
                 Encoding =>'base64',
                 Path     =>'hellonurse.gif'
		 );

Create a multipart message (i.e., one with attachments):

    ### Create a new multipart message:
    $msg = MIME::Lite->new( 
                 From    =>'me@myhost.com',
                 To      =>'you@yourhost.com',
                 Cc      =>'some@other.com, some@more.com',
                 Subject =>'A message with 2 parts...',
                 Type    =>'multipart/mixed'
		 );
    
    ### Add parts (each "attach" has same arguments as "new"):
    $msg->attach(Type     =>'TEXT',   
                 Data     =>"Here's the GIF file you wanted"
		 );  
    $msg->attach(Type     =>'image/gif',
                 Path     =>'aaa000123.gif',
                 Filename =>'logo.gif',
		 Disposition => 'attachment'
		 );

Output a message:

    ### Format as a string:
    $str = $msg->as_string;
    
    ### Print to a filehandle (say, a "sendmail" stream):
    $msg->print(\*SENDMAIL);


Send a message:

    ### Send in the "best" way (the default is to use "sendmail"):
    $msg->send;
      


=head1 DESCRIPTION

In the never-ending quest for great taste with fewer calories,
we proudly present: I<MIME::Lite>.  

MIME::Lite is intended as a simple, standalone module for generating
(not parsing!) MIME messages... specifically, it allows you to
output a simple, decent single- or multi-part message with text or binary
attachments.  It does not require that you have the Mail:: or MIME::
modules installed.

You can specify each message part as either the literal data itself (in
a scalar or array), or as a string which can be given to open() to get
a readable filehandle (e.g., "<filename" or "somecommand|").

You don't need to worry about encoding your message data:
this module will do that for you.  It handles the 5 standard MIME encodings.

If you need more sophisticated behavior, please get the MIME-tools 
package instead.  I will be more likely to add stuff to that toolkit
over this one.


=head1 EXAMPLES

=head2 Create a simple message containing just text

    $msg = MIME::Lite->new(
                 From     =>'me@myhost.com',
                 To       =>'you@yourhost.com',
                 Cc       =>'some@other.com, some@more.com',
                 Subject  =>'Helloooooo, nurse!',
                 Data     =>"How's it goin', eh?"
		 );

=head2 Create a simple message containing just an image

    $msg = MIME::Lite->new(
                 From     =>'me@myhost.com',
                 To       =>'you@yourhost.com',
                 Cc       =>'some@other.com, some@more.com',
                 Subject  =>'Helloooooo, nurse!',
                 Type     =>'image/gif',
                 Encoding =>'base64',
                 Path     =>'hellonurse.gif'
		 );


=head2 Create a multipart message

    ### Create the multipart "container":
    $msg = MIME::Lite->new( 
                 From    =>'me@myhost.com',
                 To      =>'you@yourhost.com',
                 Cc      =>'some@other.com, some@more.com',
                 Subject =>'A message with 2 parts...',
                 Type    =>'multipart/mixed'
		 );
    
    ### Add the text message part:
    ### (Note that "attach" has same arguments as "new"):
    $msg->attach(Type     =>'TEXT',   
                 Data     =>"Here's the GIF file you wanted"
		 );  
     
    ### Add the image part:
    $msg->attach(Type     =>'image/gif',
                 Path     =>'aaa000123.gif',
                 Filename =>'logo.gif',
		 Disposition => 'attachment'
		 );


=head2 Attach a GIF to a text message

This will create a multipart message exactly as above, but using the 
"attach to singlepart" hack:

    ### Start with a simple text message:
    $msg = MIME::Lite->new(
                 From    =>'me@myhost.com',
                 To      =>'you@yourhost.com',
                 Cc      =>'some@other.com, some@more.com',
                 Subject =>'A message with 2 parts...',
                 Type    =>'TEXT',
                 Data    =>"Here's the GIF file you wanted"
                 );  
    
    ### Attach a part... the make the message a multipart automatically:
    $msg->attach(Type     =>'image/gif',
                 Path     =>'aaa000123.gif',
                 Filename =>'logo.gif'
                 );


=head2 Attach a pre-prepared part to a message

    ### Create a standalone part:
    $part = MIME::Lite->new(
                 Type     =>'text/html',
                 Data     =>'<H1>Hello</H1>',
                 );
    $part->attr('content-type.charset' => 'UTF8');
    $part->add('X-Comment' => 'A message for you');
     
    ### Attach it to any message:
    $msg->attach($part);


=head2 Print a message to a filehandle

    ### Write it to a filehandle:
    $msg->print(\*STDOUT); 
     
    ### Write just the header:
    $msg->print_header(\*STDOUT); 
     
    ### Write just the encoded body:
    $msg->print_body(\*STDOUT); 


=head2 Print a message into a string

    ### Get entire message as a string:
    $str = $msg->as_string;
     
    ### Get just the header:
    $str = $msg->header_as_string;
     
    ### Get just the encoded body:
    $str = $msg->body_as_string;


=head2 Send a message

    ### Send in the "best" way (the default is to use "sendmail"):
    $msg->send;


=head2 Send an HTML document... with images included!

    $msg = MIME::Lite->new(
                 To      =>'you@yourhost.com',
                 Subject =>'HTML with in-line images!',
                 Type    =>'multipart/related'
                 );
    $msg->attach(Type => 'text/html',
                 Data => qq{ <body>
                             Here's <i>my</i> image: 
                             <img src="cid:myimage.gif"> 
                             </body> }
                 );
    $msg->attach(Type => 'image/gif',
                 Id   => 'myimage.gif',
                 Path => '/path/to/somefile.gif',
                 );
    $msg->send();


=head2 Change how messages are sent

    ### Do something like this in your 'main':
    if ($I_DONT_HAVE_SENDMAIL) {
       MIME::Lite->send('smtp', "smtp.myisp.net", Timeout=>60);
    }
     
    ### Now this will do the right thing:
    $msg->send;         ### will now use Net::SMTP as shown above






=head1 PUBLIC INTERFACE

=head2 Global configuration

To alter the way the entire module behaves, you have the following
methods/options:

=over 4


=item MIME::Lite->field_order()

When used as a L<classmethod|/field_order>, this changes the default 
order in which headers are output for I<all> messages.
However, please consider using the instance method variant instead,
so you won't stomp on other message senders in the same application.


=item MIME::Lite->quiet()

This L<classmethod|/quiet> can be used to suppress/unsuppress 
all warnings coming from this module.


=item MIME::Lite->send()

When used as a L<classmethod|/send>, this can be used to specify 
a different default mechanism for sending message.  
The initial default is:

    MIME::Lite->send("sendmail", "/usr/lib/sendmail -t -oi -oem");

However, you should consider the similar but smarter and taint-safe variant:

    MIME::Lite->send("sendmail");

Or, for non-Unix users:

    MIME::Lite->send("smtp");


=item $MIME::Lite::AUTO_CC

If true, automatically send to the Cc/Bcc addresses for send_by_smtp().
Default is B<true>.


=item $MIME::Lite::AUTO_CONTENT_TYPE

If true, try to automatically choose the content type from the file name
in C<new()>/C<build()>.  In other words, setting this true changes the
default C<Type> from C<"TEXT"> to C<"AUTO">.

Default is B<false>, since we must maintain backwards-compatibility 
with prior behavior.  B<Please> consider keeping it false,
and just using Type 'AUTO' when you build() or attach().


=item $MIME::Lite::AUTO_ENCODE

If true, automatically choose the encoding from the content type.
Default is B<true>.


=item $MIME::Lite::AUTO_VERIFY

If true, check paths to attachments right before printing, raising an exception
if any path is unreadable.
Default is B<true>.


=item $MIME::Lite::PARANOID

If true, we won't attempt to use MIME::Base64, MIME::QuotedPrint,
or MIME::Types, even if they're available.
Default is B<false>.  Please consider keeping it false,
and trusting these other packages to do the right thing.


=back

=cut

require 5.004;     ### for /c modifier in m/\G.../gc modifier

use Carp ();
use FileHandle;

use strict;
use vars qw(
            $AUTO_CC 
            $AUTO_CONTENT_TYPE
            $AUTO_ENCODE
            $AUTO_VERIFY            
            $PARANOID 
            $QUIET
            $VANILLA
            $VERSION 
            );



#==============================
#==============================
#
# GLOBALS, EXTERNAL/CONFIGURATION...

### The package version, both in 1.23 style *and* usable by MakeMaker:
$VERSION = substr q$Revision: 2.117 $, 10;

### Automatically interpret CC/BCC for SMTP:
$AUTO_CC = 1;

### Automatically choose content type from file name:
$AUTO_CONTENT_TYPE = 0;

### Automatically choose encoding from content type:
$AUTO_ENCODE = 1;

### Check paths right before printing:
$AUTO_VERIFY = 1;

### Set this true if you don't want to use MIME::Base64/QuotedPrint/Types:
$PARANOID = 0;

### Don't warn me about dangerous activities:
$QUIET = undef;

### Unsupported (for tester use): don't qualify boundary with time/pid:
$VANILLA = 0;


#==============================
#==============================
#
# GLOBALS, INTERNAL...

### Find sendmail:
my $SENDMAIL =                 "/usr/lib/sendmail";
(-x $SENDMAIL) or ($SENDMAIL = "/usr/sbin/sendmail");
(-x $SENDMAIL) or ($SENDMAIL = "sendmail");

### Our sending facilities:
my $Sender     = "sendmail";
my %SenderArgs = (
    "sendmail" => ["$SENDMAIL -t -oi -oem"],
    "smtp"     => [],
    "sub"      => [],
);

### Boundary counter:
my $BCount = 0;

### Known Mail/MIME fields... these, plus some general forms like 
### "x-*", are recognized by build():
my %KnownField = map {$_=>1} 
qw(
   bcc         cc          comments      date          encrypted 
   from        keywords    message-id    mime-version  organization
   received    references  reply-to      return-path   sender        
   subject     to

   approved
   );

### What external packages do we use for encoding?
my @Uses;

### Header order:
my @FieldOrder;

### See if we have File::Basename
my $HaveFileBasename = 0;
if (eval "require File::Basename") { # not affected by $PARANOID, core Perl
  $HaveFileBasename = 1;
  push @Uses, "F$File::Basename::VERSION";
}

### See if we have/want MIME::Types
my $HaveMimeTypes = 0;
if (!$PARANOID and eval "require MIME::Types") {
  $HaveMimeTypes = 1;
  push @Uses, "T$MIME::Types::VERSION";
}

#==============================
#==============================
#
# PRIVATE UTILITY FUNCTIONS...

#------------------------------ 
#
# fold STRING
#
# Make STRING safe as a field value.  Remove leading/trailing whitespace,
# and make sure newlines are represented as newline+space

sub fold {
    my $str = shift;
    $str =~ s/^\s*|\s*$//g;    ### trim
    $str =~ s/\n/\n /g;      
    $str;
}

#------------------------------
#
# gen_boundary
#
# Generate a new boundary to use.
# The unsupported $VANILLA is for test purposes only.

sub gen_boundary {
    return ("_----------=_".($VANILLA ? '' : int(time).$$).$BCount++);
}

#------------------------------
#
# known_field FIELDNAME
#
# Is this a recognized Mail/MIME field?

sub known_field {
    my $field = lc(shift);
    $KnownField{$field} or ($field =~ m{^(content|resent|x)-.});
}

#------------------------------
#
# is_mime_field FIELDNAME
#
# Is this a field I manage?

sub is_mime_field {
    $_[0] =~ /^(mime\-|content\-)/i;
}

#------------------------------
#
# extract_addrs STRING
#
# Split STRING into an array of email addresses: somewhat of a KLUDGE.
#
# Unless paranoid, we try to load the real code before supplying our own.

my $ATOM      = '[^ \000-\037()<>@,;:\134"\056\133\135]+';
my $QSTR      = '".*?"';
my $WORD      = '(?:' . $QSTR . '|' . $ATOM . ')';
my $DOMAIN    = '(?:' . $ATOM . '(?:' . '\\.' . $ATOM . ')*' . ')';
my $LOCALPART = '(?:' . $WORD . '(?:' . '\\.' . $WORD . ')*' . ')';
my $ADDR      = '(?:' . $LOCALPART . '@' . $DOMAIN . ')';
my $PHRASE    = '(?:' . $WORD . ')+';
my $SEP       = "(?:^\\s*|\\s*,\\s*)";     ### before elems in a list

sub my_extract_addrs {
    my $str = shift;
    my @addrs;
    $str =~ s/\s/ /g;     ### collapse whitespace

    pos($str) = 0;
    while ($str !~ m{\G\s*\Z}gco) {
	### print STDERR "TACKLING: ".substr($str, pos($str))."\n";
	if    ($str =~ m{\G$SEP$PHRASE\s*<\s*($ADDR)\s*>}gco) {push @addrs,$1}
	elsif ($str =~ m{\G$SEP($ADDR)}gco)                   {push @addrs,$1}
	elsif ($str =~ m{\G$SEP($ATOM)}gco)                   {push @addrs,$1}
	else { 
	    my $problem = substr($str, pos($str));
	    die "can't extract address at <$problem> in <$str>\n";
	}
    }
    return @addrs;
}

if (eval "require Mail::Address") {
    push @Uses, "A$Mail::Address::VERSION";
    eval q{
	sub extract_addrs {
	    return map { $_->format } Mail::Address->parse($_[0]);
	}
    }; ### q
}
else {
    eval q{
        sub extract_addrs {
	    return my_extract_addrs(@_);
	}
    }; ### q
} ### if



#==============================
#==============================
#
# PRIVATE ENCODING FUNCTIONS...

#------------------------------
#
# encode_base64 STRING
#
# Encode the given string using BASE64.
# Unless paranoid, we try to load the real code before supplying our own.

if (!$PARANOID and eval "require MIME::Base64") {
    import MIME::Base64 qw(encode_base64);
    push @Uses, "B$MIME::Base64::VERSION";
}
else {
    eval q{
sub encode_base64 {
    my $res = "";
    my $eol = "\n";

    pos($_[0]) = 0;        ### thanks, Andreas!
    while ($_[0] =~ /(.{1,45})/gs) {
	$res .= substr(pack('u', $1), 1);
	chop($res);
    }
    $res =~ tr|` -_|AA-Za-z0-9+/|;

    ### Fix padding at the end:
    my $padding = (3 - length($_[0]) % 3) % 3;
    $res =~ s/.{$padding}$/'=' x $padding/e if $padding;

    ### Break encoded string into lines of no more than 76 characters each:
    $res =~ s/(.{1,76})/$1$eol/g if (length $eol);
    return $res;
} ### sub
  } ### q
} ### if

#------------------------------
#
# encode_qp STRING
#
# Encode the given string, LINE BY LINE, using QUOTED-PRINTABLE.
# Stolen from MIME::QuotedPrint by Gisle Aas, with a slight bug fix: we
# break lines earlier.  Notice that this seems not to work unless
# encoding line by line.
#
# Unless paranoid, we try to load the real code before supplying our own.

if (!$PARANOID and eval "require MIME::QuotedPrint") {
    import MIME::QuotedPrint qw(encode_qp);
    push @Uses, "Q$MIME::QuotedPrint::VERSION";
}
else {
    eval q{
sub encode_qp {
    my $res = shift;
    local($_);
    $res =~ s/([^ \t\n!-<>-~])/sprintf("=%02X", ord($1))/eg;  ### rule #2,#3
    $res =~ s/([ \t]+)$/
      join('', map { sprintf("=%02X", ord($_)) }
	           split('', $1)
      )/egm;                        ### rule #3 (encode whitespace at eol)

    ### rule #5 (lines shorter than 76 chars, but can't break =XX escapes:
    my $brokenlines = "";
    $brokenlines .= "$1=\n" while $res =~ s/^(.{70}([^=]{2})?)//; ### 70 was 74
    $brokenlines =~ s/=\n$// unless length $res; 
    "$brokenlines$res";
} ### sub
  } ### q
} ### if


#------------------------------
#
# encode_8bit STRING
#
# Encode the given string using 8BIT.
# This breaks long lines into shorter ones.

sub encode_8bit {
    my $str = shift;
    $str =~ s/^(.{990})/$1\n/mg;
    $str;
}

#------------------------------
#
# encode_7bit STRING
#
# Encode the given string using 7BIT.
# This NO LONGER protects people through encoding.

sub encode_7bit {
    my $str = shift;
    $str =~ s/[\x80-\xFF]//g; 
    $str =~ s/^(.{990})/$1\n/mg;
    $str;
}

#==============================
#==============================

=head2 Construction

=over 4

=cut


#------------------------------

=item new [PARAMHASH]

I<Class method, constructor.>
Create a new message object.  

If any arguments are given, they are passed into C<build()>; otherwise,
just the empty object is created.

=cut

sub new {
    my $class = shift;

    ### Create basic object:
    my $self = {
	Attrs   => {},     ### MIME attributes
	Header  => [],     ### explicit message headers
	Parts   => [],     ### array of parts
    };    
    bless $self, $class;

    ### Build, if needed:
    return (@_ ? $self->build(@_) : $self);
}


#------------------------------

=item attach PART

=item attach PARAMHASH...

I<Instance method.>
Add a new part to this message, and return the new part.

If you supply a single PART argument, it will be regarded
as a MIME::Lite object to be attached.  Otherwise, this
method assumes that you are giving in the pairs of a PARAMHASH
which will be sent into C<new()> to create the new part.

One of the possibly-quite-useful hacks thrown into this is the 
"attach-to-singlepart" hack: if you attempt to attach a part (let's
call it "part 1") to a message that doesn't have a content-type
of "multipart" or "message", the following happens:

=over 4

=item *

A new part (call it "part 0") is made.

=item *

The MIME attributes and data (but I<not> the other headers)
are cut from the "self" message, and pasted into "part 0".

=item *

The "self" is turned into a "multipart/mixed" message.

=item *

The new "part 0" is added to the "self", and I<then> "part 1" is added.

=back

One of the nice side-effects is that you can create a text message
and then add zero or more attachments to it, much in the same way
that a user agent like Netscape allows you to do.

=cut

sub attach {
    my $self = shift;

    ### Create new part, if necessary:
    my $part1 = ((@_ == 1) ? shift : ref($self)->new(Top=>0, @_));

    ### Do the "attach-to-singlepart" hack:
    if ($self->attr('content-type') !~ m{^(multipart|message)/}i) {

	### Create part zero:
	my $part0 = ref($self)->new;

	### Cut MIME stuff from self, and paste into part zero: 
	foreach (qw(Attrs Data Path FH)) {
	    $part0->{$_} = $self->{$_}; delete($self->{$_});
	}
	$part0->top_level(0);    ### clear top-level attributes

	### Make self a top-level multipart:
	$self->{Attrs} ||= {};   ### reset       
	$self->attr('content-type'              => 'multipart/mixed');
	$self->attr('content-type.boundary'     => gen_boundary());
	$self->attr('content-transfer-encoding' => '7bit');
	$self->top_level(1);     ### activate top-level attributes

	### Add part 0:
	push @{$self->{Parts}}, $part0;
    }

    ### Add the new part:
    push @{$self->{Parts}}, $part1;
    $part1;
}

#------------------------------

=item build [PARAMHASH]

I<Class/instance method, initializer.>
Create (or initialize) a MIME message object.  
Normally, you'll use the following keys in PARAMHASH:

   * Data, FH, or Path      (either one of these, or none if multipart)
   * Type                   (e.g., "image/jpeg")
   * From, To, and Subject  (if this is the "top level" of a message)

The PARAMHASH can contain the following keys:

=over 4

=item (fieldname)

Any field you want placed in the message header, taken from the
standard list of header fields (you don't need to worry about case):

    Approved      Encrypted     Received      Sender         
    Bcc           From          References    Subject 
    Cc            Keywords      Reply-To      To 
    Comments      Message-ID    Resent-*      X-*
    Content-*     MIME-Version  Return-Path   
    Date                        Organization

To give experienced users some veto power, these fields will be set 
I<after> the ones I set... so be careful: I<don't set any MIME fields>
(like C<Content-type>) unless you know what you're doing!

To specify a fieldname that's I<not> in the above list, even one that's
identical to an option below, just give it with a trailing C<":">,
like C<"My-field:">.  When in doubt, that I<always> signals a mail 
field (and it sort of looks like one too).

=item Data

I<Alternative to "Path" or "FH".>
The actual message data.  This may be a scalar or a ref to an array of
strings; if the latter, the message consists of a simple concatenation 
of all the strings in the array.

=item Datestamp

I<Optional.>
If given true (or omitted), we force the creation of a C<Date:> field 
stamped with the current date/time if this is a top-level message.  
You may want this if using L<send_by_smtp()|/send_by_smtp>.  
If you don't want this to be done, either provide your own Date
or explicitly set this to false.

=item Disposition

I<Optional.>
The content disposition, C<"inline"> or C<"attachment">.
The default is C<"inline">.

=item Encoding

I<Optional.>
The content transfer encoding that should be used to encode your data:

   Use encoding:     | If your message contains:
   ------------------------------------------------------------
   7bit              | Only 7-bit text, all lines <1000 characters
   8bit              | 8-bit text, all lines <1000 characters
   quoted-printable  | 8-bit text or long lines (more reliable than "8bit")
   base64            | Largely non-textual data: a GIF, a tar file, etc.

The default is taken from the Type; generally it is "binary" (no
encoding) for text/*, message/*, and multipart/*, and "base64" for
everything else.  A value of C<"binary"> is generally I<not> suitable
for sending anything but ASCII text files with lines under 1000
characters, so consider using one of the other values instead.

In the case of "7bit"/"8bit", long lines are automatically chopped to
legal length; in the case of "7bit", all 8-bit characters are
automatically I<removed>.  This may not be what you want, so pick your
encoding well!  For more info, see L<"A MIME PRIMER">.

=item FH

I<Alternative to "Data" or "Path".>
Filehandle containing the data, opened for reading.
See "ReadNow" also.

=item Filename

I<Optional.>
The name of the attachment.  You can use this to supply a 
recommended filename for the end-user who is saving the attachment 
to disk.  You only need this if the filename at the end of the 
"Path" is inadequate, or if you're using "Data" instead of "Path".
You should I<not> put path information in here (e.g., no "/" 
or "\" or ":" characters should be used).

=item Id

I<Optional.>
Same as setting "content-id".

=item Length

I<Optional.>
Set the content length explicitly.  Normally, this header is automatically
computed, but only under certain circumstances (see L<"Limitations">).

=item Path

I<Alternative to "Data" or "FH".>
Path to a file containing the data... actually, it can be any open()able
expression.  If it looks like a path, the last element will automatically 
be treated as the filename. 
See "ReadNow" also.

=item ReadNow

I<Optional, for use with "Path".>
If true, will open the path and slurp the contents into core now.
This is useful if the Path points to a command and you don't want 
to run the command over and over if outputting the message several
times.  B<Fatal exception> raised if the open fails.

=item Top

I<Optional.>
If defined, indicates whether or not this is a "top-level" MIME message.
The parts of a multipart message are I<not> top-level.
Default is true.

=item Type

I<Optional.>
The MIME content type, or one of these special values (case-sensitive):

     "TEXT"   means "text/plain"
     "BINARY" means "application/octet-stream"
     "AUTO"   means attempt to guess from the filename, falling back
              to 'application/octet-stream'.  This is good if you have
              MIME::Types on your system and you have no idea what
              file might be used for the attachment.

The default is C<"TEXT">, but it will be C<"AUTO"> if you set
$AUTO_CONTENT_TYPE to true (sorry, but you have to enable 
it explicitly, since we don't want to break code which depends 
on the old behavior).

=back

A picture being worth 1000 words (which
is of course 2000 bytes, so it's probably more of an "icon" than a "picture",
but I digress...), here are some examples:

    $msg = MIME::Lite->build( 
               From     => 'yelling@inter.com',
               To       => 'stocking@fish.net',
               Subject  => "Hi there!",
               Type     => 'TEXT',
               Encoding => '7bit',
               Data     => "Just a quick note to say hi!");
 
    $msg = MIME::Lite->build(
               From     => 'dorothy@emerald-city.oz',
               To       => 'gesundheit@edu.edu.edu',
               Subject  => "A gif for U"
               Type     => 'image/gif',
               Path     => "/home/httpd/logo.gif");
 
    $msg = MIME::Lite->build( 
               From     => 'laughing@all.of.us',
               To       => 'scarlett@fiddle.dee.de',
               Subject  => "A gzipp'ed tar file",
               Type     => 'x-gzip',
               Path     => "gzip < /usr/inc/somefile.tar |",
               ReadNow  => 1,
               Filename => "somefile.tgz");

To show you what's really going on, that last example could also 
have been written:

    $msg = new MIME::Lite;
    $msg->build(Type     => 'x-gzip',
                Path     => "gzip < /usr/inc/somefile.tar |",
                ReadNow  => 1,
                Filename => "somefile.tgz");    
    $msg->add(From    => "laughing@all.of.us");
    $msg->add(To      => "scarlett@fiddle.dee.de");
    $msg->add(Subject => "A gzipp'ed tar file");  

=cut

sub build {
    my $self = shift;
    my %params = @_;
    my @params = @_;
    my $key;

    ### Miko's note: reorganized to check for exactly one of Data, Path, or FH
    (defined($params{Data})+defined($params{Path})+defined($params{FH}) <= 1)
	or Carp::croak "supply exactly zero or one of (Data|Path|FH).\n";

    ### Create new instance, if necessary:
    ref($self) or $self = $self->new;


    ### CONTENT-TYPE....
    ###

    ### Get content-type or content-type-macro:
    my $type = ($params{Type} || ($AUTO_CONTENT_TYPE ? 'AUTO' : 'TEXT'));

    ### Interpret content-type-macros:
    if    ($type eq 'TEXT')   { $type = 'text/plain'; }
    elsif ($type eq 'BINARY') { $type = 'application/octet-stream' }
    elsif ($type eq 'AUTO')   { $type = $self->suggest_type($params{Path}); }

    ### We now have a content-type; set it:
    $type = lc($type);
    $self->attr('content-type' => $type);
   
    ### Get some basic attributes from the content type:
    my $is_multipart = ($type =~ m{^(multipart)/}i);

    ### Add in the multipart boundary:
    if ($is_multipart) {
	my $boundary = gen_boundary();
	$self->attr('content-type.boundary' => $boundary);
    }


    ### CONTENT-ID...
    ###
    $self->attr('content-id' => $params{Id}) if defined($params{Id});


    ### DATA OR PATH...
    ###    Note that we must do this *after* we get the content type, 
    ###    in case read_now() is invoked, since it needs the binmode().

    ### Get data, as...
    ### ...either literal data:
    if (defined($params{Data})) {
	$self->data($params{Data});
    }
    ### ...or a path to data:
    elsif (defined($params{Path})) {
	$self->path($params{Path});       ### also sets filename
	$self->read_now if $params{ReadNow};
    }
    ### ...or a filehandle to data:
    ### Miko's note: this part works much like the path routine just above,
    elsif (defined($params{FH})) {
	$self->fh($params{FH});
	$self->read_now if $params{ReadNow};  ### implement later
    }
    

    ### FILENAME... (added by Ian Smith <ian@safeway.dircon.co.uk> on 8/4/97)
    ###    Need this to make sure the filename is added.  The Filename
    ###    attribute is ignored, otherwise.
    if (defined($params{Filename})) {
	$self->filename($params{Filename});
    }
  

    ### CONTENT-TRANSFER-ENCODING...
    ###

    ### Get it:
    my $enc = ($params{Encoding} ||  
	       ($AUTO_ENCODE and $self->suggest_encoding($type)) ||
	       'binary');      
    $self->attr('content-transfer-encoding' => lc($enc));
	
    ### Sanity check:
    if ($type =~ m{^(multipart|message)/}) {
	($enc =~ m{^(7bit|8bit|binary)\Z}) or 
	  Carp::croak("illegal MIME: ".
		      "can't have encoding $enc with type $type\n");
    }

    ### CONTENT-DISPOSITION...
    ###    Default is inline for single, none for multis:
    ###
    my $disp = ($params{Disposition} or ($is_multipart ? undef : 'inline'));
    $self->attr('content-disposition' => $disp);

    ### CONTENT-LENGTH...
    ###
    my $length;
    if (exists($params{Length})) {   ### given by caller:
	$self->attr('content-length' => $params{Length});
    }
    else {                           ### compute it ourselves
	$self->get_length;
    }

    ### Init the top-level fields:
    my $is_top = defined($params{Top}) ? $params{Top} : 1;
    $self->top_level($is_top);

    ### Datestamp if desired:
    my $ds_wanted    = $params{Datestamp};
    my $ds_defaulted = ($is_top and !exists($params{Datestamp}));
    if (($ds_wanted or $ds_defaulted) and !exists($params{Date})) {
	my ($u_wdy, $u_mon, $u_mdy, $u_time, $u_y4) = 
	    split /\s+/, gmtime()."";   ### should be non-locale-dependent
	my $date = "$u_wdy, $u_mdy $u_mon $u_y4 $u_time UT";
	$self->add("date", $date);
    }
    
    ### Set message headers:
    my @paramz = @params;
    my $field;
    while (@paramz) {
	my ($tag, $value) = (shift(@paramz), shift(@paramz));

	### Get tag, if a tag:
	if ($tag =~ /^-(.*)/) {      ### old style, backwards-compatibility
	    $field = lc($1);
	}
	elsif ($tag =~ /^(.*):$/) {  ### new style
	    $field = lc($1);
	}
	elsif (known_field($field = lc($tag))) {   ### known field
	    ### no-op
	}
	else {                       ### not a field:
	    next;
	}
	
	### Add it:
	$self->add($field, $value);
    }

    ### Done!
    $self;
}

=back

=cut


#==============================
#==============================

=head2 Setting/getting headers and attributes

=over 4

=cut

#------------------------------
#
# top_level ONOFF
#
# Set/unset the top-level attributes and headers.
# This affects "MIME-Version" and "X-Mailer".

sub top_level {
    my ($self, $onoff) = @_;	
    if ($onoff) {
	$self->attr('MIME-Version' => '1.0');
	my $uses = (@Uses ? ("(" . join("; ", @Uses) . ")") : '');
	$self->replace('X-Mailer' => "MIME::Lite $VERSION $uses")
	    unless $VANILLA;
    }
    else {
	$self->attr('MIME-Version' => undef);
	$self->delete('X-Mailer');
    }
}

#------------------------------

=item add TAG,VALUE

I<Instance method.>
Add field TAG with the given VALUE to the end of the header. 
The TAG will be converted to all-lowercase, and the VALUE 
will be made "safe" (returns will be given a trailing space).

B<Beware:> any MIME fields you "add" will override any MIME
attributes I have when it comes time to output those fields.
Normally, you will use this method to add I<non-MIME> fields:

    $msg->add("Subject" => "Hi there!");

Giving VALUE as an arrayref will cause all those values to be added.
This is only useful for special multiple-valued fields like "Received":

    $msg->add("Received" => ["here", "there", "everywhere"]

Giving VALUE as the empty string adds an invisible placeholder
to the header, which can be used to suppress the output of 
the "Content-*" fields or the special  "MIME-Version" field.
When suppressing fields, you should use replace() instead of add():

    $msg->replace("Content-disposition" => "");

I<Note:> add() is probably going to be more efficient than C<replace()>,
so you're better off using it for most applications if you are 
certain that you don't need to delete() the field first.

I<Note:> the name comes from Mail::Header.

=cut

sub add {
    my $self = shift;
    my $tag = lc(shift);
    my $value = shift;

    ### If a dangerous option, warn them:
    Carp::carp "Explicitly setting a MIME header field ($tag) is dangerous:\n".
	 "use the attr() method instead.\n"
	if (is_mime_field($tag) && !$QUIET);

    ### Get array of clean values:
    my @vals = ((ref($value) and (ref($value) eq 'ARRAY'))
		? @{$value} 
		: ($value.''));
    map { s/\n/\n /g } @vals;

    ### Add them:
    foreach (@vals) {
	push @{$self->{Header}}, [$tag, $_];
    }
}

#------------------------------

=item attr ATTR,[VALUE]

I<Instance method.>
Set MIME attribute ATTR to the string VALUE.  
ATTR is converted to all-lowercase.
This method is normally used to set/get MIME attributes:

    $msg->attr("content-type"         => "text/html");
    $msg->attr("content-type.charset" => "US-ASCII");
    $msg->attr("content-type.name"    => "homepage.html");

This would cause the final output to look something like this:

    Content-type: text/html; charset=US-ASCII; name="homepage.html"

Note that the special empty sub-field tag indicates the anonymous 
first sub-field.

Giving VALUE as undefined will cause the contents of the named
subfield to be deleted.

Supplying no VALUE argument just returns the attribute's value:

    $type = $msg->attr("content-type");        ### returns "text/html"
    $name = $msg->attr("content-type.name");   ### returns "homepage.html"

=cut

sub attr {
    my ($self, $attr, $value) = @_;
    $attr = lc($attr);

    ### Break attribute name up:
    my ($tag, $subtag) = split /\./, $attr;
    defined($subtag) or $subtag = '';

    ### Set or get?
    if (@_ > 2) {   ### set:
	$self->{Attrs}{$tag} ||= {};            ### force hash
	delete $self->{Attrs}{$tag}{$subtag};   ### delete first
	if (defined($value)) {                  ### set...
	    $value =~ s/[\r\n]//g;                   ### make clean
	    $self->{Attrs}{$tag}{$subtag} = $value;
	}
    }
	
    ### Return current value:
    $self->{Attrs}{$tag}{$subtag};
}

sub _safe_attr {
    my ($self, $attr) = @_;
    my $v = $self->attr($attr);
    defined($v) ? $v : '';
}

#------------------------------

=item delete TAG

I<Instance method.>
Delete field TAG with the given VALUE to the end of the header.  
The TAG will be converted to all-lowercase.

    $msg->delete("Subject");

I<Note:> the name comes from Mail::Header.

=cut

sub delete {
    my $self = shift;
    my $tag = lc(shift);

    ### Delete from the header:
    my $hdr = [];
    my $field;
    foreach $field (@{$self->{Header}}) {
	push @$hdr, $field if ($field->[0] ne $tag);
    }
    $self->{Header} = $hdr;
    $self;
}


#------------------------------

=item field_order FIELD,...FIELD

I<Class/instance method.>  
Change the order in which header fields are output for this object:

    $msg->field_order('from', 'to', 'content-type', 'subject');

When used as a class method, changes the default settings for
all objects:

    MIME::Lite->field_order('from', 'to', 'content-type', 'subject');

Case does not matter: all field names will be coerced to lowercase.
In either case, supply the empty array to restore the default ordering.

=cut

sub field_order {
    my $self = shift;
    if (ref($self)) { $self->{FieldOrder} = [ map { lc($_) } @_ ] }
    else            { @FieldOrder         =   map { lc($_) } @_ }
}

#------------------------------

=item fields

I<Instance method.>
Return the full header for the object, as a ref to an array
of C<[TAG, VALUE]> pairs, where each TAG is all-lowercase.  
Note that any fields the user has explicitly set will override the 
corresponding MIME fields that we would otherwise generate.  
So, don't say...

    $msg->set("Content-type" => "text/html; charset=US-ASCII");

unless you want the above value to override the "Content-type"
MIME field that we would normally generate.

I<Note:> I called this "fields" because the header() method of
Mail::Header returns something different, but similar enough to 
be confusing.

You can change the order of the fields: see L</field_order>. 
You really shouldn't need to do this, but some people have to
deal with broken mailers.

=cut

sub fields {
    my $self = shift;
    my @fields;
    
    ### Get a lookup-hash of all *explicitly-given* fields:
    my %explicit = map { $_->[0] => 1 } @{$self->{Header}};
    
    ### Start with any MIME attributes not given explicitly:
    my $tag;
    foreach $tag (sort keys %{$self->{Attrs}}) {	

	### Skip if explicit:
	next if ($explicit{$tag});         

	### Skip if no subtags:
	my @subtags = keys %{$self->{Attrs}{$tag}}; 
	@subtags or next;

	### Create string:
	my $value;
	defined($value = $self->{Attrs}{$tag}{''}) or next;  ### need default 
	foreach (sort @subtags) {
	    next if ($_ eq '');
	    $value .= qq{; $_="$self->{Attrs}{$tag}{$_}"};
	}
	
	### Add to running fields;
	push @fields, [$tag, $value];
    }
    
    ### Add remaining fields (note that we duplicate the array for safety):
    foreach (@{$self->{Header}}) {
	push @fields, [@{$_}];
    }

    ### Final step: 
    ### If a suggested ordering was given, we "sort" by that ordering.
    ###    The idea is that we give each field a numeric rank, which is
    ###    (1000 * order(field)) + origposition.
    my @order = @{$self->{FieldOrder} || []};      ### object-specific
    @order or @order = @FieldOrder;                ### no? maybe generic
    if (@order) {                                  ### either?

	### Create hash mapping field names to 1-based rank:
	my %rank = map {$order[$_] => (1+$_)} (0..$#order);
	
	### Create parallel array to @fields, called @ranked.
	### It contains fields tagged with numbers like 2003, where the 
	### 3 is the original 0-based position, and 2000 indicates that 
	### we wanted ths type of field to go second.
	my @ranked = map {
	    [
	     ($_ + 1000*($rank{lc($fields[$_][0])} || (2+$#order))),
	     $fields[$_]
	     ]
	    } (0..$#fields);  
	# foreach (@ranked) { 
	#     print STDERR "RANKED: $_->[0] $_->[1][0] $_->[1][1]\n";
	# }

	### That was half the Schwartzian transform.  Here's the rest:
	@fields = map { $_->[1] } 
	          sort { $a->[0] <=> $b->[0] } 
	          @ranked;
    }

    ### Done!
    return \@fields;
}


#------------------------------

=item filename [FILENAME]

I<Instance method.>
Set the filename which this data will be reported as.
This actually sets both "standard" attributes.

With no argument, returns the filename as dictated by the 
content-disposition.

=cut

sub filename {
    my ($self, $filename) = @_;
    if (@_ > 1) {
	$self->attr('content-type.name' => $filename);
	$self->attr('content-disposition.filename' => $filename);
    }
    $self->attr('content-disposition.filename');
}

#------------------------------

=item get TAG,[INDEX]

I<Instance method.>
Get the contents of field TAG, which might have been set 
with set() or replace().  Returns the text of the field.

    $ml->get('Subject', 0);

If the optional 0-based INDEX is given, then we return the INDEX'th
occurence of field TAG.  Otherwise, we look at the context:
In a scalar context, only the first (0th) occurence of the 
field is returned; in an array context, I<all> occurences are returned.  

I<Warning:> this should only be used with non-MIME fields.
Behavior with MIME fields is TBD, and will raise an exception for now.

=cut

sub get {
    my ($self, $tag, $index) = @_;
    $tag = lc($tag); 
    Carp::croak "get: can't be used with MIME fields\n" if is_mime_field($tag);
    
    my @all = map { ($_->[0] eq $tag) ? $_->[1] : ()} @{$self->{Header}};
    (defined($index) ? $all[$index] : (wantarray ? @all : $all[0]));
}

#------------------------------

=item get_length

I<Instance method.>
Recompute the content length for the message I<if the process is trivial>, 
setting the "content-length" attribute as a side-effect:

    $msg->get_length;

Returns the length, or undefined if not set.

I<Note:> the content length can be difficult to compute, since it 
involves assembling the entire encoded body and taking the length
of it (which, in the case of multipart messages, means freezing
all the sub-parts, etc.).  

This method only sets the content length to a defined value if the
message is a singlepart with C<"binary"> encoding, I<and> the body is
available either in-core or as a simple file.  Otherwise, the content
length is set to the undefined value.

Since content-length is not a standard MIME field anyway (that's right, kids:
it's not in the MIME RFCs, it's an HTTP thing), this seems pretty fair.

=cut

#----
# Miko's note: I wasn't quite sure how to handle this, so I waited to hear 
# what you think.  Given that the content-length isn't always required, 
# and given the performance cost of calculating it from a file handle,
# I thought it might make more sense to add some some sort of computelength 
# property. If computelength is false, then the length simply isn't 
# computed.  What do you think?
#
# Eryq's reply:  I agree; for now, we can silently leave out the content-type.

sub get_length {
    my $self = shift;

    my $is_multipart = ($self->attr('content-type') =~ m{^multipart/}i);
    my $enc = lc($self->attr('content-transfer-encoding') || 'binary');
    my $length;
    if (!$is_multipart && ($enc eq "binary")){  ### might figure it out cheap:
	if    (defined($self->{Data})) {               ### it's in core
	    $length = length($self->{Data});
	}
	elsif (defined($self->{FH})) {                 ### it's in a filehandle
	    ### no-op: it's expensive, so don't bother
	}
	elsif (defined($self->{Path})) {               ### it's a simple file!
	    $length = (-s $self->{Path})   if (-e $self->{Path});
	}
    }
    $self->attr('content-length' => $length);
    return $length;
}

#------------------------------

=item parts

I<Instance method.>
Return the parts of this entity, and this entity only.
Returns empty array if this entity has no parts.

This is B<not> recursive!  Parts can have sub-parts; use
parts_DFS() to get everything.

=cut

sub parts {
    my $self = shift;
    @{$self->{Parts} || []};
}

#------------------------------

=item parts_DFS

I<Instance method.>
Return the list of all MIME::Lite objects included in the entity,
starting with the entity itself, in depth-first-search order.  
If this object has no parts, it alone will be returned.

=cut

sub parts_DFS {
    my $self = shift;
    return ($self, map { $_->parts_DFS } $self->parts);
}

#------------------------------

=item preamble [TEXT]

I<Instance method.>
Get/set the preamble string, assuming that this object has subparts.
Set it to undef for the default string.

=cut

sub preamble {
    my $self = shift;
    $self->{Preamble} = shift if @_;
    $self->{Preamble};
}

#------------------------------

=item replace TAG,VALUE

I<Instance method.>
Delete all occurences of fields named TAG, and add a new
field with the given VALUE.  TAG is converted to all-lowercase.

B<Beware> the special MIME fields (MIME-version, Content-*):
if you "replace" a MIME field, the replacement text will override 
the I<actual> MIME attributes when it comes time to output that field.
So normally you use attr() to change MIME fields and add()/replace() to 
change I<non-MIME> fields:

    $msg->replace("Subject" => "Hi there!");

Giving VALUE as the I<empty string> will effectively I<prevent> that
field from being output.  This is the correct way to suppress
the special MIME fields:    

    $msg->replace("Content-disposition" => "");

Giving VALUE as I<undefined> will just cause all explicit values
for TAG to be deleted, without having any new values added.

I<Note:> the name of this method  comes from Mail::Header.

=cut

sub replace {
    my ($self, $tag, $value) = @_;
    $self->delete($tag);
    $self->add($tag, $value) if defined($value);
}


#------------------------------

=item scrub

I<Instance method.>
B<This is Alpha code.  If you use it, please let me know how it goes.>
Recursively goes through the "parts" tree of this message and tries 
to find MIME attributes that can be removed. 
With an array argument, removes exactly those attributes; e.g.:

    $msg->scrub(['content-disposition', 'content-length']);

Is the same as recursively doing:

    $msg->replace('Content-disposition' => '');
    $msg->replace('Content-length'      => '');

=cut

sub scrub {
    my ($self, @a) = @_;
    my ($expl) = @a;
    local $QUIET = 1;

    ### Scrub me:
    if (!@a) {         ### guess

	### Scrub length always:
	$self->replace('content-length', '');

	### Scrub disposition if no filename, or if content-type has same info:
	if (!$self->_safe_attr('content-disposition.filename') ||
	    $self->_safe_attr('content-type.name')) {
	    $self->replace('content-disposition', '');
	}

	### Scrub encoding if effectively unencoded:
	if ($self->_safe_attr('content-transfer-encoding') =~
	    /^(7bit|8bit|binary)$/i) {
	    $self->replace('content-transfer-encoding', '');
	}

	### Scrub charset if US-ASCII:
	if ($self->_safe_attr('content-type.charset') =~ /^(us-ascii)/i) {
	    $self->attr('content-type.charset' => undef);
	}

	### TBD: this is not really right for message/digest:
	if ((keys %{$self->{Attrs}{'content-type'}} == 1) and
	    ($self->_safe_attr('content-type') eq 'text/plain')) {
	    $self->replace('content-type', '');
	}
    }
    elsif ($expl and (ref($expl) eq 'ARRAY')) {
	foreach (@{$expl}) { $self->replace($_, ''); }
    }

    ### Scrub my kids:
    foreach (@{$self->{Parts}}) { $_->scrub(@a); }
}

=back

=cut


#==============================
#==============================

=head2 Setting/getting message data

=over 4

=cut

#------------------------------

=item binmode [OVERRIDE]

I<Instance method.>
With no argument, returns whether or not it thinks that the data 
(as given by the "Path" argument of C<build()>) should be read using 
binmode() (for example, when C<read_now()> is invoked).

The default behavior is that any content type other than 
C<text/*> or C<message/*> is binmode'd; this should in general work fine.

With a defined argument, this method sets an explicit "override"
value.  An undefined argument unsets the override.
The new current value is returned.

=cut

sub binmode {
    my $self = shift;
    $self->{Binmode} = shift if (@_);       ### argument? set override
    return (defined($self->{Binmode}) 
	    ? $self->{Binmode}
	    : ($self->attr("content-type") !~ m{^(text|message)/}i));
}

#------------------------------

=item data [DATA]

I<Instance method.>
Get/set the literal DATA of the message.  The DATA may be
either a scalar, or a reference to an array of scalars (which
will simply be joined).    

I<Warning:> setting the data causes the "content-length" attribute
to be recomputed (possibly to nothing).

=cut

sub data {
    my $self = shift;
    if (@_) {
	$self->{Data} = ((ref($_[0]) eq 'ARRAY') ? join('', @{$_[0]}) : $_[0]);
	$self->get_length;
    }
    $self->{Data};
}

#------------------------------

=item fh [FILEHANDLE]

I<Instance method.>
Get/set the FILEHANDLE which contains the message data.

Takes a filehandle as an input and stores it in the object.
This routine is similar to path(); one important difference is that 
no attempt is made to set the content length.  

=cut

sub fh {
    my $self = shift;
    $self->{FH} = shift if @_;
    $self->{FH};
}

#------------------------------

=item path [PATH]

I<Instance method.>
Get/set the PATH to the message data.

I<Warning:> setting the path recomputes any existing "content-length" field,
and re-sets the "filename" (to the last element of the path if it
looks like a simple path, and to nothing if not).

=cut

sub path {
    my $self = shift;
    if (@_) {

	### Set the path, and invalidate the content length:
	$self->{Path} = shift;

	### Re-set filename, extracting it from path if possible:
	my $filename;
	if ($self->{Path} and ($self->{Path} !~ /\|$/)) {  ### non-shell path:
	    ($filename = $self->{Path}) =~ s/^<//;    

	    ### Consult File::Basename, maybe:
	    if ($HaveFileBasename) {
		$filename = File::Basename::basename($filename);
	    } 
	    else {
		($filename) = ($filename =~ m{([^\/]+)\Z});
	    }
	}
	$self->filename($filename);

	### Reset the length:
	$self->get_length;
    }
    $self->{Path};
}

#------------------------------

=item resetfh [FILEHANDLE]

I<Instance method.>
Set the current position of the filehandle back to the beginning. 
Only applies if you used "FH" in build() or attach() for this message.

Returns false if unable to reset the filehandle (since not all filehandles
are seekable).

=cut

#----
# Miko's note: With the Data and Path, the same data could theoretically 
# be reused.  However, file handles need to be reset to be reused, 
# so I added this routine.
#
# Eryq reply: beware... not all filehandles are seekable (think about STDIN)!

sub resetfh {
    my $self = shift;
    seek($self->{FH},0,0);
}

#------------------------------

=item read_now 

I<Instance method.>
Forces data from the path/filehandle (as specified by C<build()>)
to be read into core immediately, just as though you had given it
literally with the C<Data> keyword.  

Note that the in-core data will always be used if available.

Be aware that everything is slurped into a giant scalar: you may not want 
to use this if sending tar files!  The benefit of I<not> reading in the data 
is that very large files can be handled by this module if left on disk
until the message is output via C<print()> or C<print_body()>.

=cut

sub read_now {
    my $self = shift;
    local $/ = undef;
    
    if    ($self->{FH}) {       ### data from a filehandle:
	my $chunk;
	my @chunks;
	CORE::binmode($self->{FH}) if $self->binmode;
	while (read($self->{FH}, $chunk, 1024)) { 
	    push @chunks, $chunk; 
	}
	$self->{Data} = join '', @chunks;
    }
    elsif ($self->{Path}) {     ### data from a path:
	open SLURP, $self->{Path} or Carp::croak "open $self->{Path}: $!\n";
	CORE::binmode(SLURP) if $self->binmode;
	$self->{Data} = <SLURP>;        ### sssssssssssssslurp...
	close SLURP;                    ### ...aaaaaaaaahhh!
    }
}

#------------------------------

=item sign PARAMHASH

I<Instance method.>
Sign the message.  This forces the message to be read into core,
after which the signature is appended to it.

=over 4

=item Data

As in C<build()>: the literal signature data.
Can be either a scalar or a ref to an array of scalars.

=item Path

As in C<build()>: the path to the file.

=back

If no arguments are given, the default is:

    Path => "$ENV{HOME}/.signature"

The content-length is recomputed.

=cut

sub sign {
    my $self = shift;
    my %params = @_;

    ### Default:
    @_ or $params{Path} = "$ENV{HOME}/.signature";

    ### Force message in-core:
    defined($self->{Data}) or $self->read_now;

    ### Load signature:
    my $sig;
    if (!defined($sig = $params{Data})) {      ### not given explicitly:
	local $/ = undef;
	open SIG, $params{Path} or Carp::croak "open sig $params{Path}: $!\n";
	$sig = <SIG>;                  ### sssssssssssssslurp...
	close SIG;                     ### ...aaaaaaaaahhh!
    }    
    $sig = join('',@$sig) if (ref($sig) and (ref($sig) eq 'ARRAY'));

    ### Append, following Internet conventions:
    $self->{Data} .= "\n-- \n$sig";

    ### Re-compute length:
    $self->get_length;
    1;
}

#------------------------------
#
# =item suggest_encoding CONTENTTYPE
#
# I<Class/instance method.>
# Based on the CONTENTTYPE, return a good suggested encoding.
# C<text> and C<message> types have their bodies scanned line-by-line
# for 8-bit characters and long lines; lack of either means that the
# message is 7bit-ok.  Other types are chosen independent of their body:
#
#    Major type:       7bit ok?    Suggested encoding:
#    ------------------------------------------------------------
#    text              yes         7bit
#                      no          quoted-printable    
#                      unknown     binary
#
#    message           yes         7bit
#                      no          binary    
#                      unknown     binary
#
#    multipart         n/a         binary (in case some parts are not ok)
#
#    (other)           n/a         base64
#
#=cut

sub suggest_encoding {
    my ($self, $ctype) = @_;
    $ctype = lc($ctype);

    ### Consult MIME::Types, maybe:
    if ($HaveMimeTypes) {
    
	### Mappings contain [suffix,mimetype,encoding]
	my @mappings = MIME::Types::by_mediatype($ctype);
	if (scalar(@mappings)) {
	    ### Just pick the first one:
	    my ($suffix, $mimetype, $encoding) = @{$mappings[0]};
	    if ($encoding &&
		$encoding =~/^(base64|binary|[78]bit|quoted-printable)$/i) {
		return lc($encoding);    ### sanity check
	    }
	}
    }

    ### If we got here, then MIME::Types was no help.
    ### Extract major type:
    my ($type) = split '/', $ctype;
    if (($type eq 'text') || ($type eq 'message')) {    ### scan message body?
	return 'binary';
    }
    else {
	return ($type eq 'multipart') ? 'binary' : 'base64';
    }
}

#------------------------------
#
# =item suggest_type PATH
#
# I<Class/instance method.>
# Suggest the content-type for this attached path.
# We always fall back to "application/octet-stream" if no good guess
# can be made, so don't use this if you don't mean it!
#
sub suggest_type {
    my ($self, $path) = @_;

    ### If there's no path, bail:
    $path or return 'application/octet-stream';

    ### Consult MIME::Types, maybe:
    if ($HaveMimeTypes) {
	# Mappings contain [mimetype,encoding]:
	my @mappings = MIME::Types::by_suffix($path);
	if (scalar(@mappings)) {
	    ### Just pick the first one:
	    my ($mimetype, $encoding) = @{$mappings[0]};
	    if ($mimetype && $mimetype =~ /^\S+\/\S+$/) {
		return $mimetype;  ### sanity check
	    }	    
	}
    }
    
    ### If we got here, then MIME::Types was no help.
    ### The correct thing to fall back to is the most-generic content type:
    return 'application/octet-stream';
}

#------------------------------

=item verify_data

I<Instance method.>
Verify that all "paths" to attached data exist, recursively.  
It might be a good idea for you to do this before a print(), to
prevent accidental partial output if a file might be missing.
Raises exception if any path is not readable.

=cut

sub verify_data {
    my $self = shift;

    ### Verify self:
    my $path = $self->{Path};
    if ($path and ($path !~ /\|$/)) {  ### non-shell path:
	$path =~ s/^<//;    
	(-r $path) or die "$path: not readable\n";
    }

    ### Verify parts:
    foreach my $part (@{$self->{Parts}}) { $part->verify_data }
    1;
}

=back

=cut


#==============================
#==============================

=head2 Output

=over 4

=cut

#------------------------------

=item print [OUTHANDLE]

I<Instance method.> 
Print the message to the given output handle, or to the currently-selected
filehandle if none was given.

All OUTHANDLE has to be is a filehandle (possibly a glob ref), or 
any object that responds to a print() message.

=cut

sub print {
    my ($self, $out) = @_;

    ### Coerce into a printable output handle:
    $out = wrap MIME::Lite::IO_Handle $out;

    ### Output head, separator, and body:
    $self->verify_data if $AUTO_VERIFY;       ### prevents missing parts!
    $out->print($self->header_as_string, "\n");
    $self->print_body($out);
}

#------------------------------
#
# print_for_smtp
#
# Instance method, private.
# Print, but filter out the topmost "Bcc" field.
# This is because qmail apparently doesn't do this for us!
#
sub print_for_smtp {
    my ($self, $out) = @_;

    ### Coerce into a printable output handle:
    $out = wrap MIME::Lite::IO_Handle $out;
    
    ### Create a safe head:
    my @fields = grep { $_->[0] ne 'bcc' } @{$self->fields};
    my $header = $self->fields_as_string(\@fields);

    ### Output head, separator, and body:
    $out->print($header, "\n");
    $self->print_body($out);
}

#------------------------------

=item print_body [OUTHANDLE]

I<Instance method.> 
Print the body of a message to the given output handle, or to 
the currently-selected filehandle if none was given.  

All OUTHANDLE has to be is a filehandle (possibly a glob ref), or 
any object that responds to a print() message.

B<Fatal exception> raised if unable to open any of the input files,
or if a part contains no data, or if an unsupported encoding is 
encountered.

=cut

sub print_body {
    my ($self, $out) = @_;

    ### Coerce into a printable output handle:
    $out = wrap MIME::Lite::IO_Handle $out;

    ### Output either the body or the parts.
    ###   Notice that we key off of the content-type!  We expect fewer 
    ###   accidents that way, since the syntax will always match the MIME type.
    my $type = $self->attr('content-type');
    if ($type =~ m{^multipart/}i) {	
	my $boundary = $self->attr('content-type.boundary');

	### Preamble:
	$out->print(defined($self->{Preamble})
		    ? $self->{Preamble}
		    : "This is a multi-part message in MIME format.\n");
	
	### Parts:
	my $part;
	foreach $part (@{$self->{Parts}}) {
	    $out->print("\n--$boundary\n");
	    $part->print($out);
	}

	### Epilogue:
	$out->print("\n--$boundary--\n\n");
    }
    elsif ($type =~ m{^message/}) {
	my @parts = @{$self->{Parts}};

	### It's a toss-up; try both data and parts:
	if    (@parts == 0) { $self->print_simple_body($out) }
	elsif (@parts == 1) { $parts[0]->print($out) }
	else                { Carp::croak "can't handle message with >1 part\n"; }
    }
    else {                    
	$self->print_simple_body($out); 
    }
    1;
}

#------------------------------
#
# print_simple_body [OUTHANDLE]
#
# I<Instance method, private.>
# Print the body of a simple singlepart message to the given 
# output handle, or to the currently-selected filehandle if none 
# was given.  
#
# Note that if you want to print "the portion after
# the header", you don't want this method: you want 
# L<print_body()|/print_body>.
#
# All OUTHANDLE has to be is a filehandle (possibly a glob ref), or 
# any object that responds to a print() message.
#
# B<Fatal exception> raised if unable to open any of the input files,
# or if a part contains no data, or if an unsupported encoding is 
# encountered.
#
sub print_simple_body {
    my ($self, $out) = @_;

    ### Coerce into a printable output handle:
    $out = wrap MIME::Lite::IO_Handle $out;

    ### Get content-transfer-encoding:
    my $encoding = uc($self->attr('content-transfer-encoding'));

    ### Notice that we don't just attempt to slurp the data in from a file:
    ### by processing files piecemeal, we still enable ourselves to prepare
    ### very large MIME messages...

    ### Is the data in-core?  If so, blit it out...
    if (defined($self->{Data})) {
      DATA: 
	{ local $_ = $encoding;

	  /^BINARY$/ and do {
	      $out->print($self->{Data}); 
	      last DATA;
	  };
	  /^8BIT$/ and do {
	      $out->print(encode_8bit($self->{Data})); 
	      last DATA;
	  };
	  /^7BIT$/ and do {
	      $out->print(encode_7bit($self->{Data})); 
	      last DATA;
	  };
	  /^QUOTED-PRINTABLE$/ and do {
	      ### UNTAINT since m//mg on tainted data loops forever:
	      my ($untainted) = ($self->{Data} =~ m/\A(.*)\Z/s);

	      ### Encode it line by line:
	      while ($untainted =~ m{^(.*[\r\n]*)}mg) {
		  $out->print(encode_qp($1)); ### have to do it line by line...
	      }
	      last DATA;	 
	  };
	  /^BASE64/ and do {
	      $out->print(encode_base64($self->{Data})); 
	      last DATA;
	  };
	  Carp::croak "unsupported encoding: `$_'\n";
        }
    }

    ### Else, is the data in a file?  If so, output piecemeal...
    ###    Miko's note: this routine pretty much works the same with a path 
    ###    or a filehandle. the only difference in behaviour is that it does 
    ###    not attempt to open anything if it already has a filehandle
    elsif (defined($self->{Path}) || defined($self->{FH})) {
	no strict 'refs';          ### in case FH is not an object
	my $DATA;

	### Open file if necessary:
	if (defined($self->{Path})) {
	    $DATA = new FileHandle || Carp::croak "can't get new filehandle\n";
	    $DATA->open("$self->{Path}") or 
	      Carp::croak "open $self->{Path}: $!\n";
	}
	else {
	    $DATA=$self->{FH};
	}
	CORE::binmode($DATA) if $self->binmode;
		
	### Encode piece by piece:
      PATH: 
	{   local $_ = $encoding;
	    
	    /^BINARY$/ and do {
		$out->print($_)                while read($DATA, $_, 2048); 
		last PATH;
	    };      
	    /^8BIT$/ and do {
		$out->print(encode_8bit($_))   while (<$DATA>); 
		last PATH;
	    };
	    /^7BIT$/ and do {
		$out->print(encode_7bit($_))   while (<$DATA>); 
		last PATH;
	    };
	    /^QUOTED-PRINTABLE$/ and do {
		$out->print(encode_qp($_))     while (<$DATA>); 
		last PATH;
	    };
	    /^BASE64$/ and do {
		$out->print(encode_base64($_)) while (read($DATA, $_, 45));
		last PATH;
	    };
	    Carp::croak "unsupported encoding: `$_'\n";
	}
	
	### Close file:
	close $DATA if defined($self->{Path});
    }
    
    else {
	Carp::croak "no data in this part\n";
    }
    1;
}

#------------------------------

=item print_header [OUTHANDLE]

I<Instance method.> 
Print the header of the message to the given output handle, 
or to the currently-selected filehandle if none was given.

All OUTHANDLE has to be is a filehandle (possibly a glob ref), or 
any object that responds to a print() message.

=cut

sub print_header {
    my ($self, $out) = @_;

    ### Coerce into a printable output handle:
    $out = wrap MIME::Lite::IO_Handle $out;

    ### Output the header:
    $out->print($self->header_as_string);
    1;
}

#------------------------------

=item as_string

I<Instance method.> 
Return the entire message as a string, with a header and an encoded body.

=cut

sub as_string {
    my $self = shift;
    my $buf = [];
    my $io = (wrap MIME::Lite::IO_ScalarArray $buf);
    $self->print($io);
    join '', @$buf;
}
*stringify = \&as_string;    ### backwards compatibility
*stringify = \&as_string;    ### ...twice to avoid warnings :)

#------------------------------

=item body_as_string

I<Instance method.> 
Return the encoded body as a string.
This is the portion after the header and the blank line.

I<Note:> actually prepares the body by "printing" to a scalar.
Proof that you can hand the C<print*()> methods any blessed object 
that responds to a C<print()> message.

=cut

sub body_as_string {
    my $self = shift;
    my $buf = [];
    my $io = (wrap MIME::Lite::IO_ScalarArray $buf);
    $self->print_body($io);
    join '', @$buf;
}
*stringify_body = \&body_as_string;    ### backwards compatibility
*stringify_body = \&body_as_string;    ### ...twice to avoid warnings :)

#------------------------------
#
# fields_as_string FIELDS
#
# PRIVATE!  Return a stringified version of the given header
# fields, where FIELDS is an arrayref like that returned by fields().
#
sub fields_as_string {
    my ($self, $fields) = @_;
    my @lines;
    foreach (@$fields) {
	my ($tag, $value) = @$_; 
	next if ($value eq '');          ### skip empties
	$tag =~ s/\b([a-z])/uc($1)/ge;   ### make pretty
	$tag =~ s/^mime-/MIME-/ig;       ### even prettier
	push @lines, "$tag: $value\n";
    }
    join '', @lines;
}

#------------------------------

=item header_as_string

I<Instance method.> 
Return the header as a string.

=cut

sub header_as_string {
    my $self = shift;
    $self->fields_as_string($self->fields);
}
*stringify_header = \&header_as_string;    ### backwards compatibility
*stringify_header = \&header_as_string;    ### ...twice to avoid warnings :)

=back

=cut



#==============================
#==============================

=head2 Sending

=over 4

=cut

#------------------------------

=item send

=item send HOW, HOWARGS...

I<Class/instance method.>  
This is the principal method for sending mail, and for configuring
how mail will be sent.

I<As a class method> with a HOW argument and optional HOWARGS, it sets 
the default sending mechanism that the no-argument instance method 
will use.  The HOW is a facility name (B<see below>), 
and the HOWARGS is interpreted by the facilty.  
The class method returns the previous HOW and HOWARGS as an array.

    MIME::Lite->send('sendmail', "d:\\programs\\sendmail.exe");
    ...
    $msg = MIME::Lite->new(...);
    $msg->send;

I<As an instance method with arguments> 
(a HOW argument and optional HOWARGS), sends the message in the 
requested manner; e.g.:

    $msg->send('sendmail', "d:\\programs\\sendmail.exe");

I<As an instance method with no arguments,> sends the message by 
the default mechanism set up by the class method.
Returns whatever the mail-handling routine returns: this should be true 
on success, false/exception on error:

    $msg = MIME::Lite->new(From=>...);
    $msg->send || die "you DON'T have mail!";

On Unix systems (at least), the default setting is equivalent to:

    MIME::Lite->send("sendmail", "/usr/lib/sendmail -t -oi -oem");

There are three facilities:

=over 4

=item "sendmail", ARGS...

Send a message by piping it into the "sendmail" command.
Uses the L<send_by_sendmail()|/send_by_sendmail> method, giving it the ARGS.
This usage implements (and deprecates) the C<sendmail()> method.

=item "smtp", [HOSTNAME]

Send a message by SMTP, using optional HOSTNAME as SMTP-sending host.
Uses the L<send_by_smtp()|/send_by_smtp> method.

=item "sub", \&SUBREF, ARGS...

Sends a message MSG by invoking the subroutine SUBREF of your choosing,
with MSG as the first argument, and ARGS following.

=back

I<For example:> let's say you're on an OS which lacks the usual Unix
"sendmail" facility, but you've installed something a lot like it, and
you need to configure your Perl script to use this "sendmail.exe" program.
Do this following in your script's setup:

    MIME::Lite->send('sendmail', "d:\\programs\\sendmail.exe");

Then, whenever you need to send a message $msg, just say:

    $msg->send;

That's it.  Now, if you ever move your script to a Unix box, all you
need to do is change that line in the setup and you're done.
All of your $msg-E<gt>send invocations will work as expected.

=cut

sub send {
    my $self = shift;

    if (ref($self)) {              ### instance method:
	my ($method, @args); 
	if (@_) {                            ### args; use them just this once
	    $method = 'send_by_' . shift;
	    @args   = @_;    
	}
	else {                               ### no args; use defaults
	    $method = "send_by_$Sender";
	    @args   = @{$SenderArgs{$Sender} || []};	    
	}
	$self->verify_data if $AUTO_VERIFY;  ### prevents missing parts!
	return $self->$method(@args);
    }
    else {                         ### class method:
	if (@_) {
	    my @old = ($Sender, @{$SenderArgs{$Sender}});
	    $Sender = shift;
	    $SenderArgs{$Sender} = [@_];    ### remaining args
	    return @old;
	}
	else {
	  Carp::croak "class method send must have HOW... arguments\n";
	}
    }
}

#------------------------------

=item send_by_sendmail SENDMAILCMD

=item send_by_sendmail PARAM=>VALUE, ...

I<Instance method.>
Send message via an external "sendmail" program
(this will probably only work out-of-the-box on Unix systems).

Returns true on success, false or exception on error.

You can specify the program and all its arguments by giving a single
string, SENDMAILCMD.  Nothing fancy is done; the message is simply
piped in.

However, if your needs are a little more advanced, you can specify 
zero or more of the following PARAM/VALUE pairs; a Unix-style, 
taint-safe "sendmail" command will be constructed for you:

=over 4

=item Sendmail

Full path to the program to use.  
Default is "/usr/lib/sendmail".

=item BaseArgs

Ref to the basic array of arguments we start with.  
Default is C<["-t", "-oi", "-oem"]>.

=item SetSender

Unless this is I<explicitly> given as false, we attempt to automatically
set the C<-f> argument to the first address that can be extracted from 
the "From:" field of the message (if there is one). 

I<What is the -f, and why do we use it?>
Suppose we did I<not> use C<-f>, and you gave an explicit "From:" 
field in your message: in this case, the sendmail "envelope" would 
indicate the I<real> user your process was running under, as a way 
of preventing mail forgery.  Using the C<-f> switch causes the sender 
to be set in the envelope as well.

I<So when would I NOT want to use it?>
If sendmail doesn't regard you as a "trusted" user, it will permit
the C<-f> but also add an "X-Authentication-Warning" header to the message
to indicate a forged envelope.  To avoid this, you can either 
(1) have SetSender be false, or 
(2) make yourself a trusted user by adding a C<T> configuration 
    command to your I<sendmail.cf> file 
    (e.g.: C<Teryq> if the script is running as user "eryq").

=item FromSender

If defined, this is identical to setting SetSender to true, 
except that instead of looking at the "From:" field we use 
the address given by this option.
Thus:

    FromSender => 'me@myhost.com'

=back

=cut

sub send_by_sendmail {
    my $self = shift;

    if (@_ == 1) {                    ### Use the given command...
	my $sendmailcmd = shift @_;

	### Do it:
	open SENDMAIL, "|$sendmailcmd" or Carp::croak "open |$sendmailcmd: $!\n";
	$self->print(\*SENDMAIL);
	close SENDMAIL;
	return (($? >> 8) ? undef : 1);
    }
    else {                            ### Build the command...
	my %p = @_;
	$p{Sendmail} ||= "/usr/lib/sendmail";   
	
	### Start with the command and basic args:
	my @cmd = ($p{Sendmail}, @{$p{BaseArgs} || ['-t', '-oi', '-oem']});

	### See if we are forcibly setting the sender:
	$p{SetSender} = 1 if defined($p{FromSender});

	### Add the -f argument, unless we're explicitly told NOT to:
	unless (exists($p{SetSender}) and !$p{SetSender}) {
	    my $from = $p{FromSender} || ($self->get('From'))[0];
	    if ($from) {
		my ($from_addr) = extract_addrs($from);
		push @cmd, "-f$from_addr"       if $from_addr;
	    }
	}

	### Open the command in a taint-safe fashion:
	my $pid = open SENDMAIL, "|-"; 
	defined($pid) or die "open of pipe failed: $!\n";
	if (!$pid) {    ### child
	    exec(@cmd) or die "can't exec $p{Sendmail}: $!\n";
	    ### NOTREACHED
	}
	else {          ### parent
	    $self->print(\*SENDMAIL);
	    close SENDMAIL || die "error closing $p{Sendmail}: $! (exit $?)\n";
	    return 1;
	}
    }
}

#------------------------------

=item send_by_smtp ARGS...

I<Instance method.>
Send message via SMTP, using Net::SMTP.  
The optional ARGS are sent into Net::SMTP::new(): usually, these are

    MAILHOST, OPTION=>VALUE, ...

Note that the list of recipients is taken from the 
"To", "Cc" and "Bcc" fields.

Returns true on success, false or exception on error.

=cut

### Provided by Andrew McRae. Version 0.2  anm  09Sep97
### Copyright 1997 Optimation New Zealand Ltd.
### May be modified/redistributed under the same terms as Perl.
#
sub send_by_smtp {
    my ($self, @args) = @_;

    ### We need the "From:" and "To:" headers to pass to the SMTP mailer:
    my $hdr  = $self->fields();   
    my $from = $self->get('From');
    my $to   = $self->get('To');

    ### Sanity check:
    defined($to) or Carp::croak "send_by_smtp: missing 'To:' address\n";
 	       
    ### Get the destinations as a simple array of addresses:
    my @to_all = extract_addrs($to);
    if ($AUTO_CC) {
	foreach my $field (qw(Cc Bcc)) {
	    my $value = $self->get($field);
	    push @to_all, extract_addrs($value) if defined($value);
	}
    }

    ### Create SMTP client:
    require Net::SMTP;
    my $smtp = MIME::Lite::SMTP->new(@args)
        or Carp::croak("Failed to connect to mail server: $!\n");
    $smtp->mail($from)
        or Carp::croak("SMTP MAIL command failed: $!\n".$smtp->message."\n");
    $smtp->to(@to_all)
        or Carp::croak("SMTP RCPT command failed: $!\n".$smtp->message."\n");
    $smtp->data()
        or Carp::croak("SMTP DATA command failed: $!\n".$smtp->message."\n");

    ### MIME::Lite can print() to anything with a print() method:
    $self->print_for_smtp($smtp);
    $smtp->dataend();
    $smtp->quit;
    1;
}

#------------------------------
#
# send_by_sub [\&SUBREF, [ARGS...]]
#
# I<Instance method, private.>
# Send the message via an anonymous subroutine.
#
sub send_by_sub {
    my ($self, $subref, @args) = @_;
    &$subref($self, @args);
}

#------------------------------

=item sendmail COMMAND...

I<Class method, DEPRECATED.>  
Declare the sender to be "sendmail", and set up the "sendmail" command.
I<You should use send() instead.>

=cut

sub sendmail {
    my $self = shift;
    $self->send('sendmail', join(' ', @_));
}

=back

=cut



#==============================
#==============================

=head2 Miscellaneous

=over 4

=cut

#------------------------------

=item quiet ONOFF

I<Class method.>  
Suppress/unsuppress all warnings coming from this module.

    MIME::Lite->quiet(1);       ### I know what I'm doing

I recommend that you include that comment as well.  And while
you type it, say it out loud: if it doesn't feel right, then maybe
you should reconsider the whole line.  C<;-)>

=cut

sub quiet {
    my $class = shift;
    $QUIET = shift if @_;
    $QUIET;
}

=back

=cut



#============================================================

package MIME::Lite::SMTP;

#============================================================
# This class just adds a print() method to Net::SMTP.
# Notice that we don't use/require it until it's needed!

use strict;
use vars qw( @ISA );
@ISA = qw(Net::SMTP);

sub print { shift->datasend(@_) }



#============================================================

package MIME::Lite::IO_Handle;

#============================================================

### Wrap a non-object filehandle inside a blessed, printable interface:
### Does nothing if the given $fh is already a blessed object.
sub wrap {
    my ($class, $fh) = @_;
    no strict 'refs';

    ### Get default, if necessary:
    $fh or $fh = select;        ### no filehandle means selected one
    ref($fh) or $fh = \*$fh;    ### scalar becomes a globref
    
    ### Stop right away if already a printable object:
    return $fh if (ref($fh) and (ref($fh) ne 'GLOB'));

    ### Get and return a printable interface:
    bless \$fh, $class;         ### wrap it in a printable interface
}

### Print:
sub print {
    my $self = shift;
    print {$$self} @_;
}


#============================================================

package MIME::Lite::IO_Scalar;

#============================================================

### Wrap a scalar inside a blessed, printable interface:
sub wrap {
    my ($class, $scalarref) = @_;
    defined($scalarref) or $scalarref = \"";
    bless $scalarref, $class;
}

### Print:
sub print {
    my $self = shift;
    $$self .= join('', @_);
    1;
}


#============================================================

package MIME::Lite::IO_ScalarArray;

#============================================================

### Wrap an array inside a blessed, printable interface:
sub wrap {
    my ($class, $arrayref) = @_;
    defined($arrayref) or $arrayref = [];
    bless $arrayref, $class;
}

### Print:
sub print {
    my $self = shift;
    push @$self, @_;
    1;
}

1;
__END__


#============================================================

=head1 NOTES


=head2 How do I prevent "Content" headers from showing up in my mail reader?

Apparently, some people are using mail readers which display the MIME
headers like "Content-disposition", and they want MIME::Lite not
to generate them "because they look ugly".  

Sigh.

Y'know, kids, those headers aren't just there for cosmetic purposes.
They help ensure that the message is I<understood> correctly by mail 
readers.  But okay, you asked for it, you got it... 
here's how you can suppress the standard MIME headers.  
Before you send the message, do this:

	$msg->scrub;

You can scrub() any part of a multipart message independently;
just be aware that it works recursively.  Before you scrub,
note the rules that I follow:

=over 4

=item Content-type

You can safely scrub the "content-type" attribute if, and only if, 
the part is of type "text/plain" with charset "us-ascii". 

=item Content-transfer-encoding

You can safely scrub the "content-transfer-encoding" attribute 
if, and only if, the part uses "7bit", "8bit", or "binary" encoding.
You are far better off doing this if your lines are under 1000 
characters.  Generally, that means you I<can> scrub it for plain
text, and you can I<not> scrub this for images, etc.

=item Content-disposition

You can safely scrub the "content-disposition" attribute 
if you trust the mail reader to do the right thing when it decides
whether to show an attachment inline or as a link.  Be aware
that scrubbing both the content-disposition and the content-type
means that there is no way to "recommend" a filename for the attachment!

B<Note:> there are reports of brain-dead MUAs out there that 
do the wrong thing if you I<provide> the content-disposition.
If your attachments keep showing up inline or vice-versa,
try scrubbing this attribute.

=item Content-length

You can always scrub "content-length" safely.

=back


=head2 How do I give my attachment a [different] recommended filename?

By using the Filename option (which is different from Path!):

	$msg->attach(Type => "image/gif",
	             Path => "/here/is/the/real/file.GIF",
	             Filename => "logo.gif");

You should I<not> put path information in the Filename.




=head2 Benign limitations

This is "lite", after all...

=over 4

=item *

There's no parsing.  Get MIME-tools if you need to parse MIME messages.

=item *

MIME::Lite messages are currently I<not> interchangeable with 
either Mail::Internet or MIME::Entity objects.  This is a completely 
separate module.

=item *

A content-length field is only inserted if the encoding is binary,
the message is a singlepart, and all the document data is available
at C<build()> time by virtue of residing in a simple path, or in-core.
Since content-length is not a standard MIME field anyway (that's right, kids:
it's not in the MIME RFCs, it's an HTTP thing), this seems pretty fair.

=item *

MIME::Lite alone cannot help you lose weight.  You must supplement
your use of MIME::Lite with a healthy diet and exercise. 

=back


=head2 Cheap and easy mailing

I thought putting in a default "sendmail" invocation wasn't too bad an 
idea, since a lot of Perlers are on UNIX systems. 
The out-of-the-box configuration is:

     MIME::Lite->send('sendmail', "/usr/lib/sendmail -t -oi -oem");

By the way, these arguments to sendmail are:

     -t      Scan message for To:, Cc:, Bcc:, etc.
              
     -oi     Do NOT treat a single "." on a line as a message terminator.
             As in, "-oi vey, it truncated my message... why?!"
                
     -oem    On error, mail back the message (I assume to the
             appropriate address, given in the header).
             When mail returns, circle is complete.  Jai Guru Deva -oem.

Note that these are the same arguments you get if you configure to use
the smarter, taint-safe mailing:

     MIME::Lite->send('sendmail');

If you get "X-Authentication-Warning" headers from this, you can forgo
diddling with the envelope by instead specifying:

     MIME::Lite->send('sendmail', SetSender=>0);

And, if you're not on a Unix system, or if you'd just rather send mail
some other way, there's always:

     MIME::Lite->send('smtp', "smtp.myisp.net");

Or you can set up your own subroutine to call.
In any case, check out the L<send()|/send> method. 



=head1 WARNINGS

=head2 Good-vs-bad email addresses with send_by_smtp()

If using L<send_by_smtp()|/send_by_smtp>, be aware that you are
forcing MIME::Lite to extract email addresses out of a possible list
provided in the C<To:>, C<Cc:>, and C<Bcc:> fields.  This is tricky
stuff, and as such only the following sorts of addresses will work
reliably:

    username
    full.name@some.host.com
    "Name, Full" <full.name@some.host.com>

This last form is discouraged because SMTP must be able to get
at the I<name> or I<name@domain> portion.

B<Disclaimer:>
MIME::Lite was never intended to be a Mail User Agent, so please
don't expect a full implementation of RFC-822.  Restrict yourself to
the common forms of Internet addresses described herein, and you should
be fine.  If this is not feasible, then consider using MIME::Lite
to I<prepare> your message only, and using Net::SMTP explicitly to 
I<send> your message.


=head2 Formatting of headers delayed until print()

This class treats a MIME header in the most abstract sense,
as being a collection of high-level attributes.  The actual
RFC-822-style header fields are not constructed until it's time
to actually print the darn thing.


=head2 Encoding of data delayed until print()

When you specify message bodies 
(in L<build()|/build> or L<attach()|/attach>) -- 
whether by B<FH>, B<Data>, or B<Path> -- be warned that we don't 
attempt to open files, read filehandles, or encode the data until 
L<print()|/print> is invoked.  

In the past, this created some confusion for users of sendmail
who gave the wrong path to an attachment body, since enough of 
the print() would succeed to get the initial part of the message out.  
Nowadays, $AUTO_VERIFY is used to spot-check the Paths given before
the mail facility is employed.  A whisker slower, but tons safer.

Note that if you give a message body via FH, and try to print() 
a message twice, the second print() will not do the right thing 
unless you  explicitly rewind the filehandle.

You can get past these difficulties by using the B<ReadNow> option,
provided that you have enough memory to handle your messages.


=head2 MIME attributes are separate from header fields!

B<Important:> the MIME attributes are stored and manipulated separately 
from the message header fields; when it comes time to print the 
header out, I<any explicitly-given header fields override the ones that
would be created from the MIME attributes.>  That means that this:

    ### DANGER ### DANGER ### DANGER ### DANGER ### DANGER ###
    $msg->add("Content-type", "text/html; charset=US-ASCII");

will set the exact C<"Content-type"> field in the header I write, 
I<regardless of what the actual MIME attributes are.>

I<This feature is for experienced users only,> as an escape hatch in case
the code that normally formats MIME header fields isn't doing what 
you need.  And, like any escape hatch, it's got an alarm on it:
MIME::Lite will warn you if you attempt to C<set()> or C<replace()>
any MIME header field.  Use C<attr()> instead.


=head2 Beware of lines consisting of a single dot

Julian Haight noted that MIME::Lite allows you to compose messages
with lines in the body consisting of a single ".".  
This is true: it should be completely harmless so long as "sendmail" 
is used with the -oi option (see L<"Cheap and easy mailing">).

However, I don't know if using Net::SMTP to transfer such a message
is equally safe.  Feedback is welcomed.

My perspective: I don't want to magically diddle with a user's 
message unless absolutely positively necessary.  
Some users may want to send files with "." alone on a line;
my well-meaning tinkering could seriously harm them.


=head2 Infinite loops may mean tainted data!

Stefan Sautter noticed a bug in 2.106 where a m//gc match was
failing due to tainted data, leading to an infinite loop inside
MIME::Lite.  

I am attempting to correct for this, but be advised that my fix will
silently untaint the data (given the context in which the problem
occurs, this should be benign: I've labelled the source code with
UNTAINT comments for the curious).

So: don't depend on taint-checking to save you from outputting
tainted data in a message.


=head2 Don't tweak the global configuration

Global configuration variables are bad, and should go away.
Until they do, please follow the hints with each setting
on how I<not> to change it.




=head1 A MIME PRIMER

=head2 Content types

The "Type" parameter of C<build()> is a I<content type>. 
This is the actual type of data you are sending.  
Generally this is a string of the form C<"majortype/minortype">.

Here are the major MIME types.
A more-comprehensive listing may be found in RFC-2046.

=over 4

=item application

Data which does not fit in any of the other categories, particularly 
data to be processed by some type of application program. 
C<application/octet-stream>, C<application/gzip>, C<application/postscript>...

=item audio

Audio data.
C<audio/basic>...

=item image

Graphics data.
C<image/gif>, C<image/jpeg>...

=item message

A message, usually another mail or MIME message.
C<message/rfc822>...

=item multipart

A message containing other messages.
C<multipart/mixed>, C<multipart/alternative>...

=item text

Textual data, meant for humans to read.
C<text/plain>, C<text/html>...

=item video

Video or video+audio data.
C<video/mpeg>...

=back


=head2 Content transfer encodings

The "Encoding" parameter of C<build()>.
This is how the message body is packaged up for safe transit.

Here are the 5 major MIME encodings.
A more-comprehensive listing may be found in RFC-2045.

=over 4

=item 7bit

Basically, no I<real> encoding is done.  However, this label guarantees that no
8-bit characters are present, and that lines do not exceed 1000 characters 
in length.

=item 8bit

Basically, no I<real> encoding is done.  The message might contain 8-bit 
characters, but this encoding guarantees that lines do not exceed 1000 
characters in length.

=item binary

No encoding is done at all.  Message might contain 8-bit characters,
and lines might be longer than 1000 characters long.

The most liberal, and the least likely to get through mail gateways.  
Use sparingly, or (better yet) not at all.

=item base64

Like "uuencode", but very well-defined.  This is how you should send
essentially binary information (tar files, GIFs, JPEGs, etc.). 

=item quoted-printable

Useful for encoding messages which are textual in nature, yet which contain 
non-ASCII characters (e.g., Latin-1, Latin-2, or any other 8-bit alphabet).

=back



=head1 VERSION

$Id: Lite.pm,v 2.117 2001/08/20 20:40:39 eryq Exp $


=head1 CHANGE LOG

=over 4


=item Version 2.117   (2001/08/20)

The terms-of-use have been placed in the distribution file "COPYING".  
Also, small documentation tweaks were made.


=item Version 2.116   (2001/08/17)

Added long-overdue patch which makes the instance method form
of send() do the right thing when given HOW... arguments.
I<Thanks to Casey West for the patch.>

=item Version 2.114   (2001/08/16)

New special 'AUTO' content type in new()/build() tells MIME::Lite to 
try and guess the type from file extension.  To make use of 
this, you'll want to install B<MIME::Types>.
The "AUTO" setting can be made the default default (instead of "TEXT")
if you set C<$AUTO_CONTENT_TYPE = 1, $PARANOID = 0>.
I<Thanks to> Ville SkyttE<#228> I<for these patches.>

File::Basename is used if it is available.
I<Thanks to> Ville SkyttE<#228> I<for this patch.>

SMTP failures (in send_by_smtp) now add the $smtp-E<gt>message to the
croak'ed exception, so if things go wrong, you get a better
idea of what and why.
I<Thanks to Thomas R. Wyant III for the patch.>

Made a subtle change to C<as_string> which supposedly fixes a 
failed MIME data.t test with Perl 5.004_04 on NT 4 sp6.  
The problem might only exist in this old perl, but as the patch 
author says, not everyone has climbed higher on the Perl ladder.
I<Thanks to John Gotts for the patch.>

Added C<contrib> directory, with F<MailTool.pm>.
I<Thanks to Tom Wyant for this contribution.>

Improved HTML documentation (notice the links to
the individual methods in the top menu).

Corrected some mis-docs.


=item Version 2.111   (2001/04/03)

Added long-overdue C<parts()> and C<parts_DFS()> methods.

    No instance method
       For accessing the subparts?			   
    That can't be right.  D'OH!		 

Added long-overdue auto-verify logic to C<print()> method.

Added long-overdue C<preamble()> method for getting/setting 
the preamble text.  
I<Thanks to Jim Daigle for inspiring this.>


=item Version 2.108   (2001/03/30)

New C<field_order()> allows you to set the header order, both on a 
per-message basis, and package-wide.
I<Thanks to Thomas Stromberg for suggesting this.>

Added code to try and divine "sendmail" path more intelligently.
I<Thanks to Slaven Rezic for the suggestion.>


=item Version 2.107   (2001/03/27)

Fixed serious bug where tainted data with quoted-printable encoding
was causing infinite loops.  The "fix" untaints the data in question,
which is not optimal, but it's probably benign in this case.
I<Thanks to Stefan Sautter for tracking this nasty little beast down.>
I<Thanks to Larry Geralds for a related patch.>

    "Doctor, O doctor:
       it's painful when I do *this* --" 
    "Simple: don't *do* that." 

Fixed bugs where a non-local C<$_> was being modified... again!  
Will I never learn?
I<Thanks to Maarten Koskamp for reporting this.>

    Dollar-underscore
       can poison distant waters;
   'local' must it be.

Fixed buglet in C<add()> where all value references were being treated
as arrayrefs, instead of as possibly-self-stringifying object refs.
Now you can send in an object ref as the 2nd argument.
I<Thanks to dLux for the bug report.>

    That ref is a string?
       Operator overload
    has ruined my day.

Added "Approved" as an acceptable header field for C<new()>, as per RFC1036.
I<Thanks to Thomax for the suggestion regarding MIME-tools.>

Small improvements to docs to make different uses of attach() 
and various arguments clearer.
I<Thanks to Sven Rassman and Roland Walter for the suggestions.>


=item Version 2.106   (2000/11/21)

Added Alpha version of scrub() to make it easy for people to suppress
the printing of unwanted MIME attributes (like Content-length).
I<Thanks to the many people who asked for this.>

Headers with empty-strings for their values are no longer
printed.  This seems sensible, and helps us implement scrub().


=item Version 2.105   (2000/10/14)

The regression-test failure was identified, and it was my fault.
Apparently some of the \-quoting in my "autoloaded" code was
making Perl 5.6 unhappy.  For this nesting-related idiocy, 
a nesting kaiku.
I<Thanks to Scott Schwartz for identifying the problem.>

    In a pattern, my
       backslash-s dwells peacefully,
    unambiguous --
     
       but I embed it
          in a double-quoted string    
       doubling the backslash --
     
          interpolating
             that same double-quoted string 
          in other patterns --
           
             and, worlds within worlds,
                I single-quote the function
             to autoload it -- 
    
          changing the meaning
       of the backslash and the 's';
    and Five-Point-Six growls.


=item Version 2.104   (2000/09/28)

Now attempts to load and use Mail::Address for parsing email 
addresses I<before> falling back to our own method.
I<Thanks to numerous people for suggesting this.>

    Parsing addresses
       is too damn hard. One last hope:
    Let Graham Barr do it!

For the curious, the version of Mail::Address appears 
as the "A" number in the X-Mailer:

    X-Mailer: MIME::Lite 2.104  (A1.15; B2.09; Q2.03)

Added B<FromSender> option to send_by_sendmail().
I<Thanks to Bill Moseley for suggesting this feature.>


=item Version 2.101   (2000/06/06)

Major revision to print_body() and body_as_string() so that
"body" really means "the part after the header", which is what most
people would want in this context.  This is B<not> how it was used
1.x, where "body" only meant "the body of a simple singlepart".
Hopefully, this change will solve many problems and create very few ones.  

Added support for attaching a part to a "message/rfc822", treating
the "message" type as a multipart-like container.

Now takes care not to include "Bcc:" in header when using send_by_smtp,
as a safety precaution against qmail's behavior.
I<Thanks to Tatsuhiko Miyagawa for identifying this problem.>

Improved efficiency of many stringifying operations by using 
string-arrays which are joined, instead of doing multiple appends 
to a scalar.

Cleaned up the "examples" directory.


=item Version 1.147   (2000/06/02)

Fixed buglet where lack of Cc:/Bcc: was causing extract_addrs
to emit "undefined variable" warnings.  Also, lack of a "To:" field
now causes a croak.
I<Thanks to David Mitchell for the bug report and suggested patch.>


=item Version 1.146   (2000/05/18)

Fixed bug in parsing of addresses; please read the WARNINGS section
which describes recommended address formats for "To:", "Cc:", etc.
Also added automatic inclusion of a UT "Date:" at top level unless 
explicitly told not to.
I<Thanks to Andy Jacobs for the bug report and the suggestion.>

=item Version 1.145   (2000/05/06)

Fixed bug in encode_7bit(): a lingering C</e> modifier was removed.
I<Thanks to Michael A. Chase for the patch.>


=item Version 1.142   (2000/05/02)

Added new, taint-safe invocation of "sendmail", one which also
sets up the C<-f> option.  Unfortunately, I couldn't make this automatic:
the change could have broken a lot of code out there which used 
send_by_sendmail() with unusual "sendmail" variants.  
So you'll have to configure "send" to use the new mechanism:

    MIME::Lite->send('sendmail');       ### no args!

I<Thanks to Jeremy Howard for suggesting these features.>


=item Version 1.140   (2000/04/27)

Fixed bug in support for "To", "Cc", and "Bcc" in send_by_smtp():
multiple (comma-separated) addresses should now work fine.
We try real hard to extract addresses from the flat text strings.
I<Thanks to John Mason for motivating this change.>

Added automatic verification that attached data files exist,
done immediately before the "send" action is invoked.
To turn this off, set $MIME::Lite::AUTO_VERIFY to false.

=item Version 1.137   (2000/03/22)

Added support for "Cc" and "Bcc" in send_by_smtp().
To turn this off, set $MIME::Lite::AUTO_CC to false.
I<Thanks to Lucas Maneos for the patch, and tons of others for 
the suggestion.>

Chooses a better default content-transfer-encoding if the content-type
is "image/*", "audio/*", etc.
To turn this off, set $MIME::Lite::AUTO_ENCODE to false.
I<Thanks to many folks for the suggestion.>

Fixed bug in QP-encoding where a non-local C<$_> was being modified.
I<Thanks to Jochen Stenzel for finding this very obscure bug!>

Removed references to C<$`>, C<$'>, and C<$&> (bad variables
which slow things down).

Added an example of how to send HTML files with enclosed in-line
images, per popular demand.


=item Version 1.133   (1999/04/17)

Fixed bug in "Data" handling: arrayrefs were not being handled
properly.


=item Version 1.130   (1998/12/14)

Added much larger and more-flexible send() facility.
I<Thanks to Andrew McRae (and Optimation New Zealand Ltd) 
for the Net::SMTP interface.  Additional thanks to the many folks
who requested this feature.>

Added get() method for extracting basic attributes.

New... "t" tests!


=item Version 1.124   (1998/11/13)

Folded in filehandle (FH) support in build/attach.
I<Thanks to Miko O'Sullivan for the code.>


=item Version 1.122   (1998/01/19)

MIME::Base64 and MIME::QuotedPrint are used if available.

The 7bit encoding no longer does "escapes"; it merely strips 8-bit characters.


=item Version 1.121   (1997/04/08)

Filename attribute is now no longer ignored by build().
I<Thanks to Ian Smith for finding and patching this bug.>


=item Version 1.120   (1997/03/29)

Efficiency hack to speed up MIME::Lite::IO_Scalar.
I<Thanks to David Aspinwall for the patch.>


=item Version 1.116   (1997/03/19)

Small bug in our private copy of encode_base64() was patched.
I<Thanks to Andreas Koenig for pointing this out.>

New, prettier way of specifying mail message headers in C<build()>.

New quiet method to turn off warnings.

Changed "stringify" methods to more-standard "as_string" methods.


=item Version 1.112   (1997/03/06)

Added C<read_now()>, and C<binmode()> method for our non-Unix-using brethren: 
file data is now read using binmode() if appropriate.
I<Thanks to Xiangzhou Wang for pointing out this bug.>


=item Version 1.110   (1997/03/06)

Fixed bug in opening the data filehandle.


=item Version 1.102   (1997/03/01)

Initial release.


=item Version 1.101   (1997/03/01)

Baseline code.
Originally created: 11 December 1996.  Ho ho ho.

=back


=head1 TERMS AND CONDITIONS

Copyright (c) 1997 by Eryq.  
Copyright (c) 1998 by ZeeGee Software Inc.
All rights reserved.  This program is free software; you can redistribute 
it and/or modify it under the same terms as Perl itself.  

This software comes with B<NO WARRANTY> of any kind.
See the COPYING file in the distribution for details.


=head1 NUTRITIONAL INFORMATION

For some reason, the US FDA says that this is now required by law
on any products that bear the name "Lite"...

    MIME::Lite                | 
    ------------------------------------------------------------
    Serving size:             | 1 module
    Servings per container:   | 1
    Calories:                 | 0
    Fat:                      | 0g
      Saturated Fat:          | 0g

Warning: for consumption by hardware only!  May produce 
indigestion in humans if taken internally.


=head1 AUTHOR

Eryq (F<eryq@zeegee.com>).
President, ZeeGee Software Inc. (F<http://www.zeegee.com>).

Go to F<http://www.zeegee.com> for the latest downloads
and on-line documentation for this module.  Enjoy.

=cut

NMSAI_END_PM_FILE

    'NMStreq.pm' => <<'NMSAI_END_PM_FILE',
package NMStreq;
use strict;

use CGI;
use Carp;
use IO::File;
use File::Basename;
use POSIX qw(locale_h strftime);
use NMSCharset;

use vars qw($VERSION);
$VERSION = substr q$Revision: 1.16 $, 10, -1;

=head1 NAME

NMStreq - CGI request object with output templating

=head1 SYNOPSIS

  use IO::File;
  use NMStreq;

  my $treq = NMStreq->new( ConfigRoot => '/my/config/root' );

  ....

  my $sendmail = IO::File->new('| /usr/lib/sendmail -oi -t');
  defined $sendmail or die "open sendmail pipe: $!";
  $sendmail->print($mailheader, "\n");
  $treq->process_template(
      $treq->config('email_body_template', 'main_email'),
      'email',
      $sendmail
  );
  $sendmail->close or die "close sendmail pipe: $!";

  ....

  print "Content-type: text/html; charset=iso-8859-1\n\n";

  $treq->process_template(
      $treq->config('success_page_template', 'spage'),
      'html',
      \*STDOUT
  );

  ....

=head1 DESCRIPTION

An object of the C<NMStreq> class encapsulates a CGI
request who's handing depends on a configuration file
identified by the C<_config> CGI parameter.  A
simplistic templating mechanism is provided, to ease
end user customization of the output HTML and the
bodies of any emails sent.

=head1 CONSTRUCTORS

=over

=item new ( [OPTIONS] )

Creates a new C<NMStreq> object and populates it with
data pertinent to the current CGI request.  The CGI
parameter C<_config> will be used to identify the
correct configuration file for this request.  The
OPTIONS must consist of matching name/value pairs,
and the following options are recognized:

=over

=item C<ConfigRoot>

The filesystem path to the directory that holds the
configuration files and templates.  Defaults to
F</usr/local/nmstreq/config>.

=item C<MaxDepth>

The depth to which configuration files and templates
can be placed in subdirectories of the C<ConfigRoot>.
Defaults to 0, meaning that all configuration files
must reside directly in the C<ConfigRoot> directory.

=item C<ConfigExt>

The extension that configuration files are expected to
have.  Defaults to C<.trc>.

=item C<TemplateExt>

The extension that template files are expected to have.
Defaults to C<.trt>.

=item C<DateFormat>

The default date format string that will be used to
resolve the C<date> template directive if no C<date_fmt>
configuration setting is found.  Defaults to
C<%A, %B %d, %Y at %H:%M:%S>.

=item C<EnableUploads>

Unless this is set true, file uploads will be disabled in
C<CGI.pm>.  Defaults to false.

=item C<CGIMaxPost>

The maximum total size of post data.  Defaults to 1000000
bytes.

=item C<Charset>

The name of the character set to be used for input and
output text, used to initialise an C<NMSCharset> object,
see L<NMSCharset>.  Defaults to C<iso-8859-1>.

=back

Any other options set will be ignored by this module,
but can be interpolated into templates via the C<opt>
template directive if desired.

=back

=cut

sub new
{
   my $pkg = shift;

   my $self = bless {}, ref $pkg || $pkg;

   $self->{r}{opt} = $self->{opt} = {
      ConfigRoot    => '/usr/local/nmstreq/config',
      MaxDepth      => 0,
      ConfigExt     => '.trc',
      TemplateExt   => '.trt',
      DateFormat    => '%A, %B %d, %Y at %H:%M:%S',
      EnableUploads => 0,
      CGIPostMAx    => 1000000,
      Charset       => 'iso-8859-1',
      @_
   };

   $CGI::DISABLE_UPLOADS = ($self->{opt}{EnableUploads} ? 0 : 1);
   $CGI::POST_MAX        = $self->{opt}{CGIPostMax};

   my $charset = NMSCharset->new($self->{opt}{Charset});
   $self->{strip_nonprint} = $charset->strip_nonprint_coderef;
   $self->{escape_html}    = $charset->escape_html_coderef;

   my $cgi = CGI->new;
   $self->{cgi} = $cgi;

   my $cfg_name = $cgi->param('_config');
   defined $cfg_name or $cfg_name = 'default';
   $self->{r}{config} = $self->_read_config_file($cfg_name);

   # cache location of the config file to find the templates
   
   $self->{r}{config_path} 
                      = dirname($self->{opt}{ConfigRoot} . "/" . $cfg_name);

   $self->{r}{param} = {};
   my @param_list = ();
   foreach my $param ($cgi->param)
   {
      my $key = $self->strip_nonprintable($param);
      push @param_list, $key unless exists $self->{r}{param}{$key};

      my $val = join ' ',
                map {$self->strip_nonprintable($_)}
                $cgi->param($param);
      $self->{r}{param}{$key} = $val;
   }
   $self->{param_list} = \@param_list;

   foreach my $envval (keys %ENV)
   {
      my $key = $self->strip_nonprintable($envval);
      my $val = $self->strip_nonprintable($ENV{$envval});
      $self->{r}{env}{$key} = $val;
   }

   $self->{r}{date}         = \&_interpolate_date;

   return $self;
}

=head1 METHODS

=over

=item process_template ( TEMPLATE, CONTEXT, DEST )

Reads in the template TEMPLATE, which can either be an inline
template as a multiline string or the path to a template file,
relative to the configuration root and without the file
extension.  Data is substituted for any template directives
in the template, and the resulting document is passed out to
DEST.

CONTEXT is a string describing the context of the output
document, and must be either C<html> or C<email>.  If CONTEXT
is C<html> then all HTML metacharacters in interpolated
values will be escaped.  If CONTEXT is C<email> then space
characters will be inserted at a couple of points, to reduce
the scope for malicious input values to make mail software do
bad things.

DEST can be a coderef, a file glob, an object with a
print() method, or undef.

On failure, invokes the non-returning error() method.

If DEST is undef, then all template output is accumulated
into a string, which becomes the return value.

=cut

sub process_template
{
   my ($self, $template, $context, $dest) = @_;

   my ($ret, $coderef);
   if (defined $dest)
   {
      $ret = 1;
      $coderef = $self->_dest_to_coderef($dest);
   }
   else
   {
      $ret = '';
      $coderef = sub { $ret .= $_[0] };
   }

   my $complied = $self->_compile_template($template, $context);
   $self->_run_template($complied, $context, $coderef);

   return $ret;
}

=item install_directive ( NAME, VALUE )

Installs an extra directive into the data tree used for
interpolating values into templates.  NAME must be a
string consisting of word characters only.  VALUE can
be any of:

=over

=item C<a string>

If VALUE is a string then that string will be substituted
for the NAME template directive.

=item C<a reference to a string>

If VALUE is a scalar reference then the referenced string
will be substituted for the NAME template directive, without
any context dependent processing.  The string goes directly to
the output document, without HTML metacharacter escaping in
an html context or sanitisation in an email context.

Use this only for trusted data or data that has already been
carefully filtered for HTML or other malicious constructs. 

=item C<a coderef>

If VALUE is a coderef then it will be called to produce the
substitute text whenever the NAME directive is encountered.
It will be passed a reference to the C<NMStreq> object as
its first argument, the context string ("html" or "email")
as its second argument, and a destination coderef as its
third argument.  The VALUE coderef can pass output direct
to the destination coderef, and/or return some output as
a string.

=item C<a hashref>

In this case a new tree of two-part directives is defined,
with the sub-directives corresponding to the keys in the
hash.  The values in the hash must be strings, coderefs
or further hashrefs.

=back

=cut

sub install_directive
{
   my ($self, $name, $value) = @_;

   $self->{r}{$name} = $value;
}

=item uninstall_directive ( NAME )

Removes a directive previously installed with the
install_directive() method, or disables one of the builtin
directives.

Returns a value which will reinstall the directive if passed
to the install_directive() method.

=cut

sub uninstall_directive
{
   my ($self, $name) = @_;

   my $save = $self->{r}{$name};
   delete $self->{r}{$name};
   return $save;
}

=item install_foreach ( NAME, VALUES )

Installs data to support a FOREACH directive in templates.
NAME should be the name to appear in the FOREACH directive,
and VALUES must be a ref to an array of hashes, with each
hash defining values for local variables for one iteration
of the FOREACH loop.  For example, this code:

  $treq->install_foreach( 'foobar', [
    { foo => 'foo1', bar => 'bar7' },
    { foo => 'foo2', bar => 'bar4' },
    { foo => 'foo3', bar => 'bar9' },
  ]);

would cause this template segment:

  {= FOREACH foobar =}
  The foo is {= foo =}, but the bar is {= bar =}!
  {= END =}

to produce the output:

  The foo is foo1, but the bar is bar7!
  The foo is foo2, but the bar is bar4!
  The foo is foo3, but the bar is bar9!

The values can be references to strings rather than strings, to
prevent context dependent processing, as in install_directive()
above.  Use this feature only with trusted or already filtered
data, since it bypasses HTML metacharacter escaping and could
lead to XSS vulnerabilities if misapplied.
 
=cut

sub install_foreach
{
   my ($self, $name, $values) = @_;

   $self->{'foreach'}{$name} = $values;
}

=item uninstall_foreach ( NAME )

Removes a foreach data set previously installed with
the install_foreach() method.

Returns a value which will reinstall the foreach data
if passed to the install_foreach() method.

=cut

sub uninstall_foreach
{
   my ($self, $name) = @_;

   my $save = $self->{'foreach'}{$name};
   delete $self->{'foreach'}{$name};
   return $save;
}

=item config ( SETTING_NAME, DEFAULT )

Returns the value of the configuration setting SETTING_NAME
set in the configuration file for this request, or DEFAULT
if no value for SETTING_NAME has been set.

=cut

sub config
{
   my ($self, $setting_name, $default) = @_;

   my $val = $self->{r}{config}{$setting_name};
   defined $val ? $val : $default;
}

=item param ( PARAM_NAME )

Returns the value of the CGI parameter PARAM_NAME, with
runs of nonprintable characters replaced with spaces.
If the same CGI parameter appears several times then all
the values of that parameter are joined together, using
a single space character as a separator.

Returns the empty string if no such parameter is set.

=cut

sub param
{
   my ($self, $param_name) = @_;

   my $val = $self->{r}{param}{$param_name};
   defined $val ? $val : '';
}

=item param_list ()

Returns a list of the names of all CGI parameters.  The
parameter names are returned in the order in which each
parameter first occurs in the request.  There will be no
duplicates in the list returned.

Runs of nonprintable characters in parameter names are
replaced with spaces, both in the list returned by this
method and in the parameter names recognized by the
param() method.

=cut

sub param_list
{
   my ($self) = @_;

   return @{ $self->{param_list} };
}

=item cgi ()

Returns a reference to the C<CGI> object that this modules
uses to access the CGI parameter list.

=cut

sub cgi
{
   my ($self) = @_;

   return $self->{cgi};
}

=back

=head1 METHODS TO OVERRIDE

Subclasses may override any of the following methods in
order to alter the class's behavior.

=over

=item error ( MESSAGE )

A non-returning method used to handle fatal errors.  The
MESSAGE string may contain unsafe and potentially malicious
data and so must be handled with care.

This method must not return.

The default implementation calls croak().

=cut

sub error
{
   my ($self, $message) = @_;

   croak $message;
}

=item strip_nonprintable ( STRING )

Returns a copy of STRING with runs of non-printable
characters replaced with space.  The default implementation
uses the coderef provided by the C<NMSCharset> module, see
L<NMSCharset>.

=cut

sub strip_nonprintable
{
   my ($self, $string) = @_;

   &{ $self->{strip_nonprint} }( $string );
}

=item escape_html ( STRING )

Returns a copy of STRING with any HTML metacharacters escaped.
The default implementation uses the coderef provided by the
C<NMSCharset> module, see L<NMSCharset>.

=cut

sub escape_html
{
   my ($self, $string) = @_;

   &{ $self->{escape_html} }( $string );
}

=back

=head1 INTERNAL METHODS

None of these methods should be accessed from outside this
module.

=over

=item _compile_template ( TEMPLATE, CONTEXT )

Reads in a template for context CONTEXT from TEMPLATE
(which can be either a template filename relative to the
configuration root or an inline template as a multiline
string) and compiles it to the following internal
representation:

The compiled template is an array ref, each element of
which is one of:

=over

=item C<a scalar>

Some literal text from the template

=item C<a scalar reference>

The referenced string is the contents of a template
directive other than a control structure.

=item C<a hash reference>

The referenced hash represents a control structure.  The
C<ctl> value is a string that defines the type of control
structure (C<FOREACH> and C<IF>/C<ELSE> are defined).  The
C<sub> value is an array reference, holding the control
structure body as a compiled template.  The C<arg> value
is the argument string (if any) that appeared in the control
directive.  In the case of an C<IF> directive with an
C<ELSE> block, the compiled template for the else block is
stored as C<esub>.

=back

For example, this template:

  %% NMS email template file %%
  Today is {= date =}, you are {= env.REMOTE_USER =} and
  your inputs were:
  {= FOREACH input_field =}
  {= name =}: {= value =}
  {= END =}
  {= IF param.hello =}Hello!{= ELSE =}Goodbye!{= END =}
  ----

Would compile to the array ref:

  [
    "Today is ",
    \'date',
    ", you are ",
    \'env.REMOTE_USER',
    " and\n",
    "your inputs were:\n",
    {
      'ctl' => 'FOREACH'
      'arg' => 'input_field',
      'sub' => [ \'name', ": ", \'value', "\n" ],
    },
    "----\n",
    {
      'ctl'  => 'IF',
      'arg'  => 'param.hello',
      'sub'  => [ 'Hello!' ],
      'esub' => [ 'Goodbye!' ],
    },
  ]

Returns the compiled template as an array ref, or dies on
error.

=cut

sub _compile_template
{
   my ($self, $template, $context) = @_;

   my @lines;
   if ($template =~ /%/)
   {
      # An inline template as a string
      @lines = map { /^%(.*)/ ? ("$1\n") : () } split /\n/, $template;
   }
   else
   {
      # The name of a template in an external file
      my $fh = $self->_open_file($template, "$context template");
      @lines = <$fh>;
      $fh->close;
   }

   my $compiled = [];
   my @stack = ($compiled);

   local $_;
   foreach(@lines)
   {
      # Ditch trailing whitespace, in particular get rid of the spare \r
      # if the template had \r\n line termination.
      s#\s+$#\n#;

      # Produce no output for a control directive alone on a line
      s#^ \s* (\{\= \s*[A-Z]+\s*[\s\w\-\.]+ \=\}) \n#$1#x;

      while ( s#(.*?) \{\= \s* (.*?) \s* \=\} ##x )
      {
         my ($pre, $directive) = ($1, $2);
         push @{ $stack[0] }, $pre if length $pre;
         if ($directive =~ s/^(FOREACH|IF)\s*//)
         {
            my $sub = [];
            push @{ $stack[0] }, { 'ctl' => $1,
                                   'arg' => $directive,
                                   'sub' => $sub
                                 };
            unshift @stack, $sub;
         }
         elsif ($directive =~ /^END$/i)
         {
            shift @stack;
            die 'misplaced END directive' unless scalar @stack;
         }
         elsif ($directive =~ /^ELSE$/i)
         {
            shift @stack;
            die 'misplaced ELSE directive' unless scalar @stack;

            my $if = ${ $stack[0] }[-1];
            die 'ELSE outside IF' unless $if->{ctl} eq 'IF';
            die 'only one ELSE per IF' if exists $if->{esub};

            my $esub = [];
            $if->{esub} = $esub;
            unshift @stack, $esub;
         }
         else
         {
            push @{ $stack[0] }, \$directive;
         }
      }

      push @{ $stack[0] }, $_ if length;
   }

   return $compiled;
}

=item _run_template ( TEMPLATE, CONTEXT, CODEREF )

Runs a pre-compiled template, and dies on error.

The TEMPLATE parameter must be a compiled template, as
returned by the _compile_template() method.  CONTEXT is
the context string and CODEREF is the output destination
coderef.

=cut

sub _run_template
{
   my ($self, $template, $context, $coderef) = @_;

   foreach my $part (@$template)
   {
      if (ref $part eq 'HASH')
      {
         if ($part->{ctl} eq 'FOREACH')
         {
            my $vals = $self->{'foreach'}{$part->{arg}};
            defined $vals or die "[$part->{arg}] cannot be used in a FOREACH directive";
   
            foreach my $val (@$vals)
            {
               foreach my $k (keys %$val)
               {
                  $self->install_directive($k, $val->{$k});
               }
               $self->_run_template($part->{'sub'}, $context, $coderef);
               foreach my $k (keys %$val)
               {
                  $self->uninstall_directive($k);
               }
            }
         }
         elsif ($part->{ctl} eq 'IF')
         {
            my $val = '';
            my $callback = sub { $val .= $_[0] };
            $self->_interpolate($part->{arg}, $context, $callback);
            if ($val)
            {
               $self->_run_template($part->{'sub'}, $context, $coderef);
            }
            elsif (exists $part->{'esub'})
            {
               $self->_run_template($part->{'esub'}, $context, $coderef);
            }
         }
         else
         {
            die "[$part->{ctl}] unsupported";
         }
      }
      elsif (ref $part eq 'SCALAR')
      {
         $self->_interpolate($$part, $context, $coderef);
      }
      elsif (length $part)
      {
         &{ $coderef }($part);
      }
   }
}

=item _interpolate ( DIRECTIVE, CONTEXT, CODEREF )

Resolves a single template directive in context CONTEXT
and outputs the result via the coderef CODEREF.  DIRECTIVE
is the string found between the template directive
delimiters, with leading and trailing whitespace removed.

=cut

sub _interpolate
{
   my ($self, $directive, $context, $coderef) = @_;

   my $data_src = $self->{r};
   while ($directive =~ s#^(\w+)\.##)
   {
      $data_src = $data_src->{$1};
      defined $data_src or return;
      ref $data_src eq 'HASH' or return;
   }

   my $value = $data_src->{$directive};
   defined $value or return;

   if (ref $value eq 'CODE')
   {
      $value = &{ $value }($self, $context, $coderef);
   }

   if (ref $value)
   {
      return unless ref $value eq 'SCALAR';
      # reference to value means don't munge value, see install_directive()
      $value = $$value;
   }
   else
   {
      if ($context eq 'html')
      {
         $value = $self->escape_html($value);
      }
      elsif ($context eq 'email')
      {
         # Disable HTML tags with minimum impact
         $value =~ s#<([a-z])#< $1#gi;
   
         # Don't allow multiline inputs to control the first
         # character of the line.
         $value =~ s#(\r|\n)(\S)#$1 $2#g;
   
         # Could be trying to fake a MIME boundary.
         $value =~ s/------/ ------/g;
      }
      else
      {
         $self->error("unknown template context [$context]");
      }
   }

   &{ $coderef }($value) if length $value;
   return;
}

=item _interpolate_date ( CONTEXT, CODEREF )

Resolves a C<date> template directive. Will use the date_fmt config to
determine the format of the date and locale item if present to localize
appropriate parts of the date string.

=cut

sub _interpolate_date
{
   my ($self, $context, $coderef) = @_;

   my $date_fmt = $self->{r}{'config'}{date_fmt};
   
   my $old_locale;

   
   if ( my $locale = $self->config('locale') )
   {
      $old_locale = POSIX::setlocale( LC_TIME );
      POSIX::setlocale(LC_TIME, $locale );
   }

   defined $date_fmt or $date_fmt = $self->{opt}{DateFormat};

   my $date = strftime $date_fmt, localtime;

   if ( $self->config('locale',0) )
   {
      POSIX::setlocale(LC_TIME, $old_locale);
   }
    
   # cache so that all date directives in a single request get the
   # same date.
   $self->{r}{date} = $date;

   return $date;
}

=item _dest_to_coderef ( DEST )

Converts a template output destination (which can be a
coderef, a file glob or an object reference) into a
coderef.

=cut

sub _dest_to_coderef
{
   my ($self, $dest) = @_;

   if (ref $dest eq 'CODE')
   {
      return $dest;
   }
   elsif (ref $dest eq 'GLOB')
   {
      return sub { print $dest $_[0] or $self->error("write failed: $!") };
   }
   else
   {
      return sub { $dest->print($_[0]) or $self->error("print failed: $!") };
   }
}

=item _read_config_file ( CONFIG_FILE )

Reads in and interprets the configuration file CONFIG_FILE,
which must be the path to a config file, relative to
the configuration root and without the file extension.

On success, returns a reference to a hash of configuration
settings.

On failure, invokes the non-returning error() method.

=cut

sub _read_config_file
{
   my ($self, $cfg_file) = @_;

   my $fh = $self->_open_file($cfg_file, 'configuration');

   my %config = ();
   my $key = '**NOKEY**';
   local $_;
   while(<$fh>)
   {
      next if m%^\s*(#|$)%;
      $key = $1 if s#^(\w+):##;
      s#^\s*##;
      s#\s*$##;
      next unless length;
      $config{$key} = (defined $config{$key} ? "$config{$key}\n$_" : $_);
   }
   delete $config{'**NOKEY**'};

   $fh->close;
   return \%config;
}

=item _open_file ( FILENAME, FILETYPE )

Checks that FILENAME is a valid relative file path without
file extension for a template or configuration file, opens
the file, checks that it has the correct header line and
returns an C<IO::File> object from which the remainder of
the file can be read.

The FILETYPE parameter should be one of the following
strings: "configuration", "S<html template>" or
"S<email template>".

Calls the non-returning error() method if anything goes
wrong.

=cut

sub _open_file
{
   my ($self, $filename, $filetype) = @_;


   unless ( $filename =~ m#^[a-zA-Z0-9]# and
            $filename =~ m#[a-zA-Z0-9]$# and
            $filename =~ m#^([/a-zA-Z0-9_]{1,100})$# )
   {
      $self->error("Invalid character in filename [$filename]");
   }
   $filename = $1;

   $filename =~ s#/+#/#g;
   my $slashcount = $filename =~ tr#/##;

   if ($slashcount > $self->{opt}{MaxDepth})
   {
      $self->error("$filetype filename [$filename] contains too many '/' characters");
   }

   my $ext;
   if ( $filetype eq 'configuration' )
   {
      $ext = $self->{opt}{ConfigExt};
   }
   elsif ( $filetype =~ / template$/ )
   {
      $ext = $self->{opt}{TemplateExt};
   }
   else
   {
      error("bad file type [$filetype]");
   }

   my $path = "$self->{opt}{ConfigRoot}/$filename$ext";

   my $file_exists = -f $path;

   if ( $filetype =~ / template$/ and ! $file_exists)
   {
      $path = "$self->{r}{config_path}/$filename$ext";
      $file_exists = -f $path; 
   }

   if ( !$file_exists)
   {
      $self->error("$filetype file not found: [$filename]");
   }

   my $fh = IO::File->new("<$path");
   unless (defined $fh)
   {
      $self->error("failed to open $filetype file [$filename] ($!)");
   }

   my $header = <$fh>;
   unless (defined $header and
           $header =~ m#^\%\% NMS \Q$filetype\E file \%\%\s*$#)
   {
      $self->error("$filetype file [$filename]: invalid header line [$header]");
   }

   return $fh;
}

=back

=cut

1;

__END__

=head1 CONFIGURATION FILE SYNTAX

Each configuration file sets values for a set of named keys.

The key names can consist of word characters only.  The
values can contain any character, but whitespace sequences
at the start and end of each line will be discarded when
the configuration file is parsed.

The first line of the template file must be the exact text:

  %% NMS configuration file %%

Lines starting with '#' are ignored.  Whitespace can
precede the '#' character.

Any set of one or more word characters followed by a ':'
character at the start of a line introduces a new key.  All
text until another key is introduced becomes the value for
that key.

If a key appears more than once then the values will be
joined using a space character as a delimiter.

For example:

  %% NMS configuration file %%
  #
  # This is an example of a configuration file.  It assigns
  # the value "one two three four" to key 'foo' and the value
  # "1   2 3 4" to key 'bar'.
  #

  foo: one two
       # This is an indented comment
       three

  bar: 1   2
  bar:3 4

  foo:
       four

=head1 TEMPLATE FILE SYNTAX

The first line of any template file must be either:

  %% NMS html template file %%

or

  %% NMS email template file %%

depending on the context in which the template is to be used.

All other lines in the template will be copied to the output
with template directives replaced by the corresponding data
values.  Template directives consist of the string "{=",
optional whitespace, the directive name, optional whitespace,
and the string "=}".  The directive names can be simple words
such as "date" or constructs such as "param.foo".

Template directives may not be split over multiple lines.

Here is an example of an HTML template:

  %% NMS html template file %%
  <?xml version="1.0" encoding="iso-8859-1"?>
  <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
      "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
  <html xmlns="http://www.w3.org/1999/xhtml">
   <head>
    <title>{= config.html_title =}</title>
   </head>
   <body>
    <h1>{= config.html_title =}</h1>
    <p>
     Hello, the date is {= date =} and your user agent is
     <i>{= env.HTTP_USER_AGENT =}</i>.
    </p>
    <p>
     You put <b>{= param.foo =}</b> in the <b>foo</b> input.
    </p>
   </body>
  </html>

The directives that can be used are:

=over

=item C<config.*>

The C<config.html_title> directive draws the title for the
document from a value set in the configuration file, allowing
different configuration files to use this template with
different titles.  Any configuration value can be substituted
in this way.

=item C<opt.*>

The C<opt.*> directive (not used in this example) substitutes
values passed to the C<NMStreq> object's constructor into
the output document.

=item C<env.*>

The C<env.*> directive substitutes the values of environment
variables.  Any non-printable characters will be removed
from the values using the strip_nonprintable() method.

=item C<param.*>

The C<param.*> directive substitutes the values of CGI
parameters.  Any non-printable characters will be removed
from the values using the strip_nonprintable() method.

=item C<date>

The C<date> directive outputs the current date, formatted
according to the C<date_fmt> configuration setting.

=back

=head1 SEE ALSO

L<NMSCharset>, L<CGI>

=head1 MAINTAINERS

The NMS project, E<lt>http://nms-cgi.sourceforge.net/E<gt>

To request support or report bugs, please email
E<lt>nms-cgi-support@lists.sourceforge.netE<gt>

=head1 COPYRIGHT

Copyright 2002 - 2004 London Perl Mongers, All rights reserved

=head1 LICENSE

This module is free software; you are free to redistribute it
and/or modify it under the same terms as Perl itself.

=cut

NMSAI_END_PM_FILE

    'NMSCharset.pm' => <<'NMSAI_END_PM_FILE',
package NMSCharset;
use strict;

=head1 NAME

NMSCharset - a charset-aware object for handling text strings

=head1 SYNOPSIS

   my $cs = NMSCharset->new('iso-8859-1');

   my $safe_to_put_in_html = $cs->escape($untrusted_user_input);

   my $printable = &{ $cs->strip_nonprint_coderef }( $input );
   my $escaped = &{ $cs->escape_html_coderef }( $printable );

=head1 DESCRIPTION

Each object of class C<NMSCharset> is bound to a particular
character set when it is created.  The object provides methods
to generate coderefs to perform a couple of character set
dependent operations on text strings.

=cut

=head1 CONSTRUCTORS

=over

=item new ( CHARSET )

Creates a new C<NMSCharset> object, suitable for handing text
in the character set CHARSET.  The CHARSET parameter must be a
character set string, such as C<us-ascii> or C<utf-8> for example.

=cut

sub new
{
   my ($pkg, $charset) = @_;

   my $self = { CHARSET => $charset };

   if ($charset =~ /^utf-8$/i)
   {
      $self->{SN} = \&_strip_nonprint_utf8;
      $self->{EH} = \&_escape_html_utf8;
   }
   elsif ($charset =~ /^iso-8859/i)
   {
      $self->{SN} = \&_strip_nonprint_8859;
      if ($charset =~ /^iso-8859-1$/i)
      {
         $self->{EH} = \&_escape_html_8859_1;
      }
      else
      {
         $self->{EH} = \&_escape_html_8859;
      }
   }
   elsif ($charset =~ /^us-ascii$/i)
   {
      $self->{SN} = \&_strip_nonprint_ascii;
      $self->{EH} = \&_escape_html_8859_1;
   }
   else
   {
      $self->{SN} = \&_strip_nonprint_weak;
      $self->{EH} = \&_escape_html_weak;
   }

   return bless $self, $pkg;
}

=back

=head1 METHODS

=over

=item charset ()

Returns the CHARSET string that was passed to the constructor.

=cut

sub charset
{
   my ($self) = @_;

   return $self->{CHARSET};
}

=item escape ( STRING )

Returns a copy of STRING with runs of non-printable characters
replaced with spaces and HTML metacharacters replaced with the
equivalent entities.

If STRING is undef then the empty string will be returned.

=cut

sub escape
{
   my ($self, $string) = @_;

   return &{ $self->{EH} }(  &{ $self->{SN} }($string)  );
}

=item strip_nonprint_coderef ()

Returns a reference to a sub to replace runs of non-printable
characters with spaces, in a manner suited to the charset in
use.

The returned coderef points to a sub that takes a single readonly
string argument and returns a modified version of the string.  If
undef is passed to the function then the empty string will be
returned.

=cut

sub strip_nonprint_coderef
{
   my ($self) = @_;

   return $self->{SN};
}

=item escape_html_coderef ()

Returns a reference to a sub to escape HTML metacharacters in
a manner suited to the charset in use.

The returned coderef points to a sub that takes a single readonly
string argument and returns a modified version of the string.

=cut

sub escape_html_coderef
{
   my ($self) = @_;

   return $self->{EH};
}

=back

=head1 PRIVATE TABLES

=over

=item C<%eschtml_map>

The C<%eschtml_map> hash maps C<iso-8859-1> characters to the
equivalent HTML entities.

=cut

use vars qw(%eschtml_map);
%eschtml_map = ( 
                 ( map {chr($_) => "&#$_;"} (0..255) ),
                 '<' => '&lt;',
                 '>' => '&gt;',
                 '&' => '&amp;',
                 '"' => '&quot;',
               );

=back

=head1 PRIVATE FUNCTIONS

These functions are returned by the strip_nonprint_coderef() and
escape_html_coderef() methods and invoked by the escape() method.
The function most appropriate to the character set in use will be
chosen.

=over

=item _strip_nonprint_utf8

Returns a copy of STRING with everything but printable C<us-ascii>
characters and valid C<utf-8> multibyte sequences replaced with
space characters.

=cut

sub _strip_nonprint_utf8
{
   my ($string) = @_;
   return '' unless defined $string;

   $string =~
   s%
    ( [\t\n\040-\176]               # printable us-ascii
    | [\xC2-\xDF][\x80-\xBF]        # U+00000080 to U+000007FF
    | \xE0[\xA0-\xBF][\x80-\xBF]    # U+00000800 to U+00000FFF
    | [\xE1-\xEF][\x80-\xBF]{2}     # U+00001000 to U+0000FFFF
    | \xF0[\x90-\xBF][\x80-\xBF]{2} # U+00010000 to U+0003FFFF
    | [\xF1-\xF7][\x80-\xBF]{3}     # U+00040000 to U+001FFFFF
    | \xF8[\x88-\xBF][\x80-\xBF]{3} # U+00200000 to U+00FFFFFF
    | [\xF9-\xFB][\x80-\xBF]{4}     # U+01000000 to U+03FFFFFF
    | \xFC[\x84-\xBF][\x80-\xBF]{4} # U+04000000 to U+3FFFFFFF
    | \xFD[\x80-\xBF]{5}            # U+40000000 to U+7FFFFFFF
    ) | .
   %
    defined $1 ? $1 : ' '
   %gexs;

   #
   # U+FFFE, U+FFFF and U+D800 to U+DFFF are dangerous and
   # should be treated as invalid combinations, according to
   # http://www.cl.cam.ac.uk/~mgk25/unicode.html
   #
   $string =~ s%\xEF\xBF[\xBE-\xBF]% %g;
   $string =~ s%\xED[\xA0-\xBF][\x80-\xBF]% %g;

   return $string;
}

=item _escape_html_utf8 ( STRING )

Returns a copy of STRING with any HTML metacharacters
escaped.  Escapes all but the most commonly occurring C<us-ascii>
characters and bytes that might form part of valid C<utf-8>
multibyte sequences.

=cut

sub _escape_html_utf8
{
   my ($string) = @_;

   $string =~ s|([^\w \t\r\n\-\.\,\x80-\xFD])| $eschtml_map{$1} |ge;
   return $string;
}

=item _strip_nonprint_weak ( STRING )

Returns a copy of STRING with sequences of NULL characters
replaced with space characters.

=cut

sub _strip_nonprint_weak
{
   my ($string) = @_;
   return '' unless defined $string;

   $string =~ s/\0+/ /g;
   return $string;
}
   
=item _escape_html_weak ( STRING )

Returns a copy of STRING with any HTML metacharacters escaped.
In order to work in any charset, escapes only E<lt>, E<gt>, C<">
and C<&> characters.

=cut

sub _escape_html_weak
{
   my ($string) = @_;

   $string =~ s/[<>"&]/$eschtml_map{$1}/eg;
   return $string;
}

=item _escape_html_8859_1 ( STRING )

Returns a copy of STRING with all but the most commonly
occurring printable characters replaced with HTML entities.
Only suitable for C<us-ascii> or C<iso-8859-1> input.

=cut

sub _escape_html_8859_1
{
   my ($string) = @_;

   $string =~ s|([^\w \t\r\n\-\.\,])| $eschtml_map{$1} |ge;
   return $string;
}

=item _escape_html_8859 ( STRING )

Returns a copy of STRING with all but the most commonly
occurring printable C<us-ascii> characters and characters
that might be printable in some C<iso-8859-*> charset
replaced with HTML entities.

=cut

sub _escape_html_8859
{
   my ($string) = @_;

   $string =~ s|([^\w \t\r\n\-\.\,\240-\377])| $eschtml_map{$1} |ge;
   return $string;
}

=item _strip_nonprint_8859 ( STRING )

Returns a copy of STRING with runs of characters that are not
printable in any C<iso-8859-*> charset replaced with spaces.

=cut

sub _strip_nonprint_8859
{
   my ($string) = @_;
   return '' unless defined $string;

   $string =~ tr#\t\n\040-\176\240-\377# #cs;
   return $string;
}

=item _strip_nonprint_ascii ( STRING )

Returns a copy of STRING with runs of characters that are not
printable C<us-ascii> replaced with spaces.

=cut

sub _strip_nonprint_ascii
{
   my ($string) = @_;
   return '' unless defined $string;

   $string =~ tr#\t\n\040-\176# #cs;
   return $string;
}

=back

=head1 MAINTAINERS

The NMS project, E<lt>http://nms-cgi.sourceforge.net/E<gt>

To request support or report bugs, please email
E<lt>nms-cgi-support@lists.sourceforge.netE<gt>

=head1 COPYRIGHT

Copyright 2002 - 2004 London Perl Mongers, All rights reserved

=head1 LICENSE

This module is free software; you are free to redistribute it
and/or modify it under the same terms as Perl itself.

=cut

1;

NMSAI_END_PM_FILE

);

######################################################################

use vars qw($ai_auth_done %Probe $ai_cgi_name %Config);
$ai_auth_done = 0;
%Probe = ();
$ai_cgi_name = '';
%Config = ();

eval {

    if (length $pw < 7) {
        ai_usererror(<<END);
<p>The password configured in this autoinstall script is too short to
be secure.  Please choose a password of at least 7 characters and
upload this autoinstall script again.</p>
END
    }
    elsif ($pw =~ /pa[s5][s5]|w[0o]rd/i) {
        ai_usererror(<<END);
<p>The password configured in this autoinstall script is too similar
to the default value of 'password' to be secure.  Please choose a
password that's nothing like the word 'password' and upload this
autoinstall script again.</p>
END
    }
    
    ai_probe_and_create_nmsai_files();
    
    my $lockfile = "$Probe{BASEDIR_ROOT}/.lock";
    my $lock = IO::File->new(">>$lockfile");
    defined $lock or ai_syserror("open >>$lockfile: $!");
    flock($lock, LOCK_EX) or ai_syserror("flock $lockfile: $!");
    
    if (-s $lockfile > 5) {
        ai_usererror(<<END);
<p>
This autoinstall script is disabled, because there have been too many
incorrect password attempts.
</p><p>
To re-enable it, you must go to the directory on the web server where
the autoinstall CGI is, go into the <tt>.nmsai</tt> subdirectory, go
into the subdirectory that you find there (there should only be one)
and delete the file called <tt>.lock</tt>.
</p>
END
    }
    
    ai_password_page() unless $ENV{REQUEST_METHOD} eq 'POST';
    
    local $CGI::DISABLE_UPLOADS = 1;
    local $CGI::POST_MAX        = 1000000;
    my $cgi = CGI->new;
    
    my $gotpass = $cgi->param('_nmsai_pw') || '';
    if ($gotpass eq $pw) {
        truncate LOCK, 0;
        $ai_auth_done = 1;
    }
    else {
        print $lock 'x';
        $lock->close;
        ai_password_page();
    }
    $Probe{PASSWORD} = $pw;
    
    $ai_cgi_name = $cgi->param('_nmsai_cgi_name') || '';
    unless ($ai_cgi_name =~ /^([\w\-]+)$/) {
        ai_choose_cgi_name_page('TFmail', $Probe{CGI_EXT});
    }
    $ai_cgi_name = $1;
    $Probe{BASEDIR} = $Probe{BASEDIR_ROOT} . "/$ai_cgi_name";
    $Probe{REL_BASEDIR} = ".nmsai/$Probe{RAND_NAMED_DIR}/$ai_cgi_name";
    
    unless ( app_get_config_from_inputs( $cgi) ) {
        ai_load_config();
        app_fill_in_missing_defaults();
    }
    
    my $is_valid = app_config_is_valid();
    
    if ($is_valid and $cgi->param('_nmsai_do_install')) {
        ai_save_config();
        app_prepare_for_install();
        ai_do_install();
        ai_success_page();
    }
    else {
        ai_config_edit_page();
    }

};

ai_scripterror("runtime error [$@]") if $@;

####################################################################

sub ai_password_page {
    ai_html_header();

    print <<END;
<p>
Please enter the password that you put in the autoinstall script
before you uploaded it.
</p><p>
<input type="password" name="_nmsai_pw" />
<input type="submit" value="Continue" />
</p>
</form>
</body>
</html>
END

    exit;
}

sub ai_choose_cgi_name_page {
    my ($default, $ext) = @_;

    ai_html_header();
    print <<END;
<p>
Please choose the name under which the <tt>$default</tt> CGI script
will be installed:
<input type="text" name="_nmsai_cgi_name" value="$default"
/><tt>$ext</tt>
</p>
<p><input type="submit" value="Continue" /></p>
<p>Note that the CGI name that you put in the box may contain
letters, numbers, underscores and minus characters only.</p>
</form>
</body>
</html>
END

    exit;
}

sub ai_load_config {

    my %all_config = ();
    %Config = ();

    ai_load_all_config(\%all_config);

    # Configuration variables such as 'postmaster' go into the config
    # hash for this application, to act as defaults.
    foreach my $k (keys %all_config) {
        next if $k =~ /:/;
	$Config{$k} = $all_config{$k};
    }

    # Configuration variables such as 'foo:postmaster' override
    # configuration variables such as 'postmaster' if and only if we
    # are installing a CGI called 'foo'.
    foreach my $k (keys %all_config) {
        next unless $k =~ /^\Q$ai_cgi_name\E:([^:]+)$/;
	$Config{$1} = $all_config{$k};
    }
}

sub ai_load_all_config {
    my ($all_config) = @_;

    my $in = IO::File->new("<$Probe{BASEDIR_ROOT}/.config");
    defined $in or return;
    local $/ = "\0";
    while (1) {
        my $key = <$in>;
      last unless defined $key;
	chomp $key;
	$key =~ /^([\w\-\:]+)$/ or ai_scripterror("bad config key [$key]");
	$key = $1;

	my $val = <$in>;
	defined $val or ai_scripterror("key [$key] lacks a value");
	chomp $val;
	
	$all_config->{$key} = $val;
    }
    $in->close;
}

sub ai_save_config {
    my %all_config = ();
    ai_load_all_config(\%all_config);
    foreach my $k (keys %Config) {
        $all_config{$k} = $Config{$k};
	$all_config{"$ai_cgi_name:$k"} = $Config{$k};
    }
    ai_writefile("$Probe{BASEDIR_ROOT}/.config", join("\0",%all_config)."\0", 0644);
}

sub ai_config_edit_page {

    ai_html_header();
    app_edit_config_html();
    print <<END;
<p><input type="submit" value="Check this configuration" /></p>
END

    if ( app_config_is_valid() ) {
        print <<END;
<input type="submit" name="_nmsai_do_install"
value="Install with this configuration" />
END
    }

    print <<END;
</form>
</body>
</html>
END

    exit;
}

sub ai_success_page {

    ai_html_header('noform');
    print "<h2>Installation successful</h2>\n";

    if (unlink $Probe{ME}) {
        print <<END;
<p>The autoinstall script has completed the installation and deleted
itself.</p>
END
    }
    else {
        print <<END;
<p>The autoinstall script completed the installation OK, but was
unable to delete itself afterwards.  Please delete the autoinstall
script now.</p>
END
    }

    print <<END;
<p>If you had to change the permissions on the directory into which
you uploaded the autoinstall script in order to get it to work, you
should now change them back to what they were before.</p>
END

    app_success_html();
    print <<END;
</body>
</html>
END

    exit;
}

sub ai_html_header {
    my ($no_form) = @_;

    print <<END;
Content-Type: text/html; charset=iso-8859-1

<?xml version="1.0" encoding="iso-8859-1"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
 <title>NMS TFmail AutoInstall</title>
</head>
<body>
END

    ai_form_header() unless $no_form;
}

sub ai_form_header {
    print qq{<form method="post">\n};
    if ($ai_auth_done) {
        print qq|<input type="hidden" name="_nmsai_pw" value="${\( ai_eschtml($Probe{PASSWORD}) )}" />\n|;
    }
    if ($ai_cgi_name) {
        print qq|<input type="hidden" name="_nmsai_cgi_name" value="${\( ai_eschtml($ai_cgi_name) )}" />\n|;
    } 
}

sub ai_do_install {

    foreach my $module (keys %Modules) {
        my $filename = "$Probe{LIBDIR}/$module";
        ai_mkdir_p_dirname($filename);
	ai_writefile($filename, $Modules{$module}, 0644);
    }

    foreach my $cgi (keys %Scripts) {
        my $install_as = $ai_cgi_name .$cgi;
	my $filename = "$Probe{CGI_BIN}/$install_as$Probe{CGI_EXT}";
	ai_writefile($filename, $Scripts{$cgi}, 0755);
    }
}

sub ai_probe_and_create_nmsai_files {
    
    my $me = $ENV{SCRIPT_FILENAME} || $ENV{PATH_TRANSLATED};
    $me =~ tr#\\#/#;
    $me =~ m#^([\w\-\.\/ \:]+)$# or ai_scripterror("failed to get script filename");
    $Probe{ME} = $me = $1;

    $Probe{CGI_EXT} = '';
    if ($me =~ m#(\.\w{1,10})$#) {
        $Probe{CGI_EXT} = $1;
    }

    $me =~ m#^([\w\-\.\/ \:]+)/[^/]+$# or ai_scripterror("failed to get cgi-bin directory");
    $Probe{CGI_BIN} = $1;

    my $uri = $ENV{SCRIPT_NAME};
    defined $uri or ai_scripterror("failed to get script URI");
    $uri =~ s#/[^/]+$## or ai_scripterror("failed to get cgi-bin directory uri");
    $Probe{CGI_BIN_URI} = $uri;

    $Probe{SHEBANG} = '';
    my $this_script = IO::File->new("<$me");
    defined $this_script or ai_scripterror("failed to open myself for read: $!");
    my $line = <$this_script>;
    $this_script->close;
    if ($line =~ /^(\#\!.*\n)$/) {
        $Probe{SHEBANG} = $1;
    }

    my $root = "$Probe{CGI_BIN}/.nmsai";
    unless (-d $root or mkdir $root, 0755) {
        ai_usererror(<<END);
<p>
The autoinstall script failed to create the <tt>.nmsai</tt>
subdirectory, which it needs.  The error message received from the
operating system was <tt>${\( ai_eschtml($!) )}</tt>.
</p><p>
The most likely cause of this problem is that the web server process
lacks write permission on the directory into which you uploaded the
autoinstall CGI.  You must give the web server process write
permission on that directory before the installation can continue.
Depending on server configuration, that may mean making the
directory world writable (mode 0777).
</p>
END
    }

    my $subdir = ai_find_subdir($root);
    unless (defined $subdir) {
        $subdir = ai_random_string();
    }
    $Probe{RAND_NAMED_DIR} = $subdir;

    ai_mkdir_p("$root/$subdir/.lib");
    $Probe{BASEDIR_ROOT} = "$root/$subdir";
    $Probe{LIBDIR}       = "$root/$subdir/.lib";

    ai_writefile("$root/.htaccess", <<END, 0644);

Order deny allow
deny from all

END

    foreach my $index (qw(index.html index.htm default.htm welcome.htm)) {
        ai_writefile("$root/$index", <<END, 0644);
<html>
<head>Access Denied</head>
<body>
<p>You do not have permission to access this directory</p>
</body>
</html>
END
    }
}

sub ai_random_string {

    my $random = pack 'C6', map rand(256), (1..6);

    if ( my $rnd = IO::File->new("</dev/urandom") ) {
        my $devrnd;
        $rnd->read($devrnd, 6);
	$random ^= $devrnd;
    }

    my @chars = ('a'..'z', 'A'..'Z', '0'..'9', '-', '_');
    my $buf = unpack 'b*', $random;
    $buf =~ s/([01]{1,6})/$chars[ord pack 'b*', $1]/ge;

    $buf =~ /^([\w\-]{8})$/ or ai_scripterror("bad rnd [$buf]");
    return $1;
}

sub ai_find_subdir {
    my ($dir) = @_;

    opendir D, $dir;
    my @hits = grep /^[\w\-]{4,20}$/, readdir D;
    return undef unless @hits;
    if (@hits > 1) {
         ai_scripterror( "multiple subdirs of [$dir]: ".join(':',@hits) );
    }
    $hits[0] =~ /^([\w\-]{4,20})$/ or ai_scripterror("bad subdir [$hits[0]]");
    return $1;
}

sub ai_writefile {
    my ($filename, $contents, $mode) = @_;

    my $out = IO::File->new(">$filename.tmp");
    defined $out or ai_syserror("open >$filename.tmp: $!");
    $out->print($contents) or ai_syserror("write to $filename.tmp: $!");
    $out->close or ai_syserror("close $filename.tmp: $!");

    chmod $mode, "$filename.tmp" or ai_syserror("chmod $filename.tmp: $!");
    rename "$filename.tmp", $filename or ai_syserror("rename [$filename.tmp]: $!");
}

sub ai_mkdir_p_dirname {
    my ($filename) = @_;
    $filename =~ s#/[^/]+$## or ai_scripterror("dirname [$filename]");
    ai_mkdir_p($filename);
}

sub ai_mkdir_p {
    my ($dir) = @_;
    my $orig_dir = $dir;

    my @dirs = ();
    until (-d $dir) {
        $dir =~ s#/([^/]+)$## or ai_scripterror("mkdir_p [$orig_dir]: down to [$dir]");
	unshift @dirs, $1;
    }

    foreach my $d (@dirs) {
        $dir .= "/$d";
	mkdir $dir, 0755 or ai_syserror("mkdir [$dir]: $!");
    }

    -d $orig_dir or ai_scripterror("mkdir_p failed to build [$orig_dir]");
}

sub ai_usererror {
    my ($msg) = @_;

    ai_html_header('noform');
    print <<END;
<h1>Error</h1>
$msg
<hr />
</body>
</html>
END

    exit;
}

sub ai_syserror {
    my ($msg) = @_;

    ai_html_header('noform');
    print <<END;
<h1>Server Fault</h1>
<p>
An error has occurred, and it appears to be some sort of fault on
the server.  It could be that the server's hard disk is full, your
account has exceeded its disk space quota or the permissions or
ownership of a file has changed unexpectedly.
</p>
END

    ai_output_diagnostics($msg);
    print <<END;
<hr />
</body>
</html>
END

    exit;
}

sub ai_scripterror {
    my ($msg) = @_;

    ai_html_header('noform');
    print <<END;
<h1>Installation Script Fault</h1>
<p>
An error has occurred, and it appears to be the result of a fault or
shortcoming of this installation script.
</p>
END

    ai_output_diagnostics($msg);
    print <<END;
<hr />
</body>
</html>
END

    exit;
}

sub ai_output_diagnostics {
    my ($msg) = @_;

    if ($ai_auth_done) {
        print <<END;
<p>The error that occurred was: <tt>${\( ai_eschtml($msg) )}</tt>.</p>
<p>
If you choose to report this error to the script maintainers, please
include the following diagnostic information:
</p>
END
	foreach my $k (keys %ENV) {
            print "<b>${\( ai_eschtml($k) )}</b>: ${\( ai_eschtml($ENV{$k}) )}<br />\n";
        }
    }
    else {
	print STDERR "NMS autoinstall error details: [$msg]\n";
	print <<END;
<p>Details of the error can be found in the web server's error log.</p>
END
    }
}

sub ai_eschtml {
    my ($input) = @_;

    return CGI::escapeHTML($input);
}


sub app_config_is_valid {
        ( app__sendmail_command_ok() or app__smtp_relay_ok() )
   and  app__postmaster_ok()
   and  app__charset_ok()
   ;
}

sub app__sendmail_command_ok {
    my $sc = $Config{sendmail_command} || '';
    return 0 unless length $sc;
    return 0 unless $sc =~ m#^(\S+)#;
    return 1;
}

sub app__smtp_relay_ok {
    my $relay = $Config{smtp_relay} || '';
    return 0 unless $relay =~ /^[\w\-\.]+$/;
    return 1;
}

sub app__postmaster_ok {
    my $pm = $Config{postmaster_address} || '';
    return 0 unless $pm =~ m#^[\w\-\.\+]{1,80}\@[\w\-\.]{1,100}$#;
    return 1;
}

sub app__charset_ok {
    $Config{charset} =~ /^[\w\-]+$/ ? 1 : 0;
}

sub app_edit_config_html {
    my ($sme, $pme, $cse) = ('', '', '');

    unless (app__sendmail_command_ok() or app__smtp_relay_ok()) {
        $sme = <<END;
<p>
 <font color="red">You must set either <b>sendmail location</b> or
 <b>SMTP relay</b> above.  Your system administrator or hosting provider
 should be able to tell you how to set one or other of these for your
 web server.</font>
</p>
END
    }

    unless (app__postmaster_ok()) {
        $pme = <<END;
<p><font color="red">You must enter a valid email address here.</font></p>
END
    }

    unless (app__charset_ok()) {
        $cse = <<END;
<p><font color="red">You must enter the name of a character set here, e.g.
<tt>iso-8859-1</tt>.</font></p>
END
    }

    print <<END;
<h1>Installation Settings</h1>
<p>
 Please adjust the installation settings below until you're happy with them.
 Text in <font color="red">red</font> marks places where you need to make a
 change before TFmail can be installed.
</p>
<hr />
<p>
 <b>Sendmail location</b>: the location of the <tt>sendmail</tt> program on
 the web server, e,g <tt>/usr/lib/sendmail</tt>.  If this web server has no
 <tt>sendmail</tt> program then leave this blank and set <b>SMTP relay</b>
 below instead.
</p>
<p>
 <input type="text" value="${\( ai_eschtml($Config{sendmail_command}) )}"
 name="sendmail_command" size="80" />
</p>
<hr />
<p>
 <b>SMTP relay</b>: the host name or IP address of an SMTP server that will
 relay outgoing emails for this web server.  You only need to set this if
 there is no <tt>sendmail</tt> program on the web server.
</p>
<p>
 <input type="text" value="${\( ai_eschtml($Config{smtp_relay}) )}"
 name="smtp_relay" size="50" />
</p>
$sme
<hr />
<p>
 <b>Postmaster Address</b>: the address to use as the envelope sender of
 mail sent by the script.  If in dought, put your own email address here.
</p>
<p>
 <input type="text" value="${\( ai_eschtml($Config{postmaster_address}) )}"
 name="postmaster_address" size="80" />
</p>
$pme
<hr />
<p>
 <input type="checkbox" name="debugging"
 ${\( $Config{debugging} ? ' checked="checked" value="1"' : '' )} />
 Tick here if you want debugging mode switched on.  You should have this
 on if you're having trouble getting the script working, but turn it off
 for normal operation.
</p>
<p>
 <input type="checkbox" name="enable_uploads"
 ${\( $Config{enable_uploads} ? ' checked="checked" value="1"' : '' )} />
 Tick here if you want to be able to set up forms where the user uploads
 a file and it comes through to you as an attachment to the email.
</p>
<p>
 <input type="checkbox" name="use_mime_lite"
 ${\( $Config{use_mime_lite} ? ' checked="checked" value="1"' : '' )} />
 Tick here if you want <tt>TFmail</tt> to use the <tt>MIME::Lite</tt>
 Perl module to send emails.  With this ticked your emails are less
 likely to be scrambled by mail servers but the script may start up
 slower on some systems.
</p>
<hr />
<p>
 <b>Character set</b>: <input type="text" size="20" name="charset"
 value="${\( ai_eschtml($Config{charset}) )}" />
</p>
$cse
END
}

sub app_get_config_from_inputs {
    my ($cgi) = @_;

    defined $cgi->param('postmaster_address') or return 0;

    foreach my $p (qw(postmaster_address sendmail_command smtp_relay charset)) {
        $Config{$p} = $cgi->param($p) || '';
    }

    foreach my $p (qw(debugging enable_uploads use_mime_lite)) {
        $Config{$p} = ($cgi->param($p) ? 1 : 0);
    }

    return 1;
}

sub app_fill_in_missing_defaults {
    
    unless ($Config{sendmail_command}) {
        foreach my $bin (qw(/usr/sbin/sendmail /usr/lib/sendmail
	                    /usr/bin/sendmail /bin/sendmail
                        /var/qmail/bin/sendmail
                        )) {
            if (-x $bin and not -d $bin) {
	            $Config{sendmail_command} = $bin;
              last;
            }
        }
    }

    $Config{debugging} = '1' unless $Config{debugging} eq '0';
    $Config{enable_uploads} = '0' unless $Config{enable_uploads} eq '1';
    $Config{use_mime_lite} = '1' unless $Config{use_mime_lite} eq '0';
    $Config{charset} |= 'iso-8859-1';
}

sub app_prepare_for_install {
    
    my $mailprog;
    if ( length $Config{sendmail_command} ) {
        $mailprog = $Config{sendmail_command};
        $mailprog .= " -oi -t" unless $mailprog =~ /\S\s+\S/;
    }
    else {
        $mailprog = "SMTP:$Config{smtp_relay}";
    }

    app__subst_constant('DEBUGGING',      $Config{debugging});
    app__subst_constant('LIBDIR',         $Probe{LIBDIR});
    app__subst_constant('MAILPROG',       $mailprog);
    app__subst_constant('POSTMASTER',     $Config{postmaster_address});
    app__subst_constant('CONFIG_ROOT',    "$Probe{BASEDIR}/cfg");
    app__subst_constant('LOGFILE_ROOT',   "$Probe{BASEDIR}/log");
    app__subst_constant('ENABLE_UPLOADS', $Config{enable_uploads});
    app__subst_constant('USE_MIME_LITE',  $Config{use_mime_lite});
    app__subst_constant('CHARSET',        $Config{charset});
    app__subst_constant('LOCKFILE',       "$Probe{BASEDIR}/.lock");
    app__subst_constant('PASSWORD',       $Probe{PASSWORD});

    foreach my $script (keys %Scripts) {
	$Scripts{$script} =~ s/^\#\!.*\n/$Probe{SHEBANG}/;
    }

    ai_mkdir_p("$Probe{BASEDIR}/cfg");
    ai_mkdir_p("$Probe{BASEDIR}/log");

    my $v = eval q{ require MIME::Lite ; $MIME::Lite::VERSION };
    if (defined $v and $v =~ /^2/) {
        delete $Modules{'MIME/Lite.pm'};
    }
}

sub app__subst_constant {
    my ($constant, $value) = @_;

    $value =~ s#([\\'])#\\$1#g;
    foreach my $script (keys %Scripts) {
        $Scripts{$script} =~ s#(\nuse constant \Q$constant\E[ \t]+\=\>)[^\n]+#$1 '$value';#;
    }
}

sub app_success_html {

    my $config_edit = "$Probe{CGI_BIN_URI}/${ai_cgi_name}Config$Probe{CGI_EXT}";

    print <<END;
<form method="post" action="${\( ai_eschtml($config_edit) )}">
<p>
<input type="hidden" name="password" value="${\( ai_eschtml( $Probe{PASSWORD} ) )}" />
The <input type="submit" value="TFmail Interactive Configuration Editor" />
has been installed as <tt>${\( ai_eschtml($config_edit) )}</tt>.  The password
for the configuration editor is the same as the activation password of this
autoinstall script.
</p>
</form>
<p>
If you prefer to edit the TFmail configuration files using a text editor
and then upload them, you should upload all <tt>.trc</tt> and <tt>.trt</tt>
files to the <tt>${\( ai_eschtml("$Probe{REL_BASEDIR}/cfg") )}</tt>
subdirectory of the directory into which you uploaded this autoinstall
script.
</p>
END
}

